import shlex
import subprocess
import tarfile
import os
import yaml

from kubernetes.client.rest import ApiException

from si_tests.deployments.utils import commons, file_utils, kubectl_utils
from si_tests.deployments.utils.namespace import NAMESPACE
from si_tests.utils import utils, waiters
from si_tests import settings

OS_NAMESPACE = NAMESPACE.openstack


def turn_off_service(os_manager, service):
    osdpl = os_manager.get_osdpl_deployment()
    osdpl_data = osdpl.read()
    services = osdpl_data.spec["features"].get("services", [])
    services = list(filter(lambda x: not x == service, services))

    commons.LOG.info("Turn off old %s", service)
    osdpl.patch({"spec": {"features": {"services": services}}})

    # Wait for kubernetes to delete old service
    def _wait_hb_absence():
        helmbundles = os_manager.get_os_helmbundles()
        helmbunle_name = 'openstack-{}'.format(service)
        hb_exist = list(filter(
            lambda x: x.name == helmbunle_name,
            helmbundles))
        if not hb_exist:
            return True
        return False

    waiters.wait(_wait_hb_absence, interval=10, timeout=180)
    return services


def run_tests(os_manager, report_name, run_pod, service, get_run_command):
    run_command = get_run_command(report_name)

    services = turn_off_service(os_manager, service)

    osdpl = os_manager.get_osdpl_deployment()

    svcs_body = {}
    custom_flavor = settings.TEMPEST_CUSTOM_FLAVOR
    custom_image = settings.TEMPEST_CUSTOM_IMAGE
    custom_tempest_parameters = settings.TEMPEST_CUSTOM_PARAMETERS
    custom_tempest_config = yaml.safe_load(custom_tempest_parameters)

    if custom_flavor:
        svcs_body = utils.merge_dicts(svcs_body, {"spec": {"services": {
            service: {service: {"values": {"conf": {"convert_to_uuid": {
                "compute": {"flavor_ref": custom_flavor,
                            'flavor_ref_alt': custom_flavor}}}}}}}}})
    if custom_image:
        svcs_body = utils.merge_dicts(svcs_body, {"spec": {"services": {
            service: {service: {"values": {"conf": {"convert_to_uuid": {
                "compute": {"image_ref": custom_image}}}}}}}}})
    if custom_tempest_parameters:
        svcs_body = utils.merge_dicts(svcs_body, {"spec": {"services": {
            service: {service: {"values": {"conf": {
                "tempest": custom_tempest_config}}}}}}})
    if run_command:
        commons.LOG.info("Explicitly set %s command", service)
        svcs_body = utils.merge_dicts(svcs_body, {"spec": {"services": {
            service: {service: {"values": {"conf": {
                "script": run_command}}}}}}})
        osdpl.patch(svcs_body)
    else:
        msg = "TestScheme was not defined or incorrect. " \
              "Tests will be run with default script from Chart."
        commons.LOG.warning(msg)
        if svcs_body:
            osdpl.patch(svcs_body)

    commons.LOG.info("Turn on %s", service)
    services.append(service)
    osdpl.patch({"spec": {"features": {"services": services}}})

    def _wait_job():
        job = os_manager.api.jobs.get(
            run_pod, namespace=OS_NAMESPACE)
        data = job.read()
        commons.LOG.info(data.status)
        if data.status.failed:
            raise ValueError(
                "Job {} failed to complete".format(job.name))
        assert data.status.succeeded == 1

    try:
        waiters.wait_pass(
            _wait_job, interval=60, timeout=8 * 3600,
            expected=(AssertionError, ApiException))
    finally:
        job = os_manager.api.pods.list(
            name_prefix=run_pod, namespace=OS_NAMESPACE)[0]
        logs = job._manager.api.read_namespaced_pod_log(
            name=job.name, namespace=OS_NAMESPACE)
        commons.LOG.info(logs)


def save_report(os_manager, report_name, results_pod, volume_log_dir,
                log_name=None):

    commons.LOG.info("Collect artifacts from pvc")
    kubectl = kubectl_utils.Kubectl()

    commons.LOG.info("Create pod to collect artifacts from PVC")
    basepath = os.path.dirname(os.path.abspath(__file__))
    pod_yaml = file_utils.join(
        basepath, "../yamls/{}.yaml".format(results_pod))
    kubectl.apply(pod_yaml, cwd=".")

    def _wait_pod():
        result_pod = os_manager.api.pods.get(
            results_pod, namespace=OS_NAMESPACE)
        assert result_pod.read().status.phase == "Running"
        return result_pod

    waiters.wait_pass(_wait_pod, interval=5, timeout=240)

    artifacts = [report_name]
    if log_name:
        artifacts.append(log_name)

    for artifact in artifacts:
        commons.LOG.info("Copy %s to local machine", artifact)
        source = "{}/{}:{}{}".format(
            OS_NAMESPACE, results_pod, volume_log_dir, artifact)
        destination = file_utils.get_artifact(artifact)
        try:
            kubectl.cp(source, destination)
        except Exception:
            commons.LOG.exception("File not found")

    commons.LOG.info("Delete results pod")
    kubectl.delete("", "-f {}".format(pod_yaml), OS_NAMESPACE, cwd=".")


def convert_results_to_xml(report_name):
    commons.LOG.info("Convert report from subunit to junit xml")
    report_path = file_utils.get_artifact(report_name)
    with open("{}.xml".format(report_path), "w") as f:
        p = subprocess.Popen(
            shlex.split("subunit2junitxml {}".format(report_path)),
            stdout=f)
        p.wait(timeout=120)

    subprocess.run(shlex.split("rm {}".format(report_path)))


def archive_log(log_name):
    log_path = file_utils.get_artifact(log_name)
    if os.path.exists(file_utils.get_artifact(log_name)):
        commons.LOG.info("Archive %s", log_name)
        archive = file_utils.get_artifact("{}.tar.gz".format(log_name))
        with tarfile.open(archive, "w:gz") as tar:
            tar.add(log_path)
    else:
        commons.LOG.info("%s not present in artifacts", log_name)
