import os

from time import sleep
import unittest
import tempfile
from retry import retry

import paramiko
import yaml
import json
from exec_helpers import SSHClient, SSHAuth, \
    CalledProcessError, ExecHelperTimeoutError, Subprocess
from packaging import version
from si_tests import logger
from si_tests import settings
from si_tests.clients.iam import iam_client
from si_tests.clients.iam import keycloak_client
from si_tests.clients.k8s import K8sCluster
from si_tests.managers.kaas_manager import Manager
from si_tests.managers.aws_manager import AwsManager
from si_tests.utils import templates, utils, waiters, helpers

LOG = logger.logger


class BootstrapManager(unittest.TestCase):
    def get_seed_ip(self, os_client, fip='seed_floating_ip'):
        if settings.SEED_STANDALONE_EXTERNAL_IP and fip == 'seed_floating_ip':
            LOG.debug("ssh into seed using SEED_STANDALONE_EXTERNAL_IP")
            return settings.SEED_STANDALONE_EXTERNAL_IP

        LOG.debug("getting seed ip via openstack client")
        stack_name = settings.ENV_NAME
        stack = os_client._current_stack
        if stack:
            outputs = stack.outputs
            key = next(
                (o for o in outputs if o["output_key"] == fip),
                {})
            if key:
                return key['output_value']
            else:
                LOG.warning(f"Stack: {stack_name} doesnt contain requested"
                            f"key: {fip}")
                return None
        LOG.warning("Stack {0} isn't created".format(stack_name))
        return None

    def get_seed_instance(self, os_client):
        resources = os_client._get_resources_by_type("OS::Nova::Server")
        if not resources:
            return None

        resource = resources[0]
        instance = os_client.get_server_by_id(resource.attributes['id'])

        return instance

    def get_kaas_release_file_name(self):
        kaas_release_file = settings.KAAS_RELEASE_FILE
        if kaas_release_file is None:
            kaas_release_file = ''

        if kaas_release_file:
            if kaas_release_file.lower() != 'stable':
                LOG.info('KAAS_RELEASE_FILE has been passed:{}'.format(
                    kaas_release_file))
                return kaas_release_file

        def ver(n):
            return version.Version(os.path.splitext(n)[0])

        if settings.KAAS_RELEASES_FOLDER and \
                os.path.isdir(settings.KAAS_RELEASES_FOLDER):
            kaas_releases_folder = os.path.join(
                settings.KAAS_RELEASES_FOLDER, 'kaas')
            list_of_kaas_yamls = os.listdir(kaas_releases_folder)
            list_of_kaas_yamls.sort(key=ver)
            LOG.debug("Found kaas releases files in {0} "
                      "- {1}".format(settings.KAAS_RELEASES_FOLDER,
                                     list_of_kaas_yamls))
        else:
            raise NameError(
                "Folder {} from variable"
                " settings.KAAS_RELEASES_FOLDER not exist".format(
                    settings.KAAS_RELEASES_FOLDER))

        if kaas_release_file.lower() == 'stable':
            latest_file = None
            for filename in reversed(list_of_kaas_yamls):
                if '-rc.' in filename:
                    continue
                latest_file = filename
                break
        else:
            latest_file = list_of_kaas_yamls[-1]

        LOG.info('Latest kaas release file is {}'.format(latest_file))
        return latest_file

    def get_bootstrap_version(self):
        kaas_release_filename = self.get_kaas_release_file_name()
        kaas_release_yml_path = os.path.join(settings.KAAS_RELEASES_FOLDER,
                                             'kaas',
                                             kaas_release_filename)
        with open(kaas_release_yml_path, 'r') as f:
            kaas_release_yml = yaml.load(f.read(), Loader=yaml.SafeLoader)

        bootstrap_version = kaas_release_yml['spec']['bootstrap']['version']
        LOG.info("Bootstrap version fetched from {} - {}".format(
            kaas_release_yml_path, bootstrap_version))
        return bootstrap_version

    def get_bootstrap_ssh_keyfile_path(self, remote):
        # deprecated path
        r_base = ""
        if remote.isfile("bootstrap/dev/kaas"):
            # baremetal seed custom dir
            r_base = "bootstrap/dev/"
        key_path = "{0}.ssh/openstack_tmp".format(r_base)

        # current rc solution
        key_proposal = os.path.join(settings.KAAS_BOOTSTRAP_TARGET_DIR,
                                    "ssh_key")
        # can't rely on kaas version here because
        # upgraded cluster may still use old key
        if remote.isfile(key_proposal):
            key_path = key_proposal

        return key_path

    def get_remote_default_kaas_data(self, remote, openstack_client):
        version_cmd = (
            "grep KAAS_RELEASE_YAML {}/bootstrap.env| "
            "cut -d '=' -f2 ".format(settings.KAAS_BOOTSTRAP_TARGET_DIR))
        LOG.info('Plan to execute {}'.format(version_cmd))
        release_file_path = remote.execute(version_cmd).stdout[0].rstrip()
        LOG.info("Kaas file from boot.env {}".format(
            release_file_path.decode("utf-8")))
        LOG.info("Check founded file exists")

        release_content = self.get_remote_file_yaml(
            openstack_client,
            filepath=release_file_path.decode("utf-8"))
        LOG.info("Content on kaas release:\n{0}".format(release_content))
        return release_content

    def get_bootstrap_templates_path(self, remote=None,
                                     openstack_client=None):
        if not settings.BOOTSTRAP_WITH_DEFAULTS:
            bootstrap_version = self.get_bootstrap_version()
        else:
            bootstrap_version = self.get_remote_default_kaas_data(
                remote, openstack_client)['spec']['bootstrap']['version']
            LOG.info("Get default version {}".format(bootstrap_version))
        bootstrap_templates_path = os.path.expanduser(
            settings.KAAS_BOOTSTRAP_TEMPLATES_DIR)
        bootstrap_templates_path = os.path.join(
            bootstrap_templates_path,
            'templates_{}'.format(bootstrap_version))

        LOG.info("Bootstrap template folder placed here - {}".format(
            bootstrap_templates_path))
        assert os.path.isdir(bootstrap_templates_path), \
            "Path {} not found".format(bootstrap_templates_path)
        return bootstrap_templates_path

    def get_kaas_version(self):
        kaas_release_filename = self.get_kaas_release_file_name()
        kaas_release_yml_path = os.path.join(settings.KAAS_RELEASES_FOLDER,
                                             'kaas',
                                             kaas_release_filename)
        with open(kaas_release_yml_path, 'r') as f:
            kaas_release_yml = yaml.load(f.read(), Loader=yaml.SafeLoader)

        kaas_version = kaas_release_yml['spec']['version']
        LOG.info("KaaS version fetched from {} - {}".format(
            kaas_release_yml_path, kaas_version))
        return kaas_version

    @retry(Exception, delay=1, tries=3, jitter=1, logger=LOG)
    def remote_seed(self, os_client):
        key = utils.load_keyfile(settings.SEED_SSH_PRIV_KEY_FILE)
        key = utils.get_rsa_key(key["private"])
        seed_ip = self.get_seed_ip(os_client)
        remote = SSHClient(
            host=seed_ip,
            port=22,
            auth=SSHAuth(
                username=settings.SEED_SSH_LOGIN,
                key=key))
        remote.logger.addHandler(logger.console)
        return remote

    def get_remote_file_yaml(self, os_client, filepath):
        key = utils.load_keyfile(settings.SEED_SSH_PRIV_KEY_FILE)
        seed_ip = self.get_seed_ip(os_client)
        with templates.YamlEditor(
                file_path=filepath,
                host=seed_ip,
                port=22,
                username=settings.SEED_SSH_LOGIN,
                password="",
                private_keys=[key["private"]]) as editor:
            content = editor.content or {}
        return content

    def put_remote_file_yaml(self, os_client, content, filepath):
        '''
        Save content as yaml file on remote seed node
        :param os_client:
        :param content:
        :param filepath:
        :return:
        '''
        key = utils.load_keyfile(settings.SEED_SSH_PRIV_KEY_FILE)
        seed_ip = self.get_seed_ip(os_client)
        with templates.YamlEditor(
                file_path=filepath,
                host=seed_ip,
                port=22,
                username=settings.SEED_SSH_LOGIN,
                password="",
                private_keys=[key["private"]]) as editor:
            editor.content = content

    def wait_for_seed_node_ssh(self, os_client):
        # wait for ssh on seed node is up
        seed_ip = self.get_seed_ip(os_client)
        seed_port = 22
        timeout = 1800
        msg = ("Seed node with IP {0} didn't open SSH in {1} sec"
               .format(seed_ip, timeout))
        waiters.wait_tcp(seed_ip, seed_port, timeout, msg)
        # Let cloud-init to finish "modules-final" section
        # to avoid errors like "E: Unable to locate package docker.io"
        sleep(10)

    def create_seed_node(self, os_client):
        stack = os_client._current_stack
        assert not stack
        if settings.KAAS_OFFLINE_DEPLOYMENT:
            if not settings.KAAS_EXTERNAL_PROXY_ACCESS_STR:
                LOG.error("KAAS_EXTERNAL_PROXY_ACCESS_STR is not set, but "
                          "KAAS_OFFLINE_DEPLOYMENT is. "
                          "It makes no sense to deploy mgmt cluster "
                          "in offline mode, but w/o proxy. Exiting..")
                raise RuntimeError('KAAS_OFFLINE_DEPLOYMENT is '
                                   'set, but KAAS_EXTERNAL_PROXY_ACCESS_STR '
                                   'is empty. Please fill out '
                                   'KAAS_EXTERNAL_PROXY_ACCESS_STR')
        else:
            if settings.KAAS_EXTERNAL_PROXY_ACCESS_STR:
                LOG.error("KAAS_EXTERNAL_PROXY_ACCESS_STR is set, but "
                          "KAAS_OFFLINE_DEPLOYMENT is disabled. "
                          "It makes no sense to use proxy on online "
                          "management cluster. Exiting..")
                raise RuntimeError('KAAS_OFFLINE_DEPLOYMENT is disabled, but'
                                   'KAAS_EXTERNAL_PROXY_ACCESS_STR is set. '
                                   'Please enable KAAS_OFFLINE_DEPLOYMENT')
        os_client.stack_create(settings.HEAT_TEMPLATE_PATH)
        stack = os_client._current_stack
        self.wait_for_seed_node_ssh(os_client)
        return stack.outputs

    def reboot_seed_node(self, os_client):
        remote = self.remote_seed(os_client)
        remote.check_call('touch /run/user/$(id -u)/si_rebooted')
        LOG.warning('Rebooting node..')
        with remote.sudo(enforce=True):
            remote.check_call(
                'sync;'
                'nohup bash -c "sleep 2;'
                'iptables -A INPUT -p tcp --dport 22 -j REJECT;'
                'shutdown -r -f now" &>/dev/null & exit',
                verbose=True)
            # just to be sure
            sleep(2)
        self.wait_for_seed_node_ssh(os_client)
        remote.reconnect()
        if remote.isfile('/run/user/$(id -u)/si_rebooted'):
            raise Exception('Node has not been rebooted!')

    def step_001_erase_env_before(self, openstack_client):
        LOG.info("Cleanup KaaS cluster")
        seed_ip = self.get_seed_ip(openstack_client)
        run_envs = {
            "CLUSTER_NAME": settings.CLUSTER_NAME,
            "REGIONAL_CLUSTER_NAME": settings.REGIONAL_CLUSTER_NAME,
            "KAAS_BOOTSTRAP_LOG_LVL": settings.KAAS_BOOTSTRAP_LOG_LVL
        }
        envs_string = utils.make_export_env_strting(run_envs)
        skip_clean = settings.KAAS_SKIP_BOOTSTRAP_CLEANUP

        if seed_ip and not skip_clean:
            LOG.info("IP address {0}".format(seed_ip))
            try:
                remote = self.remote_seed(openstack_client)
                bootstrap_dir = settings.KAAS_BOOTSTRAP_TARGET_DIR
                if remote.isfile("./{}/bootstrap.sh".format(bootstrap_dir)):
                    ret = remote.execute("{env}; bash -x ./{bdir}/bootstrap.sh"
                                         " cleanup".format(
                                             env=envs_string,
                                             bdir=bootstrap_dir),
                                         verbose=True)
                    if ret.exit_code == 0:
                        LOG.info("Cleanup has been finished successfully")
                    else:
                        try:
                            self.step_bs_collect_logs(remote)
                        except Exception as e:
                            LOG.warning("Unable to collect logs: {}".format(e))

                        if settings.KAAS_FORCE_FAIL_ON_BOOTSTRAP_CLEANUP:
                            raise Exception('KaaS cleanup procedure failed')

                        LOG.warning("""
                                Cleanup hasn't been finished successfully
                                Test failure was suppressed
                                set environment variable/Jenkins parameter
                                KAAS_FORCE_FAIL_ON_BOOTSTRAP_CLEANUP: True
                                to force test failure""")
                else:
                    LOG.warning("bootstrap.sh file was not found")
            except OSError:
                LOG.warning("We could not connect to server")
            except paramiko.AuthenticationException:
                LOG.warning("Could not authenticate to server")
        elif skip_clean:
            LOG.warning("Cleanup skipped")
        else:
            LOG.warning("IP address was not found")
        stack_name = settings.ENV_NAME
        LOG.info("Remove heat stack {0}".format(stack_name))
        openstack_client.delete_stack()

    def step_001_cleanup_seed_workspace(self, openstack_client=None):
        remote = self.remote_seed(openstack_client)
        LOG.info("Cleanup docker resources")
        remote.check_call("for i in $(docker ps -aq); do docker rm -f $i;"
                          "done && docker system prune -a -f",
                          verbose=True)
        LOG.info("Cleanup possible bootstrap metadata leftovers")
        # TODO vnaumov refactor cleanup after PRODX-9803
        leftovers = (".ssh/openstack_tmp ssh_key kaas_releases kubeconfig"
                     " logs.tar.gz mirantis.lic output.txt"
                     " passwords.yaml .bootstrap_resource_id .kube {}"
                     .format(settings.KAAS_BOOTSTRAP_TARGET_DIR))
        remote.check_call("rm -rf {}".format(leftovers), verbose=True)

    @utils.log_method_time()
    def step_002_provision_seed_vm(self, openstack_client):
        if openstack_client._current_stack and settings.KEEP_ENV_BEFORE:
            LOG.warning("Stack %s already created", settings.ENV_NAME)
        else:
            self.create_seed_node(openstack_client)
        seed_ip = self.get_seed_ip(openstack_client)
        LOG.info("Seed node ip address is {0}".format(seed_ip))
        assert seed_ip
        with open('{0}/seed_node_ip'.format(settings.ARTIFACTS_DIR),
                  mode='w') as f:
            # TODO: use yaml editor to store the seed info
            f.write(str(seed_ip))
        with open('{0}/env_name'.format(settings.ARTIFACTS_DIR),
                  mode='w') as f:
            # TODO: use yaml editor to store the seed info
            f.write(settings.ENV_NAME)

    def step_003_prepare_seed_node_base(self, openstack_client):
        remote = self.remote_seed(openstack_client)
        _sfile = '/tmp/si_step_003_prepare_seed_node_base_done'
        if remote.isfile(_sfile):
            LOG.warning(f'Skipping stage step_003_prepare_seed_node_base,'
                        f'since {_sfile} exists ')
            return
        apt_for_cloud_str = 'mirror.mirantis.com'
        # Guess nearest mirror for OS cases
        if settings.USE_MIRA_APT_MIRRORS:
            if 'ic-eu.ssl.mirantis.net' in settings.OS_AUTH_URL:
                apt_for_cloud_str = 'mirror-eu.mcp.mirantis.net'
            if 'ic-us.ssl.mirantis.net' in settings.OS_AUTH_URL:
                apt_for_cloud_str = 'mirror-us.mcp.mirantis.net'

        with remote.sudo(enforce=True):
            LOG.info('Stop apt self-update services')
            remote.check_call(
                'systemctl stop apt-daily-upgrade.service '
                'apt-daily.service apt-daily-upgrade.timer '
                'apt-daily.timer unattended-upgrades.service')
            remote.check_call(
                'systemctl disable apt-daily-upgrade.service '
                'apt-daily.service apt-daily-upgrade.timer '
                'apt-daily.timer unattended-upgrades.service')
            LOG.info('Block apt to install suggest pkgs')
            remote.check_call(
                'echo "APT::Get::Install-Suggests "false";" '
                ' > /etc/apt/apt.conf.d/si-tests;'
                'echo "APT::Get::Install-Recommends "false";"'
                '>> /etc/apt/apt.conf.d/si-tests')
            LOG.info('Switch apt to mirantis local mirror')

            remote.check_call(
                ''.join(['export DISTR=$(lsb_release -sc | awk "{print $2}");',
                         'echo "deb [arch=amd64] http://',
                         apt_for_cloud_str,
                         '/nightly/ubuntu/ ${DISTR} main restricted universe"',
                         ' > /etc/apt/sources.list;',
                         'echo "deb [arch=amd64] http://',
                         apt_for_cloud_str,
                         '/nightly/ubuntu/ ${DISTR}-updates main restricted',
                         ' universe" >> /etc/apt/sources.list']))
            remote.check_call('apt update')
            remote.check_call('mkdir -p /etc/docker/')
            remote.check_call(
                "echo '{ \"bip\": \"192.168.91.1/24\" }' "
                " > /etc/docker/daemon.json ")
            LOG.info('Install docker.io to seed node')
            remote.check_call('apt install docker.io')
            remote.check_call(
                "usermod -aG docker {}".format(settings.SEED_SSH_LOGIN))
            LOG.info('Install base packages to seed node')
            remote.check_call('apt install dh-autoreconf '
                              'python3.8-venv rng-tools')

            if settings.SEED_NODE_EXTRA_PKG:
                LOG.warning('Install {}'.format(settings.SEED_NODE_EXTRA_PKG))
                remote.check_call('DEBIAN_FRONTEND=noninteractive apt install '
                                  '{}'.format(settings.SEED_NODE_EXTRA_PKG))
            if settings.SEED_NODE_EXTRA_DO_UPGRADE:
                LOG.warning('Upgrading system')
                remote.check_call(
                    'DEBIAN_FRONTEND=noninteractive apt dist-upgrade')
        if settings.SEED_NODE_EXTRA_PKG_REBOOT_AFTER:
            self.reboot_seed_node(openstack_client)
        remote.reconnect()
        # TODO: those must be refactored, since resource available
        # only internally
        LOG.info('Perform docker connectivity selfcheck for'
                 ' internal resources')
        remote.check_call(
            'docker run --rm '
            'mirantiseng.docker.mirantis.net/library/alpine:3.12.0 '
            'sh -c "wget {} -O -"'.format(settings.OS_AUTH_URL),
            verbose=True,
            error_info=f'Unable to perform access check '
                       f'to IC:{settings.OS_AUTH_URL}')
        remote.check_call(f'touch {_sfile}')

    def prepare_seed_node_region_templates(
            self,
            openstack_client,
            remote_templates_dir):
        """Prepare templates on seed node."""
        remote = self.remote_seed(openstack_client)
        if not settings.BOOTSTRAP_WITH_DEFAULTS:
            bootstrap_templates_path = self.get_bootstrap_templates_path()
        else:
            bootstrap_templates_path = self.get_bootstrap_templates_path(
                remote, openstack_client)

        bootstrap_templates = []
        for dirpath, _, filenames in os.walk(bootstrap_templates_path):
            dirpath = dirpath.replace(bootstrap_templates_path, '')
            # in case when dirpath has forward `/` os.path.join drop first argument  # noqa
            dirpath = dirpath.lstrip('/')
            remote.check_call("mkdir -p {0}".format(
                os.path.join(remote_templates_dir, dirpath)))
            for file_name in filenames:
                if file_name == '.defaults':
                    continue
                bootstrap_templates.append(os.path.join(dirpath, file_name))

        for template_name in bootstrap_templates:
            template_yaml = templates.render_template(
                os.path.join(bootstrap_templates_path, template_name))
            template_filename = os.path.join(remote_templates_dir,
                                             template_name)
            LOG.debug("Writing to the seed node the template {0}"
                      .format(template_filename))
            with remote.open(template_filename, mode="w") as r_f:
                r_f.write(template_yaml)

    def step_003_prepare_seed_node_templates(self, openstack_client):
        remote_templates_dir = f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/' \
                               f'templates'
        self.prepare_seed_node_region_templates(
            openstack_client, remote_templates_dir)

    def step_003_prepare_seed_node_get_bootstrap(self, openstack_client):
        LOG.info("Wait for connection check to seed ip")
        self.wait_for_seed_node_ssh(openstack_client)
        remote = self.remote_seed(openstack_client)
        remote.upload(settings.KAAS_RELEASES_FOLDER,
                      settings.KAAS_RELEASES_REMOTE_FOLDER)
        run_envs = {"KAAS_CDN_REGION": settings.CDN_REGION}
        if settings.KAAS_EXTERNAL_PROXY_ACCESS_STR:
            run_envs['HTTP_PROXY'] = \
                settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            run_envs['HTTPS_PROXY'] = \
                settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            LOG.info("HTTP_PROXY and HTTPS_PROXY is set to {0}".format(
                settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            ))
            # PRODX-13381: add EU/US fip range to no_proxy
            if settings.KAAS_PROXYOBJECT_NO_PROXY:
                run_envs['NO_PROXY'] = \
                    f"{settings.KAAS_PROXYOBJECT_NO_PROXY}," \
                    f"{settings.OS_FIP_RANGES}"
            else:
                run_envs['NO_PROXY'] = settings.OS_FIP_RANGES
            run_envs['no_proxy'] = run_envs['NO_PROXY']
            LOG.info("NO_PROXY is set to {0}".format(
                run_envs['NO_PROXY']))
        if not settings.BOOTSTRAP_WITH_DEFAULTS:
            kaas_release_filename = self.get_kaas_release_file_name()
            run_envs['KAAS_RELEASE_YAML'] = \
                '{}/kaas/{}'.format(settings.KAAS_RELEASES_REMOTE_FOLDER,
                                    kaas_release_filename)
            run_envs['CLUSTER_RELEASES_DIR'] = \
                '{}/cluster'.format(settings.KAAS_RELEASES_REMOTE_FOLDER)
            run_envs['TARGET_DIR'] = settings.KAAS_BOOTSTRAP_TARGET_DIR
            envs_string = utils.make_export_env_strting(run_envs)
            remote.check_call(
                "{envs_string}; bash -x {kaas}/{script}".format(
                    envs_string=envs_string,
                    kaas=settings.KAAS_RELEASES_REMOTE_FOLDER,
                    script=settings.KAAS_GET_SCRIPT_NAME),
                verbose=True)
        else:

            run_envs["KAAS_RELEASES_BASE_URL"] = \
                "{}releases".format(settings.KAAS_RELEASES_BASE_URL)
            envs_string = utils.make_export_env_strting(run_envs)
            remote.check_call(
                "{envs_string}; bash -x {kaas}/{script}".format(
                    envs_string=envs_string,
                    kaas=settings.KAAS_RELEASES_REMOTE_FOLDER,
                    script=settings.KAAS_GET_SCRIPT_NAME),
                verbose=True)

        remote.check_call(
            "cd {}; "
            "mv templates templates.orig && mkdir templates".format(
                settings.KAAS_BOOTSTRAP_TARGET_DIR),
            verbose=True)
        if not settings.BOOTSTRAP_WITH_DEFAULTS:
            bootstrap_version = self.get_bootstrap_version()
            LOG.info("Save artifact with KaaS bootstrap version")
            with open("{0}/bootstrap_version".format(settings.ARTIFACTS_DIR),
                      "w") as f:
                f.write(bootstrap_version)

            kaas_version = self.get_kaas_version()
            LOG.info("Save artifact with KaaS bootstrap version")
            with open("{0}/management_version".format(settings.ARTIFACTS_DIR),
                      "w") as f:
                f.write(kaas_version)
        else:
            release_content = self.get_remote_default_kaas_data(
                remote, openstack_client)
            bootstrap_version = release_content['spec']['bootstrap']['version']
            LOG.info("Bootstrap version fetched from boot.env - {}".format(
                bootstrap_version))
            LOG.info("Save artifact with KaaS bootstrap version")
            with open("{0}/bootstrap_version".format(settings.ARTIFACTS_DIR),
                      "w") as f:
                f.write(bootstrap_version)
            kaas_version = release_content['spec']['version']
            LOG.info("KaaS version fetched from boot.env - {}".format(
                kaas_version))

            LOG.info("Save artifact with KaaS bootstrap version")
            with open("{0}/management_version".format(settings.ARTIFACTS_DIR),
                      "w") as f:
                f.write(kaas_version)

    @utils.log_method_time()
    def step_003_prepare_seed_node(self, openstack_client):
        self.wait_for_seed_node_ssh(openstack_client)
        remote = self.remote_seed(openstack_client)

        LOG.info("Verify that the seed node isn't prepared")
        if remote.isdir(settings.KAAS_BOOTSTRAP_TARGET_DIR) and \
                remote.isfile('{}/clouds.yaml'.format(
                    settings.KAAS_BOOTSTRAP_TARGET_DIR)):
            LOG.error('Seed node already prepared')
            return

        self.step_003_prepare_seed_node_base(openstack_client)
        self.step_003_prepare_seed_node_get_bootstrap(openstack_client)
        self.step_003_prepare_seed_node_templates(openstack_client)

        LOG.info("Render clouds.yaml and kaas templates to seed node")
        clouds_yaml = templates.render_template(settings.CLOUDS_YAML_PATH)
        with remote.open("{}/clouds.yaml".format(
                settings.KAAS_BOOTSTRAP_TARGET_DIR), mode='w') as r_f:
            r_f.write(clouds_yaml)

        remote.upload(
            settings.KAAS_BOOTSTRAP_LICENSE_FILE,
            settings.KAAS_BOOTSTRAP_LICENSE_REMOTE_FILE)
        LOG.info("License is uploaded")

    # works only with SEED_STANDALONE_EXTERNAL_IP
    @utils.log_method_time()
    def step_003_prepare_standalone_seed_node(self):
        seed_ip = self.get_seed_ip(os_client=None, fip='seed_floating_ip')
        assert seed_ip
        LOG.info("Seed node ip address is {0}".format(seed_ip))
        with open('{0}/seed_node_ip'.format(settings.ARTIFACTS_DIR),
                  mode='w') as f:
            # TODO: use yaml editor to store the seed info
            f.write(str(seed_ip))

        remote = self.remote_seed(None)
        self.step_003_prepare_seed_node_get_bootstrap(None)
        self.step_003_prepare_seed_node_templates(None)

        bootstrap_version = self.get_bootstrap_version()
        with open("{0}/bootstrap_version".format(settings.ARTIFACTS_DIR),
                  "w") as f:
            f.write(bootstrap_version)

        remote.upload(
            settings.KAAS_BOOTSTRAP_LICENSE_FILE,
            settings.KAAS_BOOTSTRAP_LICENSE_REMOTE_FILE)
        LOG.info("License is uploaded")

    @utils.log_method_time()
    def step_003_prepare_bm_seed_node(self, openstack_client):
        self.wait_for_seed_node_ssh(openstack_client)
        remote = self.remote_seed(openstack_client)
        kaas_rel_dir = settings.KAAS_RELEASES_FOLDER

        self.step_003_prepare_seed_node_base(openstack_client)
        # Run ansible to perform install
        with remote.sudo(enforce=True):
            # yes, we still need python3-venv even if we have 3.8-venv
            pkgs = 'curl python3.8-venv python3.8-dev ' \
                   'python3-venv python-apt ' \
                   'python-pip libvirt-dev gcc'
            LOG.info(f'Install all necessary packages: {pkgs}')
            remote.check_call(f'apt -y install {pkgs}')
        self.ansible_init(openstack_client)

        # after release 1.12 thats required only for bm single tests
        kaas_core_dir = settings.KAAS_CORE_REPO_PATH
        kaas_core_remote_dir = settings.KAAS_CORE_REMOTE_PATH
        if os.path.isdir(kaas_core_dir):
            LOG.info("Uploading '{}' dir to 'node0'".format(kaas_core_dir))
            self.rsync_run(openstack_client, from_dir=kaas_core_dir,
                           todir=f'~/{kaas_core_remote_dir}')
        else:
            LOG.warning('kaas-core dir has not been found.')
        if os.path.isdir(kaas_rel_dir):
            LOG.info("Uploading kaas-releases dir to 'node0'")
            self.rsync_run(openstack_client, from_dir=kaas_rel_dir,
                           todir=f'~/{settings.KAAS_RELEASES_REMOTE_FOLDER}')
        else:
            LOG.warning('kaas-releases dir has not been found.')

        # Upload license file
        remote.upload(
            settings.KAAS_BOOTSTRAP_LICENSE_FILE,
            settings.KAAS_BOOTSTRAP_LICENSE_REMOTE_FILE)
        LOG.info("License is uploaded")

        # Prepare and upload ansible ENV file
        env_dict = dict(CLOUD_NAME=settings.CLOUD_NAME,
                        KAAS_CDN_REGION=settings.CDN_REGION,
                        CDN_BOOTSTRAP_REGION=settings.CDN_REGION)
        if settings.DNS_NAMESERVER:
            env_dict['DNS_NAMESERVER'] = settings.DNS_NAMESERVER
        if settings.KAAS_BM_DEPLOY_KAAS_BASED_ON_SOURCE:
            env_dict[
                'KAAS_BM_DEPLOY_KAAS_BASED_ON_SOURCE'] = settings.KAAS_BM_DEPLOY_KAAS_BASED_ON_SOURCE  # noqa
        # Pass global variables to ansible.
        if settings.KAAS_BM_PUBLIC_API_CHART_URL:
            env_dict[
                'KAAS_BM_PUBLIC_API_CHART_URL'] = settings.KAAS_BM_PUBLIC_API_CHART_URL  # noqa
        if settings.KAAS_BM_OPERATOR_CHART_URL:
            env_dict[
                'KAAS_BM_OPERATOR_CHART_URL'] = settings.KAAS_BM_OPERATOR_CHART_URL  # noqa
        if settings.KAAS_BM_OPERATOR_CHART_VERSION:
            env_dict['KAAS_BM_CHARTS_REPO'] = settings.KAAS_BM_CHARTS_REPO
        if settings.KAAS_BM_OPERATOR_DOCKER_IMAGE:
            env_dict[
                'KAAS_BM_OPERATOR_DOCKER_IMAGE'] = settings.KAAS_BM_OPERATOR_DOCKER_IMAGE  # noqa
        if settings.KAAS_BM_IPAM_CHART_URL:
            env_dict[
                'KAAS_BM_IPAM_CHART_URL'] = settings.KAAS_BM_IPAM_CHART_URL
        if settings.KAAS_BM_IPAM_DOCKER_IMAGE:
            env_dict[
                'KAAS_BM_IPAM_DOCKER_IMAGE'] = settings.KAAS_BM_IPAM_DOCKER_IMAGE  # noqa
        if settings.KAAS_BM_KAAS_IPAM_DOCKER_IMAGE:
            env_dict[
                'KAAS_BM_KAAS_IPAM_DOCKER_IMAGE'] = settings.KAAS_BM_KAAS_IPAM_DOCKER_IMAGE  # noqa
        if settings.KAAS_BM_IRONIC_OPERATOR_DOCKER_IMAGE:
            env_dict[
                'KAAS_BM_IRONIC_OPERATOR_DOCKER_IMAGE'] = settings.KAAS_BM_IRONIC_OPERATOR_DOCKER_IMAGE  # noqa
        if settings.CDN_REGION:
            env_dict['KAAS_CDN_REGION'] = settings.CDN_REGION
        if settings.KAAS_BM_KAAS_ANSIBLE_TAR_GZ:
            env_dict[
                'KAAS_BM_KAAS_ANSIBLE_TAR_GZ'] = settings.KAAS_BM_KAAS_ANSIBLE_TAR_GZ  # noqa
        if settings.KAAS_BM_KAAS_ANSIBLE_TAR_GZ_MD5:
            env_dict[
                'KAAS_BM_KAAS_ANSIBLE_TAR_GZ_MD5'] = settings.KAAS_BM_KAAS_ANSIBLE_TAR_GZ_MD5  # noqa
        if settings.KAAS_BM_DEPLOY_KAAS_BASED_ON_SOURCE:
            kaas_release_filename = self.get_kaas_release_file_name()
            kaas_release_yml_remote_path = os.path.join(
                settings.KAAS_RELEASES_REMOTE_FOLDER,
                'kaas',
                kaas_release_filename)
            # Processed to get_container_cloud.sh
            env_dict['KAAS_RELEASE_YAML'] = kaas_release_yml_remote_path
            env_dict['CLUSTER_RELEASES_DIR'] = '{}/cluster'.format(
                settings.KAAS_RELEASES_REMOTE_FOLDER)
        if settings.KAAS_PLATFORM_RELEASE_REF:
            env_dict[
                'KAAS_PLATFORM_RELEASE_REF'] = settings.KAAS_PLATFORM_RELEASE_REF  # noqa
        if settings.KAAS_RELEASE_REPO_URL:
            env_dict['KAAS_RELEASE_REPO_URL'] = settings.KAAS_RELEASE_REPO_URL
        if settings.KAAS_BOOTSTRAP_LICENSE_REMOTE_FILE:
            env_dict['KAAS_BOOTSTRAP_LICENSE_REMOTE_FILE'] = \
                settings.KAAS_BOOTSTRAP_LICENSE_REMOTE_FILE
        if settings.KEYCLOAK_LDAP_ENABLED:
            env_dict['KEYCLOAK_LDAP_ENABLED'] = settings.KEYCLOAK_LDAP_ENABLED
            env_dict['KEYCLOAK_LDAP_BIND_USERNAME'] = \
                settings.KEYCLOAK_LDAP_BIND_USERNAME
            env_dict['KEYCLOAK_LDAP_BIND_PASSWORD'] = \
                settings.KEYCLOAK_LDAP_BIND_PASSWORD
        if settings.USE_MIRA_APT_MIRRORS:
            env_dict['USE_MIRA_APT_MIRRORS'] = \
                settings.USE_MIRA_APT_MIRRORS

        seed_floating_ip = self.get_seed_ip(openstack_client,
                                            fip='seed_floating_ip')
        if seed_floating_ip:
            env_dict['BM_SEED_FLOATING_IP'] = seed_floating_ip

        mgmt_floating_ip = self.get_seed_ip(openstack_client,
                                            fip='vip_floating_ip')
        if mgmt_floating_ip:
            env_dict['BM_MGMT_FLOATING_IP'] = mgmt_floating_ip

        child_floating_ip = self.get_seed_ip(openstack_client,
                                             fip='child_floating_ip')
        if child_floating_ip:
            env_dict['BM_CHILD_FLOATING_IP'] = child_floating_ip

        iam_floating_ip = self.get_seed_ip(openstack_client,
                                           fip='iam_floating_ip')
        if iam_floating_ip:
            LOG.info("IAM floating ip address is {0}".format(iam_floating_ip))
            with open('{0}/iam_floating_ip'.format(settings.ARTIFACTS_DIR),
                      mode='w') as f:
                f.write(str(iam_floating_ip))

        if settings.KAAS_RELEASES_BASE_URL:
            env_dict['KAAS_RELEASES_BASE_URL'] = settings.KAAS_RELEASES_BASE_URL   # noqa
        if settings.KAAS_BM_CI_SCALEDOWNRESOURCES:
            env_dict['KAAS_BM_CI_SCALEDOWNRESOURCES'] = settings.KAAS_BM_CI_SCALEDOWNRESOURCES # noqa
        env_dict['KAAS_BM_BOOT_UEFI'] = settings.KAAS_BM_BOOT_UEFI

        # As for release 2.3+ we need to pass correct region, even for
        # mgmt cluster deployment, especially for custom bmhp for mgmt case
        env_dict['BM_REGION'] = settings.BM_REGION
        if settings.BM_DEPLOY_REGIONAL:
            env_dict['REGIONAL_CLUSTER_NAME'] = settings.REGIONAL_CLUSTER_NAME

        if settings.IAM_FLOATING_IP:
            env_dict['IAM_FLOATING_IP'] = settings.IAM_FLOATING_IP

        if settings.KAAS_BM_FULL_PREFLIGHT:
            LOG.info('Full preflight is enabled')
            env_dict['KAAS_BM_FULL_PREFLIGHT'] = settings.KAAS_BM_FULL_PREFLIGHT # noqa

        if settings.KAAS_EXTERNAL_PROXY_ACCESS_STR:
            env_dict['HTTP_PROXY'] = \
                settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            env_dict['HTTPS_PROXY'] = \
                settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            env_dict['NO_PROXY'] = \
                settings.KAAS_NO_PROXY_ACCESS_STR

        if settings.KAAS_OFFLINE_DEPLOYMENT:
            # Use the seed node as the default GW for cluster nodes
            # those pass ignored for virtual case, see setup-kaas-bm-pbook.yml
            # FIXME(alexz): those event for virt case must be refactored
            #  For virt case, manually set SET_PXE_NW_GW=10.0.0.15
            env_dict['SET_PXE_NW_GW'] = settings.PHYS_GATEWAY

        self.put_remote_file_yaml(openstack_client, env_dict,
                                  settings.ANSIBLE_ENV_OVERRIDE_FILE)

        kaas_version = self.get_kaas_version()
        LOG.info("Save artifact with KaaS bootstrap version")
        with open("{0}/management_version".format(settings.ARTIFACTS_DIR),
                  "w") as f:
            f.write(kaas_version)

    def check_templates(self, openstack_client):
        """Match local templates to default templates from KaaS

        Match the local templates rendered with default variables
        from the file {KAAS_BOOTSTRAP_TEMPLATES_DIR}/.defaults
        to the default templates from the KaaS bootstrap archive.

        This allows to catch any changes made to default templates
        and fix it in the local templates in si-tests.

        Local templates by default should be placed in the directories
        according to the KaaS bootstrap archive version:
        templates/bootstrap/templates_{KAAS_VERSION}/

        Default KaaS templates are expected on the seed node in the
        directory '~/templates.orig/'.

        Scenario:

            1. Compare list of the local and default template files
            2. Render a local template to a string object using Jinja2
            3. Read the appropriate default template from the seed node
            4. Convert both templates to python objects using yaml.load
               to check the yaml syntax.
            5. Convert both templates back to string using yaml.dump
               to get the same result without comments, with the same
               indents and the same keys order
            6. Compare the resulting strings using assertEqual, to get
               a nice diff in the log.
        """
        if not settings.BOOTSTRAP_WITH_DEFAULTS:
            bootstrap_templates_path = self.get_bootstrap_templates_path()
        else:
            remote = self.remote_seed(openstack_client)
            bootstrap_templates_path = self.get_bootstrap_templates_path(
                remote, openstack_client)
        # Read .defaults file from the templates dir
        with open('{0}/.defaults'.format(bootstrap_templates_path)) as f:
            default_variables = yaml.load(f.read(), Loader=yaml.SafeLoader)

        def os_env(var_name, default=None, env_type=None):
            # Define variables that will produce the default template
            var = default_variables.get(var_name, default)

            if var is None:
                raise Exception("Environment variable '{0}' is undefined!"
                                .format(var_name))

            if env_type == "bool":
                var = settings.get_var_as_bool(var_name, default)

            return var

        def os_env_bool(var_name, default=None):
            return os_env(var_name, default, env_type='bool')

        remote = self.remote_seed(openstack_client)
        res = remote.execute(
            "cd {}/templates.orig; find . -type f".format(
                settings.KAAS_BOOTSTRAP_TARGET_DIR))
        orig_templates = [x.decode("utf-8").strip().replace('./', '')
                          for x in res.stdout]

        bootstrap_templates = []
        for dirpath, _, filenames in os.walk(bootstrap_templates_path):
            dirpath = dirpath.replace(
                bootstrap_templates_path, '')
            # in case when dirpath has forward `/` os.path.join drop first argument  # noqa
            dirpath = dirpath.lstrip('/')
            for file_name in filenames:
                if file_name == '.defaults':
                    continue
                bootstrap_templates.append(os.path.join(dirpath, file_name))

        self.assertEqual(set(orig_templates), set(bootstrap_templates))

        self.maxDiff = None
        render_options = {'os_env': os_env, 'os_env_bool': os_env_bool,
                          'TEST': True}

        for template_name in bootstrap_templates:
            # render with overriden os_env
            template_yaml = yaml.load_all(templates.render_template(
                os.path.join(bootstrap_templates_path, template_name),
                options=render_options), Loader=yaml.SafeLoader)

            template_name_file = os.path.join(
                '{}/templates.orig'.format(
                    settings.KAAS_BOOTSTRAP_TARGET_DIR),
                template_name)
            with remote.open(template_name_file, mode="r") as r_f:
                orig_yaml = yaml.load_all(r_f.read(), Loader=yaml.SafeLoader)
                self.assertEqual(yaml.dump_all(template_yaml),
                                 yaml.dump_all(orig_yaml))

    @utils.log_method_time()
    def step_004_prepare_for_kaas_bm_cluster(self, openstack_client):
        remote = self.remote_seed(openstack_client)
        _sfile = '/tmp/step_004_prepare_for_kaas_bm_cluster'
        if remote.isfile(_sfile):
            LOG.warning(f'Skipping stage step_004_prepare_for_kaas_bm_cluster,'
                        f'since {_sfile} exists ')
            return
        if settings.BM_HALF_VIRTUAL_ENV:
            inventory_prefix = "inventory_half_virt_deploy"
            ansible_vars = {'half_virt_lab': 'yes'}
        else:
            inventory_prefix = "inventory_virt_deploy"
            ansible_vars = {'virt_lab': 'yes'}
        ansible_inventory = [f'./{inventory_prefix}.yaml']

        if settings.BM_DEPLOY_CHILD:
            # Deploy 6 more VMs for child cluster
            ansible_inventory.append(f'./{inventory_prefix}_child.yaml')
            ansible_vars['deploy_bm_child'] = 'yes'

        if settings.BM_DEPLOY_REGIONAL:
            ansible_vars['deploy_bm_regional'] = 'yes'

        LOG.info('Base setup, render and create VMs using virsh')
        self.ansible_run(remote, "./setup-playbook.yml",
                         inventory=ansible_inventory, extra_vars=ansible_vars)
        remote.reconnect()
        remote.check_call(f'touch {_sfile}')

    # It also can deploy regional cluster
    @utils.log_method_time()
    def step_004_deploy_kaas_bm_mgmt_cluster(self, openstack_client):
        remote = self.remote_seed(openstack_client)
        bs_dir = settings.KAAS_BOOTSTRAP_TARGET_DIR
        prefix = 'management'
        if settings.BM_HALF_VIRTUAL_ENV:
            inventory_prefix = "inventory_half_virt_deploy"
            ansible_vars = {'half_virt_lab': 'yes'}
        else:
            inventory_prefix = "inventory_virt_deploy"
            ansible_vars = {"virt_lab": "yes"}
        ansible_inventory = [f"./{inventory_prefix}.yaml"]

        if settings.BM_DEPLOY_CHILD:
            # Deploy 6 more VMs for child cluster
            ansible_inventory.append(f"./{inventory_prefix}_child.yaml")
            ansible_vars["deploy_bm_child"] = "yes"

        if settings.BM_DEPLOY_REGIONAL:
            ansible_vars["deploy_bm_regional"] = "yes"
            prefix = 'regional'

        playbook = "./setup-kaas-bm-pbook.yml"
        try:
            self.ansible_run(
                remote, playbook,
                inventory=ansible_inventory,
                extra_vars=ansible_vars,
                timeout=settings.KAAS_BOOTSTRAP_TIMEOUT)
            remote.reconnect()
        except Exception as e:
            LOG.error(f"Ansible playbook '{playbook}' failed!\n{e}")
            raise e

        LOG.info(f"Bootstrap KaaS-BM {prefix} cluster")
        try:
            if not settings.BM_DEPLOY_REGIONAL:
                remote.check_call(
                    f"cd {bs_dir}; "
                    f"bash -o pipefail -xc './bootstrap.sh all |"
                    f"tee -a ./logs/fastlog.txt'",
                    verbose=True,
                    timeout=settings.KAAS_BOOTSTRAP_TIMEOUT)
        except Exception as e:
            LOG.error("Bootstrap failed: {}".format(e))
            raise e
        finally:
            try:
                self.step_bs_collect_logs(remote)
                self.check_multiple_kaasreleases(remote)
                # Download management_kubeconfig artifact
                if remote.isfile('./{}/kubeconfig'.format(bs_dir)):
                    remote.download("./{}/kubeconfig".format(bs_dir),
                                    "{0}/{1}_kubeconfig_orig"
                                    .format(settings.ARTIFACTS_DIR, prefix))
                    LOG.info(f"{prefix} kubeconfig has been downloaded")

                    with open("{0}/{1}_kubeconfig_orig".format(
                            settings.ARTIFACTS_DIR, prefix), 'r') as f:
                        kubeconfig_old = yaml.load(f.read(),
                                                   Loader=yaml.SafeLoader)
                    kubeconfig = self.expose_kubeconfig(
                        kubeconfig_old, self.get_seed_ip(
                            openstack_client,
                            fip='vip_floating_ip'))
                    with open("{0}/{1}_kubeconfig".format(
                            settings.ARTIFACTS_DIR, prefix), "w") as f:
                        f.write(yaml.dump(kubeconfig))

                # Download test-0-bmh.yaml generated just for test
                # in case of DEPLOY_CHILD=false
                if remote.isfile('bootstrap/test/test-0-bmh.yaml'):
                    remote.download("bootstrap/test/test-0-bmh.yaml",
                                    "{0}/test-0-bmh.yaml"
                                    .format(settings.ARTIFACTS_DIR))
                else:
                    LOG.warning("Test bmh file was not found")
            except Exception as e:
                LOG.error(
                    "Unable to collect bootstrap logs: {}".format(e))
                raise e

        LOG.info(f"{prefix} cluster has been deployed successfully")

    # It also can deploy regional cluster
    @utils.log_method_time()
    def step_004_deploy_kaas_bm_cluster_physical(self, openstack_client,
                                                 lab_id):

        bs_dir = settings.KAAS_BOOTSTRAP_TARGET_DIR
        remote = self.remote_seed(openstack_client)
        prefix = 'management'
        ansible_vars = {}

        if settings.BM_DEPLOY_REGIONAL:
            ansible_vars["deploy_bm_regional"] = "yes"
            prefix = 'regional'

        LOG.info('Ansible base')
        # We need to get ipmitools and networking here
        playbook = f"setup-kaas-bm-{lab_id}.yml"
        try:
            self.ansible_run(
                remote, playbook,
                inventory="./inventory.yaml",
                extra_vars=ansible_vars,
                timeout=settings.KAAS_BOOTSTRAP_TIMEOUT)
            remote.reconnect()
        except Exception as e:
            LOG.error(f"Ansible playbook '{playbook}' failed!\n{e}")
            raise e
        remote.reconnect()

        if remote.isfile(f'{bs_dir}/kaas') and settings.KEEP_ENV_BEFORE:
            LOG.warning('Power off HW skipped')
        else:
            ret = remote.check_call(
                f" source kaas-bm-env/labs/{lab_id}/patch.sh ; "
                f" set_power off all",
                verbose=True, raise_on_err=False)
            if ret.exit_code != 0:
                LOG.warning('Power off HW failed!')

        LOG.info(f"Bootstrap KaaS-BM {prefix} cluster")
        try:
            if not settings.BM_DEPLOY_REGIONAL:
                remote.check_call(
                    f"cd {bs_dir}; "
                    f"bash -o pipefail -xc './bootstrap.sh all |"
                    f"tee -a ./logs/fastlog.txt'",
                    verbose=True,
                    timeout=settings.KAAS_BOOTSTRAP_TIMEOUT)
        except Exception as e:
            LOG.error("Bootstrap failed: {}".format(e))
            raise e
        finally:
            try:
                self.step_bs_collect_logs(remote)
                self.check_multiple_kaasreleases(remote)
                if remote.isfile(f'{bs_dir}/kubeconfig'):
                    remote.download(f"{bs_dir}/kubeconfig",
                                    f"{settings.ARTIFACTS_DIR}"
                                    f"/{prefix}_kubeconfig")
                    LOG.info(f"{prefix} kubeconfig has been downloaded")
            except Exception as e:
                LOG.error(
                    "Unable to collect bootstrap logs: {}".format(e))
                raise e

        LOG.info(f"{prefix} cluster has been deployed successfully")

    @utils.log_method_time()
    def step_004_deploy_kaas_cluster(self, openstack_client):
        """Run bootstrap.sh, download logs"""
        LOG.info("Verify that seed node have been prepared")
        remote = self.remote_seed(openstack_client)

        LOG.info("Deploy KaaS cluster")
        run_envs = {
            "CLUSTER_NAME": settings.CLUSTER_NAME,
            "KAAS_BOOTSTRAP_LOG_LVL": settings.KAAS_BOOTSTRAP_LOG_LVL
        }
        kaas_release_filename = self.get_kaas_release_file_name()

        if settings.KAAS_EXTERNAL_PROXY_ACCESS_STR:
            run_envs['HTTP_PROXY'] = settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            run_envs['HTTPS_PROXY'] = settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            LOG.info("HTTP_PROXY and HTTPS_PROXY is set to {0}".format(
                settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            ))
            # PRODX-13381: add EU/US fip range to no_proxy
            if settings.KAAS_PROXYOBJECT_NO_PROXY:
                run_envs['NO_PROXY'] = \
                    f"{settings.KAAS_PROXYOBJECT_NO_PROXY},"\
                    f"{settings.OS_FIP_RANGES}"
            else:
                run_envs['NO_PROXY'] = settings.OS_FIP_RANGES
            run_envs['no_proxy'] = run_envs['NO_PROXY']
            LOG.info("NO_PROXY is set to {0}".format(
                run_envs['NO_PROXY']))

        if settings.KAAS_RELEASES_FOLDER and \
                os.path.isdir(settings.KAAS_RELEASES_FOLDER) and \
                kaas_release_filename and not \
                settings.BOOTSTRAP_WITH_DEFAULTS:
            LOG.info('Deploy with custom releases')
            run_envs['KAAS_RELEASE_YAML'] = '{}/kaas/{}'.format(
                settings.KAAS_RELEASES_REMOTE_FOLDER,
                kaas_release_filename.strip())
            run_envs['CLUSTER_RELEASES_DIR'] = '{}/cluster'.format(
                settings.KAAS_RELEASES_REMOTE_FOLDER)
        if settings.KAAS_AWS_ENABLED:
            # For AWS provider
            LOG.info('Enable AWS provider')
            run_envs['KAAS_AWS_ENABLED'] = True
            run_envs['AWS_SECRET_ACCESS_KEY'] = settings.AWS_SECRET_ACCESS_KEY
            run_envs['AWS_ACCESS_KEY_ID'] = settings.AWS_ACCESS_KEY_ID
            run_envs['AWS_DEFAULT_REGION'] = settings.AWS_DEFAULT_REGION
        elif settings.KAAS_VSPHERE_ENABLED:
            # For Vsphere provider
            LOG.info('Enable Vsphere provider')
            run_envs['KAAS_VSPHERE_ENABLED'] = True
            run_envs['KEYCLOAK_FLOATING_IP'] = settings.VSPHERE_KEYCLOAK_FLOATING_IP  # noqa
            run_envs['IAM_FLOATING_IP'] = settings.VSPHERE_IAM_FLOATING_IP
        elif settings.KAAS_EQUINIX_ENABLED:
            # For Equinix provider
            LOG.info('Enable Equinix provider')
            run_envs['KAAS_EQUINIX_ENABLED'] = True
        else:
            assert remote.isfile("{}/clouds.yaml".format(
                settings.KAAS_BOOTSTRAP_TARGET_DIR))
        assert remote.isfile("{}/kaas".format(
            settings.KAAS_BOOTSTRAP_TARGET_DIR))
        if not settings.BOOTSTRAP_WITH_DEFAULTS:
            bootstrap_version = self.get_bootstrap_version()
        else:
            bootstrap_version = self.get_remote_default_kaas_data(
                remote, openstack_client)['spec']['bootstrap']['version']
        remote.check_call('{}/kaas version 2>&1 | grep {}'.format(
            settings.KAAS_BOOTSTRAP_TARGET_DIR,
            bootstrap_version), verbose=True)
        LOG.info('Deploy with vanilla releases')
        envs_string = utils.make_export_env_strting(run_envs)
        LOG.info('Current envs_string {}'.format(envs_string))
        if settings.KAAS_AWS_ENABLED:
            # Check cloudformation created or not existed
            # to avoid conflict when stack is in CREATE_IN_PROGRESS state
            cloudformation_stack_name = settings.AWS_CLOUDFORMATION_STACK_NAME
            aws_client = AwsManager(
                aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
                aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
                region_name=settings.AWS_DEFAULT_REGION)

            stack = aws_client.get_cloudformation_stack(
                cloudformation_stack_name)
            if not stack:
                LOG.info("Cloudformation stack is not existed. "
                         "Will proceed with creation")
            else:
                stack_status = stack.get('StackStatus')
                if stack_status == 'CREATE_COMPLETE':
                    LOG.info("Stack is created. Will proceed with updating")
                elif stack_status == 'CREATE_IN_PROGRESS':
                    LOG.info("Cloudformation stack is CREATE_IN_PROGRESS "
                             "status. Waiting for CREATE_COMPLETE")
                    waiters.wait(lambda: aws_client.get_cloudformation_stack(
                        cloudformation_stack_name).get(
                            'StackStatus') == 'CREATE_COMPLETE')
                elif stack_status == 'DELETE_IN_PROGRESS':
                    LOG.info("Cloudformation stack is DELETE_IN_PROGRESS "
                             "status. Waiting for deletion")
                    waiters.wait(
                        lambda: not aws_client.get_cloudformation_stack(
                            cloudformation_stack_name))
                    LOG.info("Stack is deleted")
                else:
                    LOG.warning("Stack {} is existed and has incorrect "
                                "status. Current status: {}. Will try "
                                "to delete".format(cloudformation_stack_name,
                                                   stack_status))
                    aws_client.aws_cloudformation_client.delete_stack(
                        StackName=cloudformation_stack_name)
                    waiters.wait(
                        lambda: not aws_client.get_cloudformation_stack(
                            cloudformation_stack_name))
                    LOG.info("Stack is deleted")

            ret = remote.execute(
                "{envs_string}; "
                "./{bootstrap}/kaas bootstrap aws policy".format(
                    envs_string=envs_string,
                    bootstrap=settings.KAAS_BOOTSTRAP_TARGET_DIR),
                verbose=True)
            bootstrapper_users = [user.get('UserName', '') for user
                                  in aws_client.get_users()
                                  if 'bootstrapper'
                                  in user.get('UserName', '')]
            assert len(bootstrapper_users) > 0, ("Bootstrapper user was "
                                                 "not created")
            try:
                LOG.info("Trying to create new credentials for "
                         "user {}".format(bootstrapper_users[0]))
                response = aws_client.aws_iam_client.create_access_key(
                    UserName=bootstrapper_users[0]
                )
                AccessKeyId = response['AccessKey'].get('AccessKeyId')
                SecretAccessKey = response['AccessKey'].get('SecretAccessKey')
                run_envs['AWS_SECRET_ACCESS_KEY'] = SecretAccessKey
                run_envs['AWS_ACCESS_KEY_ID'] = AccessKeyId
                LOG.info("Credentials created successfuly: {}".format(
                    AccessKeyId))
                envs_string = utils.make_export_env_strting(run_envs)
                LOG.info('Updated envs_string with new AWS creds: {}'
                         .format(envs_string))
            except aws_client.aws_iam_client.exceptions.LimitExceededException\
                    as e:
                LOG.warning("Limit Exceeded: {}".format(e))
                LOG.info(
                    "Admin credentials will be used")
        try:
            ret = remote.execute(
                "{envs_string}; bash -x ./{bootstrap}/bootstrap.sh all".format(
                    envs_string=envs_string,
                    bootstrap=settings.KAAS_BOOTSTRAP_TARGET_DIR),
                verbose=True,
                timeout=settings.KAAS_BOOTSTRAP_TIMEOUT)
        except ExecHelperTimeoutError as e:
            LOG.error("Timeout for bootstrap exceeded: {}".format(e))
            raise e
        finally:
            try:
                self.step_bs_collect_logs(remote)
                self.check_multiple_kaasreleases(remote)
                # Download management_kubeconfig artifact
                kubeconfig_path = "{0}/management_kubeconfig" \
                    .format(settings.ARTIFACTS_DIR)
                remote.download("kubeconfig", kubeconfig_path)
                cluster_name_path = "{0}/management_cluster_name".format(
                    settings.ARTIFACTS_DIR)
                cluster_uid_path = "{0}/management_cluster_uid".format(
                    settings.ARTIFACTS_DIR)

                LOG.info("Try to fetch mgm cluster UID from kubeconfig")
                kaas_manager = Manager(kubeconfig=kubeconfig_path)
                ns = kaas_manager.get_namespace(settings.CLUSTER_NAMESPACE)
                cluster = ns.get_cluster(settings.CLUSTER_NAME)

                kaas_uid_meta = cluster.get_cluster_annotation(
                    "kaas.mirantis.com/uid")
                cluster_uid = kaas_uid_meta["kaas.mirantis.com/uid"]
                LOG.info(f"Cluster deployed with UID - {cluster_uid}")

                with open(cluster_name_path, 'w') as fp:
                    fp.write(settings.CLUSTER_NAME)
                with open(cluster_uid_path, 'w') as fp:
                    fp.write(str(cluster_uid))

                if not settings.SEED_STANDALONE_EXTERNAL_IP:
                    LOG.info('Set KaaS UID to seed instance object')
                    seed_instance = self.get_seed_instance(
                        openstack_client)
                    seed_instance.manager.set_meta_item(
                        seed_instance, 'KaaS', cluster_uid)
                    seed_instance.manager.set_meta_item(
                        seed_instance, 'ENV_NAME', settings.ENV_NAME)
            except Exception as e:
                LOG.warning(
                    "Unable to collect logs and/or cluster meta: {}".format(e))

        if ret.exit_code == 0:
            LOG.info("Mgmt cluster has been deployed successfully")
        else:
            LOG.error("Mgmt cluster has not been deployed successfully")
            raise CalledProcessError(result=ret, expected=(0,))

    def step_004_prepare_seed_node_region_templates(self, openstack_client):
        """Generate region templates."""
        remote_templates_dir = f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/' \
                               f'templates.region'
        self.prepare_seed_node_region_templates(openstack_client,
                                                remote_templates_dir)

    @utils.log_method_time()
    def step_004_deploy_kaas_regional(self, openstack_client, region):
        """Run bootstrap.sh deploy_regional, download logs"""
        LOG.info("Verify that seed node have been prepared")
        remote = self.remote_seed(openstack_client)
        os_conf_path = f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/' \
                       f'templates/clouds-{region}.yaml'
        LOG.info(f"Multi-region is enabled."
                 f"Deploying with region {region}")
        run_envs = {
            "REGION": region,
            "KUBECONFIG": '/home/ubuntu/kubeconfig',
            "REGIONAL_CLUSTER_NAME": settings.REGIONAL_CLUSTER_NAME,
            "KAAS_BOOTSTRAP_LOG_LVL": settings.KAAS_BOOTSTRAP_LOG_LVL
        }

        # Check if offline is enabled if proxy is set.
        # Remove if such cases will appear
        if (settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
                and not settings.KAAS_OFFLINE_DEPLOYMENT):
            LOG.error("KAAS_EXTERNAL_PROXY_ACCESS_STR is set but "
                      "KAAS_OFFLINE_DEPLOYMENT variable is absent."
                      "It makes no sense do use proxy on online management "
                      "cluster.")
            raise RuntimeError('KAAS_OFFLINE_DEPLOYMENT is '
                               'absent but KAAS_EXTERNAL_PROXY_ACCESS_STR is '
                               'set. Please fill out KAAS_OFFLINE_DEPLOYMENT')

        if settings.KAAS_EXTERNAL_PROXY_ACCESS_STR:
            run_envs['HTTP_PROXY'] = settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            run_envs['HTTPS_PROXY'] = settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            LOG.info("HTTP_PROXY and HTTPS_PROXY is set to {0}".format(
                settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
            ))
            # PRODX-13381: add EU/US fip range to no_proxy
            if settings.KAAS_PROXYOBJECT_NO_PROXY:
                run_envs['NO_PROXY'] = \
                    f"{settings.KAAS_PROXYOBJECT_NO_PROXY},"\
                    f"{settings.OS_FIP_RANGES}"
            else:
                run_envs['NO_PROXY'] = settings.OS_FIP_RANGES
            run_envs['no_proxy'] = run_envs['NO_PROXY']
            LOG.info("NO_PROXY is set to {0}".format(
                run_envs['NO_PROXY']))

        if 'os' in region:
            clouds_yaml = templates.render_template(
                settings.CLOUDS_YAML_PATH)
            with remote.open(os_conf_path, mode='w') as r_f:
                r_f.write(clouds_yaml)
            run_envs['OS_CLIENT_CONFIG_FILE'] = '/home/ubuntu/{}'.format(
                os_conf_path)

        if settings.BM_DEPLOY_REGIONAL:
            # For BM region
            run_envs['SSH_KEY_NAME'] = f"bootstrap-key-{settings.BM_REGION}"
            if remote.isdir('/home/ubuntu/.ssh'):
                LOG.info(f'Copying /home/ubuntu/.ssh to '
                         f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/')
                remote.check_call(f'cp -r /home/ubuntu/.ssh '
                                  f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/')
            LOG.info(f'Copying /home/ubuntu/kaas-bootstrap/ssh_key to '
                     f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/')
            remote.check_call(f'cp /home/ubuntu/kaas-bootstrap/ssh_key '
                              f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/',
                              raise_on_err=False)
        if settings.KAAS_AWS_ENABLED:
            # For AWS region
            LOG.info('Enable AWS region')
            run_envs['KAAS_AWS_ENABLED'] = True
            run_envs['AWS_SECRET_ACCESS_KEY'] = settings.AWS_SECRET_ACCESS_KEY
            run_envs['AWS_ACCESS_KEY_ID'] = settings.AWS_ACCESS_KEY_ID
            run_envs['AWS_DEFAULT_REGION'] = settings.AWS_DEFAULT_REGION

        if settings.KAAS_VSPHERE_ENABLED:
            # For vSphere region
            LOG.info('Enable vSphere region')
            run_envs['KAAS_VSPHERE_ENABLED'] = True
            run_envs['KEYCLOAK_FLOATING_IP'] = settings.VSPHERE_KEYCLOAK_FLOATING_IP  # noqa
            run_envs['IAM_FLOATING_IP'] = settings.VSPHERE_IAM_FLOATING_IP

            if os.path.exists(
                    os.path.join(self.get_bootstrap_templates_path(
                                     remote=remote,
                                     openstack_client=openstack_client),
                                 'vsphere', 'regional')):
                LOG.info("Templates contain prerendered region template")
                region_templates = f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/' \
                                   f'templates/vsphere/regional'
            else:
                LOG.info("Render a vsphere tamplate for region cluster")
                self.step_004_prepare_seed_node_region_templates(
                    openstack_client)
                region_templates = f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/' \
                                   f'templates.region/vsphere'
            run_envs['TEMPLATES_DIR'] = f'{region_templates}'

        if settings.KAAS_EQUINIX_ENABLED:
            LOG.info('Enable Equinix region')
            run_envs['KAAS_EQUINIX_ENABLED'] = True
            LOG.info("Render a equinix tamplate for region cluster")
            self.step_004_prepare_seed_node_region_templates(
                openstack_client)
            region_templates = f'{settings.KAAS_BOOTSTRAP_TARGET_DIR}/' \
                f'templates.region/equinix'
            run_envs['TEMPLATES_DIR'] = f'{region_templates}'

        envs_string = utils.make_export_env_strting(run_envs)
        LOG.info('Current envs_string {}'.format(envs_string))
        try:
            ret_region = remote.execute(
                "{envs_string}; bash -x "
                "./{bootstrap}/bootstrap.sh deploy_regional".format(
                    envs_string=envs_string,
                    bootstrap=settings.KAAS_BOOTSTRAP_TARGET_DIR),
                verbose=True,
                timeout=settings.KAAS_BOOTSTRAP_TIMEOUT)
        except ExecHelperTimeoutError as e:
            LOG.error("Timeout for deploy regional cluster "
                      "is exceeded: {}".format(e))
            raise e
        finally:
            try:
                self.step_bs_collect_logs(remote)
                self.check_multiple_kaasreleases(remote)
                # Download management_kubeconfig artifact
                kubeconfig_path = "{0}/management_kubeconfig" \
                    .format(settings.ARTIFACTS_DIR)
                remote.download("kubeconfig", kubeconfig_path)
                # Download regional_kubeconfig artifact
                regional_kubeconfig_path = "{0}/regional_kubeconfig" \
                    .format(settings.ARTIFACTS_DIR)
                remote.download(
                    "kubeconfig-{}".format(settings.REGIONAL_CLUSTER_NAME),
                    regional_kubeconfig_path)
            except Exception as e:
                LOG.warning(
                    "Unable to collect logs and kubeconfig: {}".format(e))
            LOG.info("Mgmt cluster has been deployed successfully")
        if ret_region.exit_code == 0:
            LOG.info("Mgmt and regional clusters has been "
                     "deployed successfully")
        else:
            LOG.error("Mgmt or regional clusters has "
                      "not been deployed successfully:"
                      "result {0} ".format(ret_region))
            raise CalledProcessError(result=ret_region, expected=(0,))

    def step_005_check_versions(self, openstack_client):
        kubeconfig = self.get_remote_file_yaml(openstack_client, "kubeconfig")
        with tempfile.NamedTemporaryFile(mode="w+") as f:
            f.write(yaml.dump(kubeconfig))
            f.flush()
            kaas_manager = Manager(kubeconfig=f.name)
            kaasreleases = kaas_manager.get_kaasrelease_names()

        # define actual version of kaasrelease for mgmt cluster
        ns = kaas_manager.get_namespace(settings.CLUSTER_NAMESPACE)
        cluster = ns.get_cluster(settings.CLUSTER_NAME)
        actual_kaasrelease = cluster.get_kaasrelease_version()
        LOG.info("Actual (deployed) kaasrelease version is {}".format(
            actual_kaasrelease))

        # fetch actual kaasrelease crd
        kaas_release_crd = kaas_manager.get_kaasrelease(
            actual_kaasrelease).data['spec']
        # fetch actual helmbundle crd
        helm_bundle_crd = kaas_manager.get_helmbundle(
            settings.CLUSTER_NAME).data['spec']

        # "kaas-0-3-0" > "kaas-0-3-0-rc" > "kaas-0-2-0" > "kaas-0-1-0-rc"
        latest_kaasrelease = str(max([version.parse(x) for x in kaasreleases]))

        # just a warning if we use not the latest version
        if latest_kaasrelease != actual_kaasrelease:
            LOG.warning("Actual kaasrelease ({0}) is not the latest one ({1})"
                        .format(actual_kaasrelease, latest_kaasrelease))

        # get kaasrelease file
        kaas_release_filename = self.get_kaas_release_file_name()
        if settings.KAAS_RELEASES_FOLDER and \
                os.path.isdir(settings.KAAS_RELEASES_FOLDER) and \
                kaas_release_filename and not \
                settings.BOOTSTRAP_WITH_DEFAULTS:
            kaas_release_path = "{}/kaas/{}".format(
                settings.KAAS_RELEASES_REMOTE_FOLDER,
                kaas_release_filename)
        else:
            actualrelease = actual_kaasrelease.replace(
                "kaas-", "").replace("-", ".")
            if 'rc' in actual_kaasrelease:
                release_parts = actualrelease.split('.')
                kaas_release_path = "releases/" \
                                    "kaas/{0}.{1}.{2}-{3}.yaml" \
                    .format(release_parts[0], release_parts[1],
                            release_parts[2], release_parts[3])
            else:
                kaas_release_path = (
                    "{0}/releases/kaas/{1}.yaml").format(
                    settings.KAAS_BOOTSTRAP_TARGET_DIR,
                    actualrelease)

        LOG.info("Accessing kaasrelease file in {}".format(kaas_release_path))
        kaas_release_file = self.get_remote_file_yaml(
            openstack_client, kaas_release_path)['spec']

        LOG.debug("Provided kaas_release_file {0}:\n{1}".format(
            kaas_release_path, yaml.dump(kaas_release_file)))
        LOG.debug("Active kaas_release_crd from the management cluster:\n{0}"
                  .format(yaml.dump(kaas_release_crd)))

        # compare helmRepositories
        cdn_region_definition = settings.CDN_REGION
        if version.parse(actual_kaasrelease) <= version.parse("kaas-2-3-0"):
            cdn_region_definition = "{version}-{region}".format(
                    version=actual_kaasrelease, region=settings.CDN_REGION)
        assert cdn_region_definition in settings.CDN_REGION_DICT.keys(), \
            ("Failed to find suitable CDN URL according to {} definition"
             .format(cdn_region_definition))

        cdn_url = settings.CDN_REGION_DICT[cdn_region_definition]
        crd_helm_repos = kaas_release_crd.get('helmRepositories', [])
        file_helm_repos = kaas_release_file.get('helmRepositories', [])
        self.assertEqual(crd_helm_repos,
                         [{'name': x['name'],
                           'url': cdn_url + '/' + x['url']}
                          for x in file_helm_repos],
                         msg="helmRepositories in kaasrelease crd "
                             "and file are not the same")

        LOG.info("Prepare objects for kaas releases comparison")
        obj_types = {
                'kaas_release_k8s_object': kaas_release_crd,
                'kaas_release_file_template': kaas_release_file,
                }
        for entity, release in obj_types.items():
            LOG.debug(f"Before: {release}")

            LOG.info(f"Removing redundant helmRepositories in {entity}")
            release.pop('helmRepositories', None)

            LOG.info(f"Removing empty provider sections in {entity}")
            for r in release['supportedClusterReleases']:
                if 'providers' in r and not r['providers']:
                    LOG.info(f"Removing 'providers' from {r}")
                    del r['providers']
            LOG.debug(f"After: {release}")

        # compare kaasrelease
        # solution for yaml aliases - this dumps full yaml w/o aliases
        yaml.Dumper.ignore_aliases = lambda *args: True
        self.assertEqual(yaml.dump(kaas_release_crd),
                         yaml.dump(kaas_release_file),
                         msg="Actual kaasrelease crd {0} does not match the "
                             "expected kaasrelease from {1}".format(
                             actual_kaasrelease, kaas_release_filename))

        # compare helmbundles crd with helmbundles data in kaas_release_file
        helm_bundle_file = kaas_release_file['management']
        for chart_in_crd in helm_bundle_crd['releases']:
            for chart_in_file in helm_bundle_file['helmReleases']:
                if chart_in_file['name'] == chart_in_crd['name']:
                    ver_in_file = helpers.get_chart_version(chart_in_file)
                    ver_in_crd = helpers.get_chart_version(chart_in_crd)
                    msg = "Actual version of the chart {0} is {1}, while" \
                          " expected is {2}".format(chart_in_crd['name'],
                                                    ver_in_crd,
                                                    ver_in_file)
                    res = ver_in_file == ver_in_crd
                    assert res, msg

        # Fetch data from cluster object
        cluster = ns.get_cluster(settings.CLUSTER_NAME)
        # cluster_release_version is deployed version of tne cluster
        cluster_release_version = \
            cluster.data['spec']['providerSpec']['value']['release']
        cluster_release_version_current = \
            cluster.data['status']['providerStatus'][
                'releaseRefs']['current']['name']
        LOG.info("Actual (deployed) clusterrelease version is {}".format(
            cluster_release_version))
        assert cluster_release_version == cluster_release_version_current

        # Make sure we have cluster_release_version
        # in the list of clusterreleases crd objects
        assert cluster_release_version in [
            x.name for x in kaas_manager.get_clusterreleases()], \
            "Deployed cluster release version is not in clusterreleases list"

        # Comparing vs kaas_release object
        assert cluster_release_version == kaas_release_crd['clusterRelease'], \
            "Deployed cluster release version doesn't match version " \
            "specified in kaasrelease"
        assert cluster_release_version in [
            x['name'] for x in kaas_release_crd['supportedClusterReleases']], \
            "Deployed cluster release version is not in kaasrelease" \
            "Supported Cluster Releases list"

        # Comparing clusterrelease crd vs clusterrelease file
        # first, fetch short 'version' value which is filename for
        # cluster yaml
        cluster_release_filename = \
            [x['version'] for x in kaas_release_crd['supportedClusterReleases']
             if x['name'] == cluster_release_version][0]

        # get cluster file
        if settings.KAAS_RELEASES_FOLDER and \
                os.path.isdir(settings.KAAS_RELEASES_FOLDER) and \
                kaas_release_filename and not \
                settings.BOOTSTRAP_WITH_DEFAULTS:
            cluster_release_path = "{}/cluster/{}.yaml".format(
                settings.KAAS_RELEASES_REMOTE_FOLDER,
                cluster_release_filename)
        else:
            cluster_release_path = (
                "{0}/releases/cluster/{1}.yaml").format(
                settings.KAAS_BOOTSTRAP_TARGET_DIR,
                cluster_release_filename)

        LOG.info("Accessing clusterrelease file in {}".format(
            cluster_release_path))
        cluster_release_file = self.get_remote_file_yaml(
            openstack_client, cluster_release_path)['spec']

        assert cluster_release_file == kaas_manager.get_clusterrelease(
            cluster_release_version).read().spec

    @utils.log_method_time()
    def step_006_wait_for_pods(self):

        kubeconfig_path = f"{settings.ARTIFACTS_DIR}/management_kubeconfig"
        if not os.path.isfile(kubeconfig_path):
            raise Exception(f"KUBECONFIG {kubeconfig_path} not found in"
                            f" artifacts!")
        kaas_manager = Manager(kubeconfig=kubeconfig_path)
        ns = kaas_manager.get_namespace(settings.CLUSTER_NAMESPACE)
        cluster = ns.get_cluster(settings.CLUSTER_NAME)
        cluster.check.check_k8s_pods()

    @utils.log_method_time()
    def step_006_postdeployment_checks(self, openstack_client,
                                       check_svc=False):

        kubeconfig_path = f"{settings.ARTIFACTS_DIR}/management_kubeconfig"
        if not os.path.isfile(kubeconfig_path):
            raise Exception(f"KUBECONFIG {kubeconfig_path} not "
                            f"found in artifacts!")

        kaas_manager = Manager(kubeconfig=kubeconfig_path)
        ns = kaas_manager.get_namespace(settings.CLUSTER_NAMESPACE)
        cluster = ns.get_cluster(settings.CLUSTER_NAME)
        cluster.check.check_helmbundles()
        cluster.check.check_persistent_volumes_mounts()
        if cluster.lcm_type_is_ucp:
            # Check/wait for correct docker service replicas in cluster
            cluster.check.check_actual_expected_docker_services()
        cluster.check.check_actual_expected_pods(timeout=900)
        # skip this check for 2.0.0 version
        if cluster.get_kaasrelease_version() != '2.0.0':
            cluster.check.check_resource_discovery()
        if check_svc:
            cluster.check.check_svc_access(self.remote_seed(openstack_client))

    @utils.log_method_time()
    def step_007_download_bootstrap_artifacts(self,
                                              openstack_client,
                                              nsname=settings.CLUSTER_NAMESPACE, # noqa
                                              clname=settings.CLUSTER_NAME):
        """"
         Also support bootstrap from region.
         For region:
            acces to cluster - via mgmt kubeconf
            access to helmbundle - via region mgmt kubeconf
        """
        remote = self.remote_seed(openstack_client)
        # to collect image version from
        sys_namespaces = ['kaas', 'stacklight']

        # Here must be passed exactly management_kubeconfig, fetched on prev.
        # steps, to be able get access to all clusters
        kubeconfig_path_mgmt = f"{settings.ARTIFACTS_DIR}/" \
                               f"management_kubeconfig"
        kaas_manager_mgmt = Manager(kubeconfig=kubeconfig_path_mgmt)
        # define actual version of kaasrelease, based on mgmt cluster
        ns = kaas_manager_mgmt.get_namespace(nsname)
        cluster = ns.get_cluster(clname)

        if cluster.is_regional:
            # TODO(alexz): should be named with clustername-etc-region format?
            cluster_prefix = 'regional'
            # if we work with region, we need to init new one k8s Manager
            # do be able work with exactly region. kubeconfig must be fetched
            # from prev.steps
            kubeconfig_path = f"{settings.ARTIFACTS_DIR}/regional_kubeconfig"
            kaas_manager = Manager(kubeconfig=kubeconfig_path)
            kubectl_client = K8sCluster(kubeconfig=kubeconfig_path)
        else:
            cluster_prefix = 'management'
            kaas_manager = kaas_manager_mgmt
            kubectl_client = K8sCluster(kubeconfig=kubeconfig_path_mgmt)

        LOG.info(f"Cluster: {nsname}/{clname} is {cluster_prefix}")

        actual_kaasrelease = \
            cluster.get_kaasrelease_version().replace("kaas-",
                                                      "").replace("-", ".")
        # get kaasrelease file
        remote.download(
            f"{settings.KAAS_BOOTSTRAP_TARGET_DIR}releases/kaas"
            f"/{actual_kaasrelease}.yaml",
            f"{settings.ARTIFACTS_DIR}/{cluster_prefix}_kaasrelease.yaml")

        # copy key
        key_path = self.get_bootstrap_ssh_keyfile_path(remote)
        assert remote.download(
            key_path,
            f"{settings.ARTIFACTS_DIR}/{cluster_prefix}_id_rsa"), \
            f"Failed to download {key_path} SSH key from remote machine"

        key = utils.load_keyfile(f"{settings.ARTIFACTS_DIR}/"
                                 f"{cluster_prefix}_id_rsa")
        pub_key = key["public"]
        with open(f'{settings.ARTIFACTS_DIR}/{cluster_prefix}_id_rsa_pub',
                  mode='w') as f:
            f.write("ssh-rsa " + str(pub_key))

        ver = kubectl_client.api_version.get_code()
        k8s_version = "{0}.{1}".format(ver.major, ver.minor).replace("+", ".x")
        with open(f"{settings.ARTIFACTS_DIR}/{cluster_prefix}_k8s_version",
                  "w") as f:
            f.write(k8s_version)

        LOG.info(f"collect versions of images for all containers, "
                 f"in {sys_namespaces}")
        containers = []
        for ns in sys_namespaces:
            ns_containers = {}
            pods = \
                kubectl_client.pods.list_raw(namespace=ns).to_dict()['items']
            for pod in pods:
                for container in pod['spec']['containers']:
                    ns_containers[container['name']] = container['image']
            containers.append({ns: ns_containers})
        with open(f"{settings.ARTIFACTS_DIR}/{cluster_prefix}_containers",
                  "w") as f:
            f.write(yaml.dump(containers))

        remote.download(
            f"{settings.KAAS_BOOTSTRAP_TARGET_DIR}/out/cluster.yaml",
            f"{settings.ARTIFACTS_DIR}/{cluster_prefix}_cluster.yaml")
        remote.download(
            f"{settings.KAAS_BOOTSTRAP_TARGET_DIR}/out/machines.yaml",
            f"{settings.ARTIFACTS_DIR}/{cluster_prefix}_machines.yaml")

        bootstrap_version = self.get_bootstrap_version()
        with open(f"{settings.ARTIFACTS_DIR}/bootstrap_version",
                  "w") as f:
            f.write(bootstrap_version)
        helm_bundle_crd = kaas_manager.get_helmbundle(clname).data['spec']
        helm_charts_versions = [{"chart": x['name'],
                                 "version": helpers.get_chart_version(x)}
                                for x in helm_bundle_crd['releases']]
        with open(f"{settings.ARTIFACTS_DIR}/{cluster_prefix}_helmbundles",
                  "w") as f:
            f.write(yaml.dump(helm_charts_versions))

        public_services = kaas_manager.describe_public_services()
        services_txt = '\n'.join([
            "{0}: {1}".format(s, ' '.join([
                "<a href={0}>{0}</a>".format(h) for h in public_services[s]]))
            for s in public_services.keys()])
        passwords_file = f"{settings.KAAS_BOOTSTRAP_TARGET_DIR}/passwords.yaml"
        if remote.isfile(passwords_file):
            res = remote.execute('cat {0}'.format(passwords_file))
            services_txt += '\n' + res.stdout_str
        with open(f"{settings.ARTIFACTS_DIR}/{cluster_prefix}_public_urls",
                  "w") as f:
            f.write(services_txt)
        # Required only for kaas-bm virt lab. child_config rendered by ansible
        if remote.isfile(
                f"{settings.KAAS_BOOTSTRAP_TARGET_DIR}/child_config.yml"):
            remote.download(
                f"{settings.KAAS_BOOTSTRAP_TARGET_DIR}/child_config.yml",
                f"{settings.ARTIFACTS_DIR}/child_config.yml")
        cluster.provider_resources.save_artifact()

    def step_007_erase_env_after(self, openstack_client):
        remote = self.remote_seed(openstack_client)
        run_envs = {
            "CLUSTER_NAME": settings.CLUSTER_NAME,
            "REGIONAL_CLUSTER_NAME": settings.REGIONAL_CLUSTER_NAME,
            "KAAS_BOOTSTRAP_LOG_LVL": settings.KAAS_BOOTSTRAP_LOG_LVL
        }
        envs_string = utils.make_export_env_strting(run_envs)

        LOG.info("Cleanup KaaS cluster")
        remote.check_call(
            "{envs_string}; bash -x ./{bootstrap}/bootstrap.sh "
            "cleanup".format(
                envs_string=envs_string,
                bootstrap=settings.KAAS_BOOTSTRAP_TARGET_DIR),
            verbose=True)

    @staticmethod
    def step_bs_collect_logs(remote):
        LOG.info("Try collect logs")
        run_envs = {
            "CLUSTER_NAME": settings.CLUSTER_NAME,
            "REGIONAL_CLUSTER_NAME": settings.REGIONAL_CLUSTER_NAME,
            "KAAS_BOOTSTRAP_LOG_LVL": "0"  # set default verbosity level
        }
        if remote.isfile('./{}/kubeconfig'.format(
                settings.KAAS_BOOTSTRAP_TARGET_DIR)):
            run_envs['KUBECONFIG'] = './{}/kubeconfig'.format(
                settings.KAAS_BOOTSTRAP_TARGET_DIR)
        envs_string = utils.make_export_env_strting(run_envs)
        if settings.BM_COLLECT_LOGS:
            remote.execute(
                "{envs_string}; bash -x ./{bootstrap}/bootstrap.sh "
                "collect_logs".format(
                    envs_string=envs_string,
                    bootstrap=settings.KAAS_BOOTSTRAP_TARGET_DIR),
                verbose=True)
        else:
            LOG.warning("bootstap.sh collect_logs skipped!")
        LOG.info('Try fetch logs')
        if remote.isdir(f"./{settings.KAAS_BOOTSTRAP_TARGET_DIR}/logs"):
            remote.check_call(f"tar zcfv logs.tar.gz "
                              f"./{settings.KAAS_BOOTSTRAP_TARGET_DIR}/logs")
            remote.download("logs.tar.gz",
                            f"{settings.ARTIFACTS_DIR}/management_logs.tar.gz")
            LOG.info(f"Remote logs placed to {settings.ARTIFACTS_DIR}"
                     f"/management_logs.tar.gz")
        # preflight.txt supposed to be created by ansible
        pflog = "./{}/logs/preflight.txt".format(
            settings.KAAS_BOOTSTRAP_TARGET_DIR)
        if remote.isfile(pflog):
            remote.download(pflog,
                            "{0}/preflight.log".format(settings.ARTIFACTS_DIR))
        # fastlog supposed to be created by ansible
        flog = "./{}/logs/fastlog.txt".format(
            settings.KAAS_BOOTSTRAP_TARGET_DIR)
        if remote.isfile(flog):
            remote.download(flog,
                            "{0}/bootstrap.log".format(settings.ARTIFACTS_DIR))

    def step_collect_logs(self,
                          remote,
                          cluster_name=settings.CLUSTER_NAME,
                          cluster_namespace=settings.CLUSTER_NAMESPACE,
                          kubeconfig_path="kubeconfig",
                          keyfile_path="",
                          bootstrap_dir=settings.KAAS_BOOTSTRAP_TARGET_DIR,
                          log_level=settings.KAAS_BOOTSTRAP_LOG_LVL):

        if not settings.BM_COLLECT_LOGS:
            LOG.warning("Step step_collect_logs skipped!")
            # simulate exit_code == 0
            res = remote.execute("exit 0")
            return res
        if not keyfile_path:
            keyfile_path = self.get_bootstrap_ssh_keyfile_path(remote)

        logs_dir = "./{0}/logs/{1}".format(bootstrap_dir, cluster_name)
        LOG.info("Cleanup path '{}' before bootstrap.sh collect_logs"
                 .format(logs_dir))
        remote.execute("rm -rf {}".format(logs_dir), verbose=True)

        mgmt_kubeconf = "kubeconfig" if remote.isfile("kubeconfig") else \
            "{0}/kubeconfig".format(bootstrap_dir)

        if kubeconfig_path:
            kubeconfig = "--kubeconfig {}".format(kubeconfig_path)
        else:
            kubeconfig = ""
        LOG.info("Try to collect logs")
        res = remote.execute(
            "./{0}/kaas collect logs "
            "{1} "
            "--v {2} "
            "--cluster-name {3} "
            "--cluster-namespace {4} "
            "--key-file {5} "
            "--management-kubeconfig {6} "
            "--output-dir ./{0}/logs".format(bootstrap_dir, kubeconfig,
                                             log_level, cluster_name,
                                             cluster_namespace,
                                             keyfile_path,
                                             mgmt_kubeconf),
            verbose=True)

        LOG.info('Try to fetch logs')
        tar_name = "{}_logs.tar.gz".format(cluster_name)
        if remote.isdir("./{}/logs/{}".format(bootstrap_dir, cluster_name)):
            remote.check_call("tar zcfv logs.tar.gz ./{}/logs/{}".format(
                bootstrap_dir, cluster_name))
            remote.download("logs.tar.gz",
                            "{0}/{1}"
                            .format(settings.ARTIFACTS_DIR, tar_name))
            LOG.info("Remote logs placed to {0}"
                     "{1}".format(settings.ARTIFACTS_DIR, tar_name))
        return res

    # Always run only after self.step_bs_collect_logs()
    @staticmethod
    def check_multiple_kaasreleases(remote):
        """Check for multiple kaasreleases after bootstrap

        Raises Exception if more than one KaaSRelease appeared
        in the kind cluster (there should not be KaaS updates)

        Logs error if more than one KaaSRelease appeared in the
        management cluster (can be normal when bug PRODX-4819
        is fixed, if the update is started right after successful
        deploy and logs are collected at this moment)
        """
        ls_res = remote.execute(
            "ls -1 {0}/logs/kaas-bootstrap-cluster/"
            "objects/cluster/kaas.mirantis.com/kaasreleases/"
            .format(settings.KAAS_BOOTSTRAP_TARGET_DIR))
        if ls_res.exit_code == 0:
            kaasreleases = ls_res.stdout
            if len(kaasreleases) > 1:
                raise Exception("KaaSReleases number more than one: {0}"
                                .format(','.join(kaasreleases)))

        ls_res = remote.execute(
            "ls -1 {0}/logs/kaas-management-cluster/"
            "objects/cluster/kaas.mirantis.com/kaasreleases/")
        if ls_res.exit_code == 0:
            kaasreleases = ls_res.stdout
            if len(kaasreleases) > 1:
                LOG.error("KaaSReleases number more than one: {0}"
                          .format(','.join(kaasreleases)))

    @staticmethod
    def expose_kubeconfig(kconfig, toIP):
        """
        In virtual case we do not have access to fake VIP address
        10.0.0.100. We should replace it to FIP from openstack assigned
        to neutron port.
        Also we can't connect to k8s using another server
        address and we must use insecure option there.

        :param kconfig: input Dict()
        :param toIP:
        :return:
        """
        import copy
        from urllib import parse
        result = copy.deepcopy(kconfig)

        for index, kcluster in enumerate(result.get('clusters', [])):
            oldIp = parse.urlparse(kcluster['cluster']['server']
                                   ).netloc.split(':')[0]
            LOG.info("Replace server: <{}> addr in kubeconfig to <{}>"
                     .format(oldIp, toIP))
            kcluster['cluster']['server'] = kcluster['cluster'][
                'server'].replace(oldIp, toIP)
            kcluster['cluster']['insecure-skip-tls-verify'] = True
            result['clusters'][index] = kcluster
        return result

    def ansible_init(self, openstack_client):
        """
        Install venv with ansible.
        TODO: parametrize?
        :return:
        """
        remote = self.remote_seed(openstack_client)
        seed_key = settings.SEED_SSH_PRIV_KEY_FILE
        if remote.check_call('[ -d ~/ansible-venv ]',
                             verbose=True,
                             raise_on_err=False).exit_code == 0:
            LOG.info('Ansible already init')
            return

        LOG.info('Init ansible venv')
        remote.check_call('python3.8 -m venv  --system-site-packages'
                          ' --copies  ~/ansible-venv')
        remote.check_call('source ~/ansible-venv/bin/activate;'
                          ' pip3 install --upgrade pip')
        remote.check_call('source ~/ansible-venv/bin/activate;'
                          ' pip3 install -U wheel jmespath'
                          ' netaddr ansible Jinja2')
        remote.check_call('mkdir -vp ~/bootstrap ~/.ansible_state',
                          verbose=True)
        kaas_bm_env_dir = os.path.join(settings.SI_TESTS_REPO_ROOT,
                                       'si_tests/kaas-bm-env')
        self.rsync_run(openstack_client, from_dir=kaas_bm_env_dir,
                       todir='~/kaas-bm-env/')
        remote.check_call("cd ~/kaas-bm-env/; "
                          "source ~/ansible-venv/bin/activate;"
                          "ansible-galaxy install -r requirements.yml;",
                          verbose=True)
        # key required to get acesss to KVM nodes accross env
        LOG.info('Upload kaas-bm-env-key to seed_node')
        self.rsync_run(openstack_client, from_file=seed_key,
                       tofile='~/kaas-bm-env-key')

    def ansible_run(self, remote, playbook,
                    inventory=None, extra_vars=None, **kwargs):
        """
        Run pre-setup ansible-playbook.
        :param cmds:
        :return:
        """

        if not extra_vars:
            extra_vars = {}
        elif not isinstance(extra_vars, dict):
            raise Exception("'extra_vars' could be only 'dict', "
                            "got {}".format(extra_vars))

        cmd = []
        if inventory:
            if isinstance(inventory, list):
                for item in inventory:
                    cmd.append("-i '{}'".format(item))
            elif isinstance(inventory, str):
                cmd.append("-i '{}'".format(inventory))
            else:
                raise Exception("Inventory could be either string or list, "
                                "got '{}'".format(inventory))

        if settings.FEATURE_FLAGS.len():
            extra_vars['feature_flags'] = settings.FEATURE_FLAGS.list()

        if extra_vars:
            cmd.append("-e '{}'".format(json.dumps(extra_vars)))

        cmd.append(playbook)

        return remote.check_call(
            "cd ~/kaas-bm-env/; "
            "source ~/ansible-venv/bin/activate; "
            "ANSIBLE_CONFIG=~/kaas-bm-env/ansible.cfg "
            "$(which ansible-playbook) {}".format(' '.join(cmd)),
            verbose=True, **kwargs)

    @retry(Exception, delay=5, tries=3, logger=LOG)
    def rsync_run(self, openstack_client, username=settings.SEED_SSH_LOGIN,
                  sshkeyfile=settings.SEED_SSH_PRIV_KEY_FILE, reverse=False,
                  hostname=None, from_dir=None, todir=None, from_file=None,
                  tofile=None):
        """
        Run local rsync call
        :param username:
        :param hostname:
        :param sshkeyfile:
        :param from_dir:
        :param todir:
        :param from_file:
        :param tofile:
        :param openstack_client:
        :param reverse: False -> sync from local to host
                        True -> sync from host to local
        :return:
        """
        hostname = hostname or self.get_seed_ip(openstack_client)
        sshopts = f'"ssh -i {sshkeyfile} -o StrictHostKeyChecking=no' \
                  ' -o ConnectionAttempts=10 ' \
                  ' -o UserKnownHostsFile=/dev/null "'
        if not (todir or tofile):
            raise Exception("Target not set")
        r = Subprocess()
        r.logger.addHandler(logger.console)
        if from_file:
            _to = tofile or todir
            if reverse:
                LOG.info(f"Downloading '{hostname}:{from_file}' to "
                         f"'{_to}'")
                cmd = f'rsync -e {sshopts} -azqP ' \
                      f'{username}@{hostname}:{from_file} {_to}'
                r.check_call(cmd, raise_on_err=True, verbose=True)
                LOG.info(f"Downloaded '{hostname}:{from_file}' to "
                         f"'{_to}'")
            else:
                cmd = f'rsync -e {sshopts} -azqP ' \
                      f'{from_file} {username}@{hostname}:{_to}'
                LOG.info(f"Uploading '{from_file}' to "
                         f"'{hostname}:{_to}'")
                r.check_call(cmd, raise_on_err=True, verbose=True)
                LOG.info(f"Uploaded '{from_file}' to "
                         f"'{hostname}:{tofile or todir}'")
        elif from_dir:
            if reverse:
                from_dir = from_dir + '/'
                cmd = f'rsync -e {sshopts}  -azqP ' \
                      f'{username}@{hostname}:{from_dir} .'
                with utils.pushd(todir):
                    LOG.info(f"Downloading '{hostname}:{from_dir}' dir to"
                             f" '{todir}'")
                    r.check_call(cmd, raise_on_err=True, verbose=True)
                    LOG.info(f"Downloaded '{from_dir}' dir to"
                             f" '{todir}'")
            else:
                cmd = f'rsync -e {sshopts}  -azqP . ' \
                      f'{username}@{hostname}:{todir}'
                with utils.pushd(from_dir):
                    LOG.info(f"Uploading '{from_dir}' dir to"
                             f" '{hostname}:{todir}'")
                    r.check_call(cmd, raise_on_err=True, verbose=True)
                    LOG.info(f"Uploaded '{from_dir}' dir to"
                             f" '{hostname}:{todir}'")
        else:
            raise Exception('Wrong invocation')

    def check_keystone_ldap_integration(self):
        LOG.info("Check keystone LDAP integration")
        # Get cluster
        kubeconfig_path = "{0}/management_kubeconfig" \
            .format(settings.ARTIFACTS_DIR)
        if not os.path.isfile(kubeconfig_path):
            raise Exception("KUBECONFIG {0} not found in artifacts!"
                            .format(kubeconfig_path))
        kaas_manager = Manager(kubeconfig=kubeconfig_path)
        ns = kaas_manager.get_namespace(settings.CLUSTER_NAMESPACE)
        cluster = ns.get_cluster(settings.CLUSTER_NAME)

        # Get iam_api_client
        keycloak_ip = kaas_manager.get_keycloak_ip()
        iam_ip_port = kaas_manager.get_iam_ip_port()
        cacert_path = 'keycloakCA.pem'
        cacert = kaas_manager.get_secret_data('iam-api-secrets',
                                              'kaas', cacert_path)
        with open(cacert_path, 'w') as f:
            f.write(cacert)
        iam_api = iam_client.IAMClient(keycloak_ip=keycloak_ip,
                                       iam_ip_port=iam_ip_port,
                                       cacert_path=cacert_path)

        # keycloak_admin_client
        pwd = kaas_manager.get_secret_data('iam-api-secrets',
                                           'kaas', 'keycloak_password')

        keycloak_admin_client = keycloak_client.KeycloakAdminClient(
            ip=keycloak_ip, user=settings.KEYCLOAK_USER, password=pwd)

        # iam_api_client
        admin_username = 'si-test'
        admin_password = '123'
        keycloak_admin_client.user_create(username=admin_username,
                                          password=admin_password)
        # add admin role
        keycloak_admin_client.assign_roles(username=admin_username,
                                           roles=["m:iam@admin"])
        iam_api_client = iam_api.connect(username=admin_username,
                                         password=admin_password)

        check_resource_path = "apis/kaas.mirantis.com/v1alpha1/clusterreleases"

        # Negative check
        # Also required to force sync for the LDAP user to Keycloak "Users"
        # so we can grant the role to that LDAP user later.
        LOG.info(f"Login to UI using {settings.KAAS_UI_USER}, and check that "
                 f"access to the {check_resource_path} is restricted")
        d_client = cluster.get_decc_dashboardclient()
        d_resources = d_client.get_resource(check_resource_path,
                                            raise_on_error=False)
        LOG.debug(f"Response from {check_resource_path} using a new LDAP user "
                  f"(must be restricted for the user with no roles):\n"
                  f"{d_resources}")
        assert 'code' in d_resources and d_resources['code'] == 403, (
            f"Access to {check_resource_path} is allowed "
            f"for the LDAP user with no roles, while it should be restricted")

        # Grant 'writer' role
        LOG.info(f"Grant 'writer' role to {settings.KAAS_UI_USER}")
        username = settings.KAAS_UI_USER
        iam_api_client.give_grants(username=username, role='writer',
                                   scope='m:kaas')
        iam_api_client.give_grants(username=username, role='admin',
                                   scope='m:os')
        grants = [{'role': 'writer', 'scope': 'm:kaas'},
                  {'role': 'admin', 'scope': 'm:os'}]
        result = iam_api_client.has_grants(username, grants)
        assert not result['missing_grants'], \
            "Grant lists are not the same. Result: {0}".format(result)

        # Re-login with new grants
        LOG.info(f"Re-login to UI using {settings.KAAS_UI_USER}, and check"
                 f"that access to the {check_resource_path} is allowed now")
        d_client = cluster.get_decc_dashboardclient()
        # Positive check
        d_resources = d_client.get_resource(check_resource_path)
        LOG.debug(f"{check_resource_path} read with LDAP credentials:\n"
                  f"{d_resources}")
        # Remove temporary admin user
        keycloak_admin_client.user_delete(username=admin_username)
        LOG.info("Keystone LDAP integration check is PASSED")
