import ast
import re
from kubernetes.client.rest import ApiException

from si_tests.managers.openstack_manager import OpenStackManager
from si_tests.deployments.utils import commons
from si_tests.clients.openstack_client.resources import OpenStackResource
from si_tests.clients.openstack_client.resource_providers import VMProvider


class OpenStackClientManager:
    pattern = re.compile(r'(?<!^)(?=[A-Z])')

    def __new__(cls, *args, **kwargs):
        cls.os_manager = OpenStackManager()
        cls.client = cls.os_manager.api.pods.list_starts_with(
            'keystone-client', cls.os_manager.openstack_namespace)[0]

        for sub_cls in OpenStackResource.__subclasses__():
            name = cls.pattern.sub('_', sub_cls.__name__).lower()
            setattr(cls, name, sub_cls(cls.client))
        return super().__new__(cls, *args, **kwargs)

    def __init__(self):
        self._refapp_url = None

    def reload_client(self):
        self.client = self.os_manager.api.pods.list_starts_with(
            'keystone-client', self.os_manager.openstack_namespace)[0]
        for sub_cls in OpenStackResource.__subclasses__():
            name = self.pattern.sub('_', sub_cls.__name__).lower()
            setattr(self, name, sub_cls(self.client))

    def _pod_cmd_client(self, label_selector, hostname):
        pod_client = self.os_manager.api.pods.list_starts_with(
            '', self.os_manager.openstack_namespace,
            label_selector=label_selector,
            field_selector=f"spec.nodeName={hostname}"
        )[0]
        return pod_client

    def exec_neutron_ovs_agent(self, hostname, request):
        label_selector = "application=neutron,component=neutron-ovs-agent"
        return self._pod_cmd_client(label_selector, hostname).exec(request)

    def exec_vrouter_agent(self, hostname, request):
        label_selector = "app=tf-vrouter-agent-dpdk"
        pod_client = self.os_manager.api.pods.list_starts_with(
            '', self.os_manager.tf_namespace,
            label_selector=label_selector,
            field_selector=f"spec.nodeName={hostname}"
        )
        return pod_client[0].exec(request, container='agent')

    def exec_libvirt(self, hostname, request):
        label_selector = "application=libvirt,component=libvirt"
        return self._pod_cmd_client(label_selector, hostname).exec(
            request, container='libvirt'
        )

    def virsh_dumpXML(self, hypervisor_hostname, domain):
        return self.exec_libvirt(
            hypervisor_hostname, ["virsh", "dumpxml", domain])

    def virsh_capabilities(self, hypervisor_hostname):
        return self.exec_libvirt(
            hypervisor_hostname, ["virsh", "capabilities"])

    def provide_vm(self, vm_name):
        return VMProvider(self, vm_name)

    def _exec(self, request):
        try:
            response = self.client.exec(request)
        except ApiException as e:
            if e.reason != 'Handshake status 404 Not Found':
                raise e
            self.reload_client()
            response = self.client.exec(request)
        return response

    def _refapp_request(self, request):
        response = self._exec(request)
        try:
            return ast.literal_eval(response)
        except Exception:
            commons.LOG.error("Request %s failed with: %s",
                              request, response)
            return {}

    def _get_refapp_url(self):
        stack_outputs = self.stack.show(
            ["refapp-test", "-c", "outputs"])["outputs"]
        url = list(filter(
            lambda x: x["output_key"] == "app_url", stack_outputs)
        )[0]["output_value"]
        return url

    @property
    def refapp_url(self):
        if not self._refapp_url:
            self._refapp_url = self._get_refapp_url()
        return self._refapp_url

    def get_refapp_node_identifier(self):
        request = ["curl", "-s", self.refapp_url]
        return self._refapp_request(request).get("host", None)

    def create_refapp_record(self, data="{\"record\":{}}"):
        request = [
            "curl", "-X", "POST", "-s", "{}/records".format(self.refapp_url),
            "-H", "Content-Type: application/json",
            "--data", data]
        return self._refapp_request(request).get(
            'record', {})

    def get_refapp_record_by_id(self, record_id):
        url = "{url}/records/{id}".format(url=self.refapp_url,
                                          id=record_id)
        return self._refapp_request(["curl", "-s", url]).get(
            'record', {})

    def get_refapp_records(self):
        url = "{}/records".format(self.refapp_url)
        request = ["curl", "-s", url]
        return self._refapp_request(request).get(
            'records', {})
