import os
import json
import random
import time
from kubernetes.client.rest import ApiException

from si_tests import settings
from si_tests.clients.k8s import K8sCluster
from si_tests import logger
from si_tests.utils import waiters

LOG = logger.logger


class OpenStackManager(object):
    kubeconfig = None

    def __init__(self, kubeconfig=settings.KUBECONFIG_PATH):
        self._api = None
        self.kubeconfig = kubeconfig
        self.openstack_namespace = 'openstack'
        self.os_helm_system_namespace = 'osh-system'
        self.tf_namespace = 'tf'

    @property
    def api(self):
        """
            :rtype: cluster.K8sCluster
        """
        if self._api is None:
            if not os.path.isfile(self.kubeconfig):
                raise FileNotFoundError(
                    'kubeconfig file={} not found!'.format(self.kubeconfig))
            self._api = K8sCluster(kubeconfig=self.kubeconfig)
        return self._api

    def get_os_resources(self, resources_type, read=False):
        resources_api = getattr(self.api, resources_type)
        resources = resources_api.list(self.openstack_namespace)
        if read:
            return [resource.read() for resource in resources]
        return resources

    def get_os_resource(self,
                        resources_type,
                        resource_name,
                        namespace,
                        read=False):
        resources_api = getattr(self.api, resources_type)
        resource = resources_api.get(name=resource_name,
                                     namespace=namespace)
        if read:
            return resource.read()
        return resource

    def get_os_deployments(self, read=False):
        return self.get_os_resources('deployments',
                                     read)

    def get_os_helmbundles(self, read=False):
        return self.get_os_resources('kaas_helmbundles',
                                     read)

    def get_os_statefulsets(self, read=False):
        return self.get_os_resources('statefulsets',
                                     read)

    def get_os_secrets(self, read=False):
        return self.get_os_resources('secrets',
                                     read)

    def get_os_daemonsets(self, read=False):
        return self.get_os_resources('daemonsets',
                                     read)

    def get_os_configmaps(self, read=False):
        return self.get_os_resources('configmaps',
                                     read)

    def get_os_pods(self, read=False):
        return self.get_os_resources('pods', read)

    def get_os_operator_helmbundle(self, name, read=False):
        return self.get_os_resource('kaas_helmbundles',
                                    name,
                                    self.os_helm_system_namespace,
                                    read)

    def get_os_helmbundle(self, name, read=False):
        return self.get_os_resource('kaas_helmbundles',
                                    name,
                                    self.openstack_namespace,
                                    read)

    def get_openstackdeployment(self, name, read=False):
        return self.get_os_resource('openstackdeployment',
                                    name,
                                    self.openstack_namespace,
                                    read)

    def get_os_statefulset(self, name, read=False):
        return self.get_os_resource('statefulsets',
                                    name,
                                    self.openstack_namespace,
                                    read)

    def check_os_chart_contain_key_value(self, chart_pattern, key, value):
        os_helmbundles = self.get_os_helmbundles(read=True)
        for hb in os_helmbundles:
            for chart in hb.spec['releases']:
                chart_name = chart['chart']
                if chart_pattern in chart_name:
                    err_msg = "{key}: {value} does not contains in chart " \
                              "{chart}".format(key=key,
                                               value=value,
                                               chart=chart_name)
                    assert (chart['values'][key] == value), err_msg

    def get_os_helmbundle_versions(self):
        os_helmbundles = self.get_os_helmbundles(read=True)
        os_helmbundles_version_map = {}
        for hb in os_helmbundles:
            fp = json.loads(hb.metadata.annotations.get(
                'lcm.mirantis.com/openstack-controller-fingerprint'))
            os_helmbundles_version_map[hb.metadata.name] = fp
        return os_helmbundles_version_map

    def get_os_helmbundle_images(self):
        os_helmbundles = self.get_os_helmbundles(read=True)
        os_helmbundles_images_map = {}
        for hb in os_helmbundles:
            hb_name = hb.metadata.name
            os_helmbundles_images_map[hb_name] = {}
            for release in hb.spec['releases']:
                if 'images' in release['values']:
                    os_helmbundles_images_map[hb_name][release['name']] = \
                        release['values']['images']['tags']
        return os_helmbundles_images_map

    def check_resource_version_updated(self,
                                       resource_list_before,
                                       resource_list_after,
                                       updated=True,
                                       skip_list=None):
        """Check metadata resourceVersion parameter update
        Args:
            :param resource_list_before: (list) list of objects before
            :param resource_list_after: (list) list of objects after
            :param skip_list: (list) list of resource names to skip check
            :param updated: (bool) should be version updated or not
        """

        skip_list = skip_list or []
        before = {resource.metadata.name: resource for resource in
                  resource_list_before}
        for resource_after in resource_list_after:
            resource_before = before[resource_after.metadata.name]
            if resource_before.metadata.name in skip_list:
                continue
            resource_version_before = \
                resource_before.metadata.resource_version
            resource_version_after = \
                resource_after.metadata.resource_version

            err_msg = ("resourceVersion was updated for"
                       " {name} {uid} from {before} to {after}".format(
                           name=resource_before.metadata.name,
                           uid=resource_before.metadata.uid,
                           before=resource_version_before,
                           after=resource_version_after))
            actual = not (
                    resource_version_before == resource_version_after)
            assert updated == actual, err_msg

    def check_resource_generation_bumped(self,
                                         resource_list_before,
                                         resource_list_after,
                                         bumped=True):
        """Check metadata generation bumped
        Args:
            :param resource_list_before: (list) list of objects before
            :param resource_list_after: (list) list of objects after
            :param bumped: (bool) should be version bumped or not
        """

        before = {resource.metadata.name: resource for resource in
                  resource_list_before}
        for resource_after in resource_list_after:
            resource_before = before[resource_after.metadata.name]
            gen_before = resource_before.metadata.generation
            gen_after = resource_after.metadata.generation

            err_msg = ("Generation bump check failed for {name}. "
                       "Should updated: {bumped}, before: {before}, "
                       "after: {after}.".format(
                           name=resource_before.metadata.name,
                           bumped=bumped,
                           before=gen_before,
                           after=gen_after))
            actual = not (gen_before == gen_after)
            assert bumped == actual, err_msg

    def get_osdpl_deployment(self, read=False):
        """
        Get OSDPL deployment
        :return: osdpl deployment
        :rtype: OpenStackDeployment
        """
        return self.get_openstackdeployment(
            settings.OSH_DEPLOYMENT_NAME, read)

    def check_osdpl_status_value(self, section, value, equal=True):
        """Wait osdpl status section value
        Args:
            :param section: (string) section to check value
            :param value: value to check
            :param equal: (bool) should be equal or not
        """

        osdpl = self.get_osdpl_deployment(read=True)
        actual = osdpl.status[section] == value
        return actual == equal

    def wait_openstack_helmbundles_fingerprint(self, name,
                                               timeout=300, interval=60):
        """Wait till all helmbundles gets fingerprint of current
           OpenStack deployment
        Args:
            :param name: name of OpenStack deployment
            :param timeout: timeout to wait
            :param interval: time between checks
        """

        def _check_helmbundles_fingerprint():
            helmbundles_version_map = self.get_os_helmbundle_versions()
            if all(version['fingerprint'] == expected_fingerprint
                   for version in helmbundles_version_map.values()):
                return True
            return False

        osdpl = self.get_openstackdeployment(name,
                                             read=True)
        expected_fingerprint = osdpl.status['fingerprint']

        err = "Openstack helmbundles fingerprint wasn't" \
              " updated to actual version of" \
              " openstackdeployment: {expected_fingerprint}".format(
                expected_fingerprint=expected_fingerprint)
        waiters.wait(
            _check_helmbundles_fingerprint,
            timeout=timeout,
            interval=interval,
            timeout_msg=err)
        LOG.info("OpenStack Helmbundles fingerprints are in actual state")

    def wait_openstack_helmbundles_version_update(self, name,
                                                  timeout=600, interval=60):
        """Wait till all helmbundles gets version of current
           OpenStack deployment
        Args:
            :param name: name of OpenStack deployment
            :param timeout: timeout to wait
            :param interval: time between checks
        """

        def _check_helmbundles_version():
            helmbundles_version_map = self.get_os_helmbundle_versions()
            if all(version['version'] == expected_version
                   for version in helmbundles_version_map.values()):
                return True
            return False

        osdpl = self.get_openstackdeployment(name)
        expected_version = osdpl.read().status['version']

        err = "Openstack helmbundles version wasn't" \
              " updated to actual version of" \
              " openstackdeployment: {expected_version}".format(
                expected_version=expected_version)
        waiters.wait(
            _check_helmbundles_version,
            timeout=timeout,
            interval=interval,
            timeout_msg=err)
        LOG.info("OpenStack Helmbundles versions are in actual state")

    def wait_openstackdeployment_health_status(self,
                                               expected_status='Ready',
                                               timeout=1800,
                                               interval=90):

        def _check_osdpl_health():
            failed_services = []
            health_map = self.get_openstackdeployment(
                settings.OSH_DEPLOYMENT_NAME).read().status['health']
            for service_name, subservice_map in health_map.items():
                for subservice_name, subservice_statuses in \
                        subservice_map.items():
                    if subservice_statuses['status'] != expected_status:
                        failed_services.append(
                            "{sn}/{sbn}: {status}".format(
                                sn=service_name,
                                sbn=subservice_name,
                                status=subservice_statuses['status']))

            if failed_services:
                msg = "Timeout waiting for services statuses." \
                      "After {timeout} sec next services are not " \
                      "in {expected_status} state: {failed_services}.".format(
                          timeout=timeout,
                          expected_status=expected_status,
                          failed_services=failed_services)
                LOG.warning(msg)
                return False
            return True

        waiters.wait(_check_osdpl_health,
                     timeout=timeout,
                     interval=interval)
        LOG.info("All Services in OpenStack deployment are in "
                 "{status} status.".format(status=expected_status))

    def wait_osdpl_status(self, timeout):
        LOG.info("Wait osdpl.status.deployed is True")
        waiters.wait((lambda expected:
                      expected == self.get_openstackdeployment(
                          settings.OSH_DEPLOYMENT_NAME
                      ).read().status['deployed']),
                     predicate_kwargs={'expected': True},
                     timeout=timeout,
                     timeout_msg="OpenStack deployment doesn't reached"
                                 " status `deployed=True`")

        LOG.info("Wait until all osdpl children statuses success=True")
        self.wait_all_osdpl_children_status()

    def wait_all_osdpl_children_status(self,
                                       success=True,
                                       timeout=600,
                                       interval=60):
        """
        Wait all osdpl children status
        :param success: expected children status
        :param timeout: timeout to wait
        :param interval: time between checks
        """
        def _check_children_statuses():
            osdpl_children_status_map = self.get_openstackdeployment(
                settings.OSH_DEPLOYMENT_NAME).read().status['children']
            for hb, status in osdpl_children_status_map.items():
                if status == success:
                    continue
                return False
            return True

        err = "Doesn't reached status success={success} for " \
              "Helmbundles in OpenStack deployment".format(
                  success=success)
        waiters.wait(_check_children_statuses,
                     timeout=timeout,
                     interval=interval,
                     timeout_msg=err)

    @staticmethod
    def _is_owner_of_kind(resource, kind):
        references = resource.metadata.owner_references
        if references:
            return references[0].kind == kind
        return False

    def wait_os_resources(self, timeout=300, interval=60):
        start = time.time()
        self.wait_all_os_pods(timeout, interval)
        timeout = max(1, int(timeout - (time.time() - start)))
        self.wait_all_os_jobs(timeout, interval)

    def wait_all_os_pods(self, timeout=300, interval=60):
        def _wait_pods():
            _pods = self.api.pods.list(namespace=self.openstack_namespace)
            pods = []
            for p in _pods:
                try:
                    p_data = p.read()
                    pods.append(p_data)
                except ApiException:
                    continue
            inactive = [p for p in pods if p.status.phase.lower()
                        not in ('running', 'succeeded')]
            # filter failed pods owned by jobs, as
            # some jobs can have restartPolicy Never
            inactive_job_pods = [pod for pod in inactive
                                 if self._is_owner_of_kind(pod, "Job")]

            for pod in inactive_job_pods:
                inactive.remove(pod)

            def _pod_mapper(pod):
                return pod.metadata.name, pod.status.phase

            LOG.info('Namespace has inactive service %s pods and job %s pods',
                     list(map(_pod_mapper, inactive)),
                     list(map(_pod_mapper, inactive_job_pods)))

            assert not inactive, \
                "Some pods {} wasn't active/finished in " \
                "proper time {}".format(pods, timeout)

        waiters.wait_pass(
            _wait_pods, AssertionError, interval, timeout)

    def wait_all_os_jobs(self, timeout=300, interval=60):
        def _wait_jobs():
            jobs = self.api.jobs.list(namespace=self.openstack_namespace)
            not_cron = []
            for job in jobs:
                try:
                    if not self._is_owner_of_kind(job.read(), "CronJob"):
                        not_cron.append(job)
                except ApiException:
                    continue

            not_succeeded = []
            for job in not_cron:
                try:
                    if not job.read().status.succeeded == 1:
                        not_succeeded.append(job.name)
                except ApiException:
                    continue
            assert not not_succeeded, "Not all jobs succeeded {}".format(
                not_succeeded)

        waiters.wait_pass(_wait_jobs, AssertionError, interval, timeout)

    @staticmethod
    def get_k8s_object_field(status, field):
        return getattr(status, field) or 0

    def wait_k8s_object_replicas_restored(self, k8s_object,
                                          k8s_object_type,
                                          replicas):
        """
        Wait for k8s object replicas restored to needed quantity.

        :param k8s_object: k8s object that has read and
            patch_scale methods

        :param k8s_object_type: tuple of k8s_object.name
            and k8s_object.resource_type

        :param replicas: number of desired replicas
        """
        get_field = self.get_k8s_object_field

        def check():
            status = k8s_object.read().status
            return get_field(status, "replicas") == get_field(
                status, "current_replicas") == get_field(
                status, "ready_replicas") == replicas

        LOG.info("Wait for %s of type %s replicas successfully reached"
                 " required value",
                 *k8s_object_type)
        waiters.wait(check, interval=30,
                     timeout=settings.OPENSTACK_LCM_OPERATIONS_TIMEOUT)

    def change_k8s_object_replicas_count(self,
                                         k8s_object,
                                         k8s_object_type,
                                         replicas):
        """
        Change replicas count for k8s_object.
        Verify that replicas are successfully changed.

        :param k8s_object: k8s object that has read and
            patch_scale methods

        :param k8s_object_type: tuple of k8s_object.name
            and k8s_object.resource_type

        :param replicas: number of desired replicas
        """
        if self.get_k8s_object_field(
                k8s_object.read().status, "ready_replicas") == replicas:
            LOG.info("%s of type %s already has "
                     "required number of replicas",
                     *k8s_object_type)
            return

        LOG.info("Set %s of type %s replicas to %s",
                 *k8s_object_type,
                 replicas)

        k8s_object.patch_scale([
            {'op': 'replace', 'path': '/spec/replicas',
             'value': replicas}])
        self.wait_k8s_object_replicas_restored(k8s_object, k8s_object_type,
                                               replicas)

    def change_and_restore_replicas_count(self,
                                          k8s_object,
                                          replicas,
                                          deployment_replicas):
        """
        Change replicas of k8s object to some number
        and then restore to original count.

        :param k8s_object: k8s_object: k8s object that has read and
            patch_scale methods

        :param replicas: number of desired replicas

        :param deployment_replicas: original number of replicas that
            should be restored
        """
        k8s_object_type = (k8s_object.name, k8s_object.resource_type)
        LOG.info("Change %s of type %s replicas to %s", *k8s_object_type,
                 replicas)
        self.change_k8s_object_replicas_count(k8s_object, k8s_object_type,
                                              replicas)

        LOG.info("Restore %s of type %s replicas to %s", *k8s_object_type,
                 deployment_replicas)
        self.change_k8s_object_replicas_count(k8s_object, k8s_object_type,
                                              deployment_replicas)

    def get_pods_to_delete(self, k8s_object, num_pods_to_delete):
        """
        Get all pods with k8s object name as prefix, and then select
        n of them.

        :param k8s_object: k8s_object: k8s object that has read and
            patch_scale methods

        :param num_pods_to_delete: number of pods that should be deleted

        :return: list of randomly selected pods
        """
        LOG.info("Get all %s pods", k8s_object.name)
        pods = self.api.pods.list_starts_with(k8s_object.name,
                                              self.openstack_namespace)

        LOG.info("Select pods to delete from %s", pods)
        pods_to_delete = random.sample(pods, num_pods_to_delete)
        LOG.info("Selected pods are %s", pods_to_delete)
        return pods_to_delete

    def wait_pods_are_deleted_and_then_restored(self,
                                                k8s_object,
                                                k8s_object_type,
                                                deployment_replicas):
        """
        Wait till at least one of the  pod is deleted and then wait for
        replicas count restores to deployment quantity.

        :param k8s_object: k8s_object: k8s object that has read and
            patch_scale methods

        :param k8s_object_type: tuple of k8s_object.name
            and k8s_object.resource_type

        :param deployment_replicas: original number of replicas that
            should be restored
        """
        get_field = self.get_k8s_object_field

        def check():
            replicas = get_field(k8s_object.read().status, "ready_replicas")
            return replicas < deployment_replicas

        LOG.info("Wait for at least one pod is deleted")
        waiters.wait(check, interval=5, timeout=180)

        self.wait_k8s_object_replicas_restored(k8s_object, k8s_object_type,
                                               deployment_replicas)

    def delete_pod_and_wait_for_rescale(self,
                                        k8s_object,
                                        num_pods_to_delete,
                                        deployment_replicas):
        """
        Select pods from list of pods, delete them and then wait till
        k8s restore original quantity of replicas.

        :param k8s_object: k8s_object: k8s object that has read and
            patch_scale methods

        :param num_pods_to_delete: number of pods that should be deleted

        :param deployment_replicas: original number of replicas that
            should be restored
        """
        k8s_object_type = (k8s_object.name, k8s_object.resource_type)

        pods_to_delete = self.get_pods_to_delete(k8s_object,
                                                 num_pods_to_delete)

        for pod in pods_to_delete:
            LOG.info("Delete pod %s", pod)
            pod.delete(async_del=True)

        self.wait_pods_are_deleted_and_then_restored(k8s_object,
                                                     k8s_object_type,
                                                     deployment_replicas)

    def delete_force_pod_and_wait_for_rescale(self,
                                              k8s_object,
                                              num_pods_to_delete,
                                              deployment_replicas,
                                              bash_command):
        """
        Select pods from list of pods, delete them and then wait till
        k8s restore original quantity of replicas.

        :param k8s_object: k8s_object: k8s object that has read and
            patch_scale methods

        :param num_pods_to_delete: number of pods that should be deleted

        :param deployment_replicas: original number of replicas that
            should be restored

        :param bash_command: bash command that should be executed to
        delete pod
        """
        k8s_object_type = (k8s_object.name, k8s_object.resource_type)

        pods_to_delete = self.get_pods_to_delete(k8s_object,
                                                 num_pods_to_delete)

        for pod in pods_to_delete:
            LOG.info("Delete pod %s with command %s", pod, bash_command)
            resp = pod.exec(bash_command, tty=True)
            LOG.info("Response: %s", resp)

        self.wait_pods_are_deleted_and_then_restored(k8s_object,
                                                     k8s_object_type,
                                                     deployment_replicas)
