import pytest

import os
import yaml
import time
import base64
from packaging import version
from exec_helpers import SSHClient, SSHAuth
from si_tests import settings
from si_tests import logger
from si_tests.utils.gen_upgrade_paths import get_child_cluster_release_name
from si_tests.utils import waiters, utils, exceptions, templates

LOG = logger.logger


def check_networks(ns, seed_ip):
    # The Plan
    # 1. get all ipaddr objects in ns and extract IP addresses from them
    # 2. get all ipamhost objects in ns
    # 3. For each ipamhost:
    # 3.1 get PXE address and make ssh_client to it
    # 3.2 get corresponding l2template
    # 3.2.1 Get si/expected_ips from labels of the l2template
    # 3.2.2 If there is no expected_ips skip test for the ipamhost
    # 3.3 sort out IP addresses that belongs to the ipamhost from list
    #     of all IPs
    # 3.4 run ping for remaining IP addresses on the ipamhost
    ipaddrs = [i.read().to_dict()['status']['address']
               for i in ns.get_ipam_ipaddrs()]
    LOG.debug(f"Got IP addresses: {ipaddrs}")
    ipamhosts = [i.read().to_dict() for i in ns.get_ipam_hosts()]
    LOG.debug(f"Got IpamHosts: {ipamhosts}")

    seed_key_file = os.path.expanduser(settings.SEED_SSH_PRIV_KEY_FILE)
    seed_key = utils.load_keyfile(seed_key_file)
    seed_key_rsa = utils.get_rsa_key(seed_key["private"])
    remote_seed = SSHClient(
        host=seed_ip,
        port=22,
        # verbose=True,
        auth=SSHAuth(
            username=settings.SEED_SSH_LOGIN,
            key=seed_key_rsa))
    remote_seed.logger.addHandler(logger.console)
    LOG.debug(f"SSH client to seed node: {remote_seed}")
    child_key_file = os.path.expanduser(
        settings.KAAS_CHILD_CLUSTER_PRIVATE_KEY_FILE)
    child_key = utils.load_keyfile(child_key_file)
    child_key_rsa = utils.get_rsa_key(child_key["private"])

    failed_pings = {}
    for ipamhost in ipamhosts:
        ipamhost_name = ipamhost['metadata']['name']
        # Build ssh client to the ipamhost through its PXE address
        bastion_ip = ipamhost['status']['osMetadataNetwork'][
            'networks'][0]['ip_address']
        LOG.debug("Check networking on {0} ({1})".format(
            ipamhost_name, bastion_ip))
        # proxy_to() method available only in exec-helpers>=6.0.0
        bastion_kwargs = {
            'hostname': bastion_ip,
            # 'verbose': True,
            'auth': SSHAuth(
                username=settings.KAAS_CHILD_CLUSTER_SSH_LOGIN,
                key=child_key_rsa),
        }
        LOG.debug(f"SSH params for {ipamhost_name}: {bastion_kwargs}")
        # Get netconfig
        netconfig = ipamhost['status'].get('netconfigV2', {})
        # Get corresponding l2template
        l2template_ref = ipamhost['status'].get('l2TemplateRef', '')
        if not (netconfig and l2template_ref):
            LOG.info("No l2template applied to imaphost '%s'. "
                     "Skip extra network checking", ipamhost_name)
            continue
        l2template_name = l2template_ref.split('/')[1]
        l2template = ns.get_l2template(l2template_name).read().to_dict()
        LOG.debug("Got l2template for ipamhost '%s': %s",
                  ipamhost_name, l2template)
        expected_ips = int(l2template['metadata'].get('labels', {})
                           .get('si/expected_ips', 0))
        LOG.debug("Got expected_ips for ipamhost '%s': %s",
                  ipamhost_name, expected_ips)
        if not expected_ips:
            LOG.info("No 'si/expected_ips' in labels of l2template '%s'. "
                     "Skip extra network checking for ipamhost '%s'",
                     l2template_name, ipamhost_name)
            continue

        # Get ip addresses owned by the ipamhost
        own_ipaddrs = []
        for ifaces in netconfig.values():
            if isinstance(ifaces, dict):
                for iface in ifaces.values():
                    own_ipaddrs += [ip.split('/')[0]
                                    for ip in iface.get('addresses', [])
                                    if isinstance(iface, dict)]
            else:
                LOG.debug("Incorrect intarface object in netconfigV2: "
                          f"{ifaces}")
        LOG.debug("IP addresses owned by {0}: {1}".format(
            ipamhost_name, own_ipaddrs))
        # Ipamhost should have expected by l2template number of IPs
        assert len(own_ipaddrs) == expected_ips, \
            "Incorrect number of IP addresses ({0}) for '{2}' ipamhost. " \
            "Expecting to have {1} IPs".format(
                len(own_ipaddrs), expected_ips, ipamhost_name)
        # Ping all IP addresses except its own
        for ipaddr in set(ipaddrs).difference(own_ipaddrs):
            LOG.debug("Run ping on {0} to check availability "
                      "of {1}".format(ipamhost_name, ipaddr))
            ping_success = waiters.icmp_ping(
                host=ipaddr,
                ssh_client=remote_seed,
                bastion_kwargs=bastion_kwargs)
            if not ping_success:
                failed_pings.setdefault(bastion_ip, []).append(ipaddr)
            LOG.debug(f"Was ping of {ipaddr} from {ipamhost_name} "
                      f"({bastion_ip}) successfull: %s", ping_success)
    # All pings should be successfull
    LOG.info(f"Failed pings: {failed_pings}")
    return not failed_pings


def check_child_ceph_cluster(ch_cluster,
                             wait_status='Created',
                             check_health=False):
    """
    ch_cluster = child cluster object
    """
    try:
        # switch to child KUBECONFIG api
        rookceph = ch_cluster.k8sclient.rookcephclusters.get(
            name='rook-ceph', namespace='rook-ceph').data or {}
        rookcephstatus = rookceph.get('status', {}).get('state',
                                                        'NotExist')
        LOG.info("Rookceph status is {}".format(rookcephstatus))
        if wait_status != rookcephstatus:
            return False
        if check_health:
            rookceph_health = \
                rookceph.get('status', {}).get('ceph', {}).get('health',
                                                               'NotExist')
            rookceph_health_detail = rookceph.get(
                'status', {}).get('ceph', {}).get('details', '')
            LOG.info(f'Rookceph health is: {rookceph_health}\n'
                     f'Details: {rookceph_health_detail}')
            if rookceph_health not in ['HEALTH_OK', 'HEALTH_WARN']:
                return False
        return True
    except Exception:
        LOG.warning('rook-ceph cluster does not exist')
        return False


def get_ceph_nodes(data, ns, _bm_nodes):
    LOG.info("Getting CEPH nodes")
    c_nodes_data = {}
    if data.get('cephClusterMapping', False) != 'profiled':
        return c_nodes_data
    # Collect all machines, that supposed to be used in ceph
    for_storage_exp = ns._get_storage_machines()
    # Check, that we have mapping for those node
    for machine in for_storage_exp:
        hwid = machine['spec']['providerSpec']['value'].get(
            'hostSelector', {}).get('matchLabels', {}).get(
            'kaas.mirantis.com/baremetalhost-id', False)
        if not hwid:
            LOG.info(f"Machine {machine['metadata']['name']} "
                     f"not supposed "
                     f"to be used in ceph,since don't have"
                     f"'kaas.mirantis.com/baremetalhost-id' hostSelector,"
                     f"continue..")
            continue
        for _node in _bm_nodes:
            if _node.get('name') == hwid and _node.get(
                    'ceph_cluster_node'):
                LOG.info(f"Found custom ceph mapping:\n"
                         f"{_node['ceph_cluster_node']}\n"
                         f"for node: {_node['name']}")
                c_nodes_data[machine['metadata']['name']] = \
                    _node.get('ceph_cluster_node')
    LOG.info(f"Ceph node data for cluster:\n{c_nodes_data}")
    return c_nodes_data


@pytest.mark.usefixtures('log_method_time')
@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                         .format(settings.TARGET_CLUSTER)])
@pytest.mark.usefixtures("store_cluster_description")
def test_kaas_create_baremetal_child_lma_istio(kaas_manager, proxy_manager, _):
    """Deploy child cluster on top of management cluster."""
    # Allow test to re-use already defined `namespaces` with some
    # depended resources, instead failing.
    # Should be used only for debug purposes.
    reuse_resources = False

    # Collecting env data
    cluster_name = settings.TARGET_CLUSTER
    namespace_name = settings.TARGET_NAMESPACE
    region = settings.TARGET_REGION
    release_name = get_child_cluster_release_name(kaas_manager)
    LOG.info("Release name: {}".format(release_name))
    LOG.info("Available releases: {}".format(
        kaas_manager.get_clusterrelease_names()))
    assert release_name in kaas_manager.get_clusterrelease_names()
    mgmt_version = \
        kaas_manager.get_mgmt_cluster().spec['providerSpec']['value']['kaas'][
            'release']
    LOG.info(f"KaaS mgmt version is:{mgmt_version}")
    child_dns = settings.KAAS_CHILD_CLUSTER_DNS

    # Gather nodes information from yaml file
    child_data = yaml.safe_load(
        templates.render_template(settings.BM_CHILD_SETTINGS_YAML_PATH,
                                  options=dict(mgmt_version=mgmt_version)))[0]
    LOG.debug(f"Rendered child_data from file:\n"
              f"{settings.BM_CHILD_SETTINGS_YAML_PATH}:\n"
              f"{child_data}")
    bm_hosts_data = child_data['nodes']
    child_cluster_data = child_data['child_config']
    tf = child_cluster_data.get('tungstenfabric', {})

    # Creating namespace
    LOG.info("Namespace name - %s", namespace_name)
    namespaces = [n.name for n in kaas_manager.get_namespaces()]
    if namespace_name not in namespaces:
        ns = kaas_manager.create_namespace(namespace_name)
    elif namespace_name in namespaces and reuse_resources:
        ns = kaas_manager.get_namespace(namespace_name)
    else:
        raise Exception(f'Namespace {namespace_name} already exist!')

    ns.create_bm_statics(child_data)
    # Create ssh key
    public_key_name = "{cluster_name}-key".format(
        cluster_name=cluster_name)
    if public_key_name not in [k.name for k in ns.get_publickeys()]:
        LOG.info("Public key name - {public_key_name}".format(
            public_key_name=public_key_name))
        with open(settings.KAAS_CHILD_CLUSTER_PUBLIC_KEY_FILE) as pub_key:
            pub_key_content = pub_key.read()
        ns.create_publickey(public_key_name, pub_key_content)
    LOG.info('Creating secret for each BMH')
    bm_nodes = []
    bm_ctrls = []
    existing_secrets = [sec.name for sec in ns.get_secrets()]
    for node in bm_hosts_data:
        if not any(item in node.get('si_roles', []) for item in
                   ['nodeforscale', 'child-scale-controller']):
            # Dirty hack, for not copy-paste bmh_name|cred_name across whole
            # test.
            bmh_name = utils.render_bmh_name(node['name'],
                                             cluster_name,
                                             node.get('bootUEFI', True),
                                             node['bmh_labels'],
                                             si_roles=node.get('si_roles',
                                                               False))
            cred_name = node['name'] + '-cred'
            _node = node.copy()
            _node.update({'bmh_name': bmh_name,
                          'cred_name': cred_name})
            bm_nodes.append(_node)
            secret_data = {
                "username": node['ipmi']['username'],
                "password": node['ipmi']['password']
            }
            if cred_name not in existing_secrets:
                ns.create_secret(name=cred_name, data=secret_data,
                                 region=region, provider="baremetal")
            else:
                LOG.warning(f'Secret: {cred_name} already exist, skipping')

    existing_bmh = [bmh.name for bmh in ns.get_baremetalhosts()]
    osdpl_node_overrides = dict()
    for node in bm_nodes:
        bmh_name = node['bmh_name']
        if bmh_name not in existing_bmh:
            if node['bmh_labels'].get(
                    "hostlabel.bm.kaas.mirantis.com/controlplane"):
                bm_ctrls.append(bmh_name)
            node_override = node.get('osdpl_node_override')
            if node_override:
                osdpl_node_overrides[node['name']] = node_override
            ns.create_baremetalhost(bmh_name=bmh_name,
                                    bmh_secret=node['cred_name'],
                                    bmh_mac=node['networks'][0]['mac'],
                                    bmh_ipmi=node['ipmi']['ipmi_ip'],
                                    hardwareProfile=node.get('hardwareProfile',
                                                             False),
                                    labels=node['bmh_labels'],
                                    bootUEFI=node.get('bootUEFI', True))
        else:
            LOG.warning(f'BaremetalHost: {bmh_name} already exist,'
                        f' skipping')

        # dnsmasq bug PRODX-7751; epic with workaround PRODX-8106
        # dnsmasq DHCP provides IPs one by one with ping-check
        # for each IP for about 12sec.
        time.sleep(15)

    LOG.info('Wait baremetal hosts commissioning')
    # Commissioning timeout has been increased, since after 2.6 release
    # bm-operator limited to PROVISIONING_LIMIT=6, so only 6 nodes operated
    # at once. Same for provision stage. Otherwise, we will monitor only
    # control nodes - cluster provision can be started when we have
    # control nodes only.
    if not settings.ASYNC_BMH_CONTROL_NODES_WAIT:
        bm_ctrls = None
    else:
        LOG.warning(f'Bmh provision wait scope limited to:{bm_ctrls}')
    ns.wait_baremetalhosts_statuses(
        wait_status='ready', retries=90, interval=60, nodes=bm_ctrls)

    LOG.info('Create cluster')
    helm_releases = []
    for hrelease in settings.KAAS_CHILD_CLUSTER_EXTRA_HELM_RELEASES:
        helm_releases.append({"name": hrelease.strip()})
    LOG.info('helm_releases: {0}'.format(helm_releases))
    LOG.info('lma_enabled: {0}'.format(settings.KAAS_CHILD_CLUSTER_DEPLOY_LMA))

    if ((settings.KAAS_EXTERNAL_PROXY_ACCESS_STR or
         settings.KAAS_INSTALL_FAKE_PROXY) and
            not settings.KAAS_OFFLINE_DEPLOYMENT):
        LOG.error("Proxy variables is set but KAAS_OFFLINE_DEPLOYMENT "
                  "is False.Using Proxy on online deployments makes no sense.")
        raise RuntimeError("Proxy variables is set but KAAS_OFFLINE_DEPLOYMENT"
                           " is False")

    if settings.KAAS_EXTERNAL_PROXY_ACCESS_STR:
        LOG.info("Found KAAS_EXTERNAL_PROXY_ACCESS_STR. Creating "
                 "proxyobject with proxy {0}".
                 format(settings.KAAS_EXTERNAL_PROXY_ACCESS_STR))
        proxy_access_str = settings.KAAS_EXTERNAL_PROXY_ACCESS_STR
        proxy_object = ns.create_proxyobject(proxy_str=proxy_access_str)
        proxy_object_name = proxy_object.data['metadata']['name']
    elif settings.KAAS_INSTALL_FAKE_PROXY:
        LOG.info("Variable KAAS_EXTERNAL_PROXY_ACCESS_STR is not set but"
                 "KAAS_INSTALL_FAKE_PROXY is enabled. Fake squid proxy will "
                 "be installed on management cluster to use it in child.")
        proxy_manager.deploy_proxy_sts()
        proxy_manager.deploy_proxy_svc()
        proxy_access_str = proxy_manager.generate_proxy_string()
        proxy_manager.check_proxy_conn()
        proxy_object = ns.create_proxyobject(proxy_str=proxy_access_str)
        proxy_object_name = proxy_object.data['metadata']['name']
    else:
        proxy_object_name = None

    existing_clusters = [_cluster.name for _cluster in ns.get_clusters()]
    if cluster_name not in existing_clusters:
        cluster = ns.create_cluster(
            cluster_name,
            release_name,
            region=region,
            provider="baremetal",
            loadbalancer_host=child_cluster_data['loadbalancer_host'],
            metallb_ip_range=child_cluster_data.get('metallb_ip_range', None),
            services_cidr="10.232.0.0/18",
            pods_cidr="192.168.0.0/16",
            nodes_cidr="10.12.10.0/24",
            dns=child_dns,
            extra_helm_releases=helm_releases,
            lma_enabled=settings.KAAS_CHILD_CLUSTER_DEPLOY_LMA,
            dedicated_controlplane=child_cluster_data.get(
                'dedicated_controlplane', True),
            public_key_name=public_key_name,
            proxy_name=proxy_object_name)
    else:
        cluster = ns.get_cluster(cluster_name)
        LOG.warning(f'Re-use existing cluster: {cluster_name}')

    for node in bm_nodes:
        LOG.info('Create BareMetal machine')
        custom_bmhp = False
        if node.get('bmh_profile'):
            custom_bmhp = {
                'namespace': namespace_name,
                'name': node['bmh_profile']
            }

        cluster.create_baremetal_machine(
            genname=node['bmh_name'],
            node_pubkey_name=public_key_name,
            matchlabels={'kaas.mirantis.com/'
                         'baremetalhost-id':
                             node['bmh_labels']
                             ['kaas.mirantis.com/baremetalhost-id']},
            baremetalhostprofile=custom_bmhp,
            l2TemplateSelector=node.get('l2TemplateSelector', dict()),
            labels=node['machine_labels'],
            node_labels=node.get('node_labels', {})
        )
        # dnsmasq bug PRODX-7751; epic with workaround PRODX-8106
        # dnsmasq DHCP provides IPs one by one with ping-check
        # for each IP for about 12sec.
        time.sleep(15)

    LOG.info('Wait baremetal hosts to be provisioned')
    ns.wait_baremetalhosts_statuses(
        wait_status='provisioned', retries=240, interval=60, nodes=bm_ctrls)

    # Extend a secret which contains child's kubeconfig with some fields needed
    # for testing BM on virtual envs. Kubeconfig itself does not modified here
    if 'external_ip' in child_cluster_data:
        kubeconfig_secret_name = "{cluster_name}-kubeconfig".format(
            cluster_name=cluster_name)
        kubeconfig_secret = ns.get_secret(kubeconfig_secret_name)
        external_ip = child_cluster_data['external_ip'].encode('utf-8')
        kubeconfig_secret.patch(
            {'data': {
                'external_ip': str(base64.b64encode(external_ip), 'utf-8'),
                'insecure': str(base64.b64encode(b'true'), 'utf-8')
            }})

    LOG.info('Waiting for machines are Ready and helmbundles are deployed')
    try:
        cluster.check.check_machines_status()
        cluster.check.check_cluster_nodes()
        cluster.check.check_helmbundles()
        cluster.check.check_k8s_nodes()
    finally:
        cluster.store_machines_artifacts()
    # This WA to set labels for OpenStack nodes. Once
    # KaaS can set labels this WA should be removed.
    cluster.apply_hotfix_labels(bm_nodes=bm_nodes)

    # Apply WA (PRODX-7905) for TF Contrail
    # Change /etc/hosts (delete hostname record without suffix):
    # https://docs.mirantis.com/mosk/master/mosk-deployment-guide/configure-host-os/configure-hosts-file-tf.html
    # TODO(alexz) most probably - dead code due get_machines().
    # probably here should be get_nodes()
    machines = ns.get_cluster(cluster_name).get_machines()
    if tf:
        LOG.info("Tungsten Fabric Steps!")
        fake_public_cidr = tf.get('fake_public_cidr')
        vsrx_ip_in_public = tf.get('vsrx_ip_in_public')

        for machine in machines:
            # routes for fake vsrx public network (PRODX-9663)
            machine.run_cmd(f"sudo ip route add {fake_public_cidr} "
                            f"via {vsrx_ip_in_public}|true")

    if settings.OPENSTACK_DEPLOY_DPDK:
        LOG.info("Install DPDK driver on compute")
        lkey, lvalue = settings.OPENSTACK_NODES_WITH_HUGEPAGES_LABEL.split('=')
        for machine in machines:
            node_labels = machine.get_k8s_node().read().metadata.labels
            for k, v in node_labels.items():
                if k == lkey and v == lvalue:
                    machine.run_cmd(
                        "sudo apt install -y "
                        f"{settings.OPENSTACK_DPDK_DRIVER_PACKAGE}",
                        timeout=600)
                    break

    # Check that networks have been correctly configured on machines so they
    # can reach each other over all configured network interfaces
    if 'seed_ip' in child_cluster_data:
        LOG.info('Check network connectivity')
        try:
            waiters.wait(
                check_networks,
                predicate_args=[ns, child_cluster_data['seed_ip']],
                interval=5, timeout=60 * 30)
        except exceptions.TimeoutError as e:
            kaas_manager.print_events(
                kaas_manager.get_events(namespace=cluster.namespace))
            LOG.error("Tired waiting for network connectivity check")
            raise e
        LOG.info('Network connectivity between all machines is OK')
    else:
        LOG.error(
            "Cannot verify network connectivity between child "
            "nodes without seed IP. "
            "Check that 'seed_ip' field is set in child_config.")

    if 'ceph-controller' in [r['name'] for r in helm_releases]:
        nodes_data = get_ceph_nodes(child_cluster_data, ns, bm_nodes)
        ceph_cluster_pools = child_cluster_data.get('ceph_pools_config')

        ceph_rgw_config = None
        if settings.CEPH_DEPLOY_RGW:
            ceph_rgw_config = child_cluster_data.get('ceph_rgw_config')
            LOG.info(f"Adding Ceph rgw metadata {ceph_rgw_config}")

        LOG.info('Create Ceph cluster')
        ns.create_ceph_cluster(
            name=settings.CEPH_CLUSTER_NAME,
            cluster_name=cluster_name,
            hdd_type=child_cluster_data.get('cephClusterDeviceClass', 'hdd'),
            nodes_data=nodes_data,
            pools=ceph_cluster_pools,
            rgw=ceph_rgw_config,
            network=child_cluster_data.get('ceph_network_config', None))
        try:
            waiters.wait(lambda: check_child_ceph_cluster(ch_cluster=cluster,
                                                          check_health=True),
                         timeout=60 * 60, interval=30)
        except exceptions.TimeoutError as e:
            LOG.error("Tired waiting for ceph cluster correct status,"
                      "events from management cluster:")
            kaas_manager.print_events(
                kaas_manager.get_events(namespace=cluster.namespace))
            # TODO(alexz) extend cluster.k8sclient:
            # LOG.error(f'Tired waiting for ceph cluster correct status,'
            #           'events from child cluster {cluster.name} cluster:')
            # cluster.k8sclient.print_events(
            #     cluster.k8sclient.get_events(namespace=cluster.namespace))
            raise e
    if version.parse(mgmt_version) > version.parse('kaas-2-4-0') and tf:
        LOG.info("Enable TF monitoring ")
        child_pr_spec = cluster.data['spec']['providerSpec']
        LOG.info("Current stacklight spec: {}".format(child_pr_spec))

        body = {
            "spec": {
                "providerSpec": {
                    "value": {
                        "helmReleases": cluster.data.get(
                            "spec").get("providerSpec").get("value").get(
                            "helmReleases")
                    }
                }
            }
        }
        lma_body = [x for x in
                    body['spec']['providerSpec']['value']['helmReleases']
                    if x['name'] == "stacklight"][0]
        lma_body['values'].update(
            {'tungstenFabricMonitoring': {'enabled': True}})
        kaas_manager.api.kaas_clusters.update(
            name=cluster_name, namespace=namespace_name, body=body)

        LOG.info("Cluster providerSpec after update {}".format(
            cluster.data['spec']['providerSpec']))

    # openstack-controller requires ceph cluster
    cluster.add_osdpl_overrides(osdpl_node_overrides)

    # Collecting artifacts
    cluster.store_k8s_artifacts()

    # Check helmbundles after ceph cluster deployed (e.g. Stacklight)
    cluster.check.check_helmbundles()

    if not cluster.openstack_lma_enabled():
        cluster.check.check_cluster_readiness(timeout=3000, interval=60)
    # Check/wait for correct docker service replicas in cluster
    cluster.check.check_actual_expected_docker_services()
    # Check/wait for correct pods statuses in child cluster
    cluster.check.check_k8s_pods()
    cluster.check.check_actual_expected_pods(timeout=2400, interval=60)
