import pytest
import yaml
import socket

from kubernetes.client.rest import ApiException

from retry import retry

from si_tests.deployments.utils.file_utils import save_to_yaml
from si_tests.utils import waiters
from si_tests.utils import exceptions
from si_tests import settings
from si_tests import logger

LOG = logger.logger


@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.TARGET_CLUSTER)])
@pytest.mark.usefixtures("store_cluster_description")
def test_delete_child_cluster(kaas_manager, _):
    """Delete child cluster"""

    ns = kaas_manager.get_namespace(settings.TARGET_NAMESPACE)
    child_cluster = ns.get_cluster(settings.TARGET_CLUSTER)
    child_cluster_provider = child_cluster.provider
    child_cluster_name = child_cluster.name
    if settings.CHECK_RESOURCES_DELETION:
        child_cluster_resources = child_cluster.provider_resources.resources

    if child_cluster_provider == 'VsphereClusterProviderSpec':
        LOG.info("Check that rhel subscription doesn't stuck lcmmachines")
        LOG.info(
            "Traffic to RHEL subsription portal should be block before delete")

        @retry(Exception, delay=10, tries=3, logger=LOG)
        def get_ip_to_block():
            LOG.info("Trying to get ip of subscription.rhn.redhat.com")
            ip = socket.gethostbyname('subscription.rhn.redhat.com')
            LOG.info(f"subscription.rhn.redhat.com has {ip}")
            return ip
        block_ip = get_ip_to_block()

        default_ns = kaas_manager.get_namespace(settings.CLUSTER_NAMESPACE)
        mgm_cluster = default_ns.get_cluster(settings.CLUSTER_NAME)
        mgm_machies = mgm_cluster.get_machines()
        for m in mgm_machies:
            LOG.info(
                "Going to block RHEL subscription server via "
                f"iptables on {m.name}")
            drop_cmd = \
                f"sudo iptables -I OUTPUT -p tcp --dst {block_ip} -j DROP"
            m.run_cmd(drop_cmd)

    LOG.info("Deleting child cluster - {0}".format(child_cluster_name))
    child_cluster.delete()

    if child_cluster_provider == 'VsphereClusterProviderSpec':

        def check_delete_lcmmachines():
            try:
                if len(child_cluster.get_lcmmachines()) == 0:
                    return True
            except ApiException as ex:
                if ex.status == '404' and \
                        ex.reason == "Not Found" and \
                        ex.body.get('Reason') == "NotFound" and \
                        ex.body.get('Code') == "404" and \
                        ex.body.get(
                            'details', {}).get(
                                "kind") == "clusters" and \
                        ex.body.get(
                            'details', {}).get(
                                "name") == settings.CLUSTER_NAME:
                    return True
            return False

        LOG.info("Wait when lcmmachines will be deleted")
        waiters.wait(
            check_delete_lcmmachines,
            timeout=1200,
            interval=60)

        for m in mgm_machies:
            LOG.info(
                "Going to unblock RHEL subscription server via "
                f"iptables on {m.name}")
            undrop_cmd = \
                f"sudo iptables -D OUTPUT -p tcp --dst {block_ip} -j DROP"
            m.run_cmd(undrop_cmd)

    LOG.info("Waiting for cluster deletion")
    child_cluster.check.check_cluster_deleted()

    if child_cluster_provider == 'BaremetalClusterProviderSpec':
        LOG.info('Check that BMHs are in ready state')
        ns.wait_baremetalhosts_statuses(
            wait_status='ready', retries=60, interval=30)

    if settings.CHILD_NAMESPACE_FORCE_CLEANUP:
        LOG.info("Cleanup child namespace")
        if child_cluster_provider == 'BaremetalClusterProviderSpec' and \
                settings.KAAS_BM_WA_PRODX_6630:
            LOG.warning("Removing bmh before ns")
            bmhs_for_delete = []
            for bmh in ns.get_baremetalhosts():
                bmhs_for_delete.append(bmh.name)
                bmh.delete()
            for bmh in bmhs_for_delete:
                ns.wait_baremetalhost_deletion(bmh)
        ns.delete()
        LOG.info("Waiting for child namespace deletion")
        ns.wait_for_deletion()

    if settings.CHECK_RESOURCES_DELETION:
        LOG.info("Will check resources deletion for {} provider".format(
            child_cluster_provider))
        try:
            waiters.wait(
                lambda:
                not child_cluster.provider_resources.check_resources_deletion(
                    child_cluster_resources), timeout=2400, interval=300)
            save_to_yaml(
                    child_cluster.provider_resources.not_deleted_resources,
                    settings.ARTIFACTS_DIR +
                    "/not_deleted_resources_{}.yaml".format(
                        child_cluster_name))
        except exceptions.TimeoutError:
            save_to_yaml(
                    child_cluster.provider_resources.not_deleted_resources,
                    settings.ARTIFACTS_DIR +
                    "/not_deleted_resources_{}.yaml".format(
                        child_cluster_name))
            raise Exception(
                    "Some resources were not deleted: {}".format(
                        yaml.dump(
                            child_cluster.provider_resources.not_deleted_resources))) # noqa
