import pytest
import yaml

from si_tests import settings
from si_tests import logger
from si_tests.utils import waiters
from kubernetes.client.rest import ApiException

LOG = logger.logger

cluster_name = settings.TARGET_CLUSTER
namespace_name = settings.TARGET_NAMESPACE
rook_ns = settings.ROOK_CEPH_NS


def get_hyperconverge_params():
    with open(settings.CEPH_HC_DATA_PATH, 'r') as data:
        hc = yaml.safe_load(data)
    tolerations = hc.get('hyperconverge', {}).get('tolerations')
    resources = hc.get('hyperconverge', {}).get('resources')
    return tolerations, resources


@pytest.mark.usefixtures('log_method_time')
def test_ceph_hyperconverge_tolerations(kaas_manager, show_step):
    """Test ceph pods taints and tolerations.

    Scenario:
        1. Gather current ceph data
        2. Check ceph health is OK before proceed
        3. Patch cluster with new data for tolerations (mon and mgr pods)
        4. Patch kaascephcluster with mon roles on control nodes
        5. Check pods are spawned,
           placed on corresponding nodes and health is OK
    """

    # Get namespace
    LOG.info(f"Namespace name: {namespace_name}")
    ns = kaas_manager.get_namespace(namespace_name)
    cluster = ns.get_cluster(cluster_name)
    k8sclient = cluster.k8sclient
    mgmt_cluster = kaas_manager.get_mgmt_cluster()
    ceph_crd = mgmt_cluster.get_ceph_data(
        namespace=settings.TARGET_NAMESPACE)
    tolerations, _ = get_hyperconverge_params()
    ceph_health_timeout = 900

    # We will add 2 additional monitors. So we need 2 nodes
    control_machines = cluster.get_machines(machine_type='control')[0:2]

    # Third control node will be used for mgr pod migration and check
    control_machine_for_mgr = cluster.get_machines(machine_type='control')[-1]
    machines_k8s_nodes_names = []
    for machine in control_machines:
        machines_k8s_nodes_names.append(machine.get_k8s_node().name)

    mon_deploys = [
        deploy for deploy in k8sclient.deployments.list(
            namespace=rook_ns) if 'rook-ceph-mon' in deploy.name]
    mon_deploys_names = [deploy.name for deploy in mon_deploys]
    mgr_pod = [
        pod for pod in k8sclient.pods.list(
            namespace=rook_ns) if 'rook-ceph-mgr' in pod.name][0]
    mgr_pod_name = mgr_pod.name
    show_step(1)
    cl_data = cluster.data
    helm_r = cl_data.get('spec').get('providerSpec').get('value', {}).get(
        'helmReleases', [])

    # Add tolerations for mon pods
    for r in helm_r:
        if r['name'] == 'ceph-controller':
            r['values'] = {'hyperconverge': {'tolerations': tolerations}}

    nodes_data = ceph_crd.data.get('spec').get('cephClusterSpec').get('nodes')

    # Will add 2 new nodes (control) to ceph cluster with mon role
    for machine in control_machines:
        nodes_data[machine.name] = {'roles': ['mon']}
    # Add new node with mgr role and remove role from other nodes
    for k, v in nodes_data.items():
        if 'mgr' in v.get('roles', []):
            v['roles'].remove('mgr')
    nodes_data[control_machine_for_mgr.name] = {'roles': ['mgr']}

    show_step(2)
    # Check ceph health
    health_info = cluster.check.get_ceph_health_status()
    assert health_info == "HEALTH_OK", (
        "Health is not OK. Will not proceed."
        "Current ceph health info: {}".format(health_info))

    patch_data = {'spec': {'cephClusterSpec': {'nodes': nodes_data}}}
    show_step(3)
    # Patch cluster with tolerations
    cluster.patch({
        'spec': {'providerSpec': {'value': {'helmReleases': helm_r}}}})

    show_step(4)
    # Patch kaascephcluster with new nodes roles
    mgmt_cluster.patch_ceph_data(data=patch_data, crd=ceph_crd)

    show_step(5)
    # Wait for mon pods number is increased by 2
    LOG.info("Waiting for mons quorum")
    waiters.wait(lambda: len([deploy for deploy in k8sclient.deployments.list(
        namespace=rook_ns) if 'rook-ceph-mon' in deploy.name
                              and 'canary' not in deploy.name]) == (
            len(mon_deploys) + 2), timeout=ceph_health_timeout, interval=30)
    LOG.info('Mons quorum found')
    cluster.check.check_k8s_pods(target_namespaces=rook_ns)
    # Mgr pod is required to be deleted for spawning on new node
    # https://mirantis.jira.com/browse/PRODX-12807
    # https://gerrit.mcp.mirantis.com/c/kaas/kaas-docs/+/95733/2/doc/release-notes/source/known-bugs/12807.rst # noqa
    # However in case when operator migrates other pods (e.g. mons),
    # it can be restarted too. So, we will try and except for it has been
    # already restarted
    try:
        LOG.info("Attempt to delete mgr pod")
        mgr_pod.delete()
        LOG.info("Waiting for new mgr pod is spawned")
        waiters.wait(lambda: len([pod for pod in k8sclient.pods.list(
            namespace=rook_ns) if 'rook-ceph-mgr' in pod.name and
                                 pod.name != mgr_pod_name]) > 0, timeout=300,
                    interval=30)
    except ApiException:
        LOG.info("Pod has already been restarted")

    new_mgr_pod = [
        pod for pod in k8sclient.pods.list(
            namespace=rook_ns) if 'rook-ceph-mgr' in pod.name][0]
    LOG.info("New mgr pod is {}".format(new_mgr_pod.name))
    # Check mgr pod is spawned on correct node
    assert new_mgr_pod.data['spec']['node_name'] == \
        control_machine_for_mgr.get_k8s_node().name, (
        "New mgr pod is not spawned on correct node.")
    LOG.info("New manager pod {} is spawned on node {}".format(
        new_mgr_pod.name, control_machine_for_mgr.get_k8s_node().name))
    # Wait for pods up
    cluster.check.check_k8s_pods(target_namespaces=rook_ns)

    # Check ceph health
    LOG.info("Waiting for HEALTH is OK")
    cluster.check.wait_ceph_health_status()
    new_mon_deploys = [deploy for deploy in k8sclient.deployments.list(
        namespace=rook_ns) if 'rook-ceph-mon' in deploy.name
                       and deploy.name not in mon_deploys_names]
    LOG.info("New mon deploys: {}".format([d.name for d in new_mon_deploys]))
    new_mon_pods = []
    for deploy in new_mon_deploys:
        for pod in [pod for pod in k8sclient.pods.list(namespace=rook_ns) if
                    'rook-ceph-mon' in pod.name]:
            if deploy.name in pod.name:
                new_mon_pods.append(pod)
    LOG.info("New mon pods: {}".format([pod.name for pod in new_mon_pods]))

    # Check that new mons are spawned on control nodes which were added to
    # cluster by tolerations

    LOG.info("Check pods are spawned on corresponding nodes")
    for pod in new_mon_pods:
        node = pod.data['spec']['node_name']
        if node in machines_k8s_nodes_names:
            machines_k8s_nodes_names.remove(node)
            LOG.info("Pod {} spawned on node {}".format(pod.name, node))
        else:
            LOG.warning(
                "New mon pod {} is spawned on node {}. "
                "This node was not added to tolerations!".format(
                    pod.name, node))
    assert len(machines_k8s_nodes_names) == 0, (
        "No mon pods are spawned on next nodes: {}".format(
            machines_k8s_nodes_names))


@pytest.mark.usefixtures('log_method_time')
def test_ceph_hyperconverge_resourcess(kaas_manager, show_step):
    """Test ceph pods resources assignment.
    Scenario:
        1. Gather current ceph data
        2. Check ceph health is OK before proceed
        3. Patch cluster with new data with pods resources
        4. Check pods are respawned,
           resources are applied and health is OK
    """
    ns = kaas_manager.get_namespace(namespace_name)
    cluster = ns.get_cluster(cluster_name)
    k8sclient = cluster.k8sclient
    _, resources = get_hyperconverge_params()
    show_step(1)
    cl_data = cluster.data
    helm_r = cl_data.get('spec').get('providerSpec').get('value', {}).get(
        'helmReleases', [])
    for r in helm_r:
        if r['name'] == 'ceph-controller':
            if r.get('values', {}) and 'hyperconverge' in r.get('values', {}):
                r['values']['hyperconverge']['resources'] = resources
            else:
                r['values'] = {'hyperconverge': {'resources': resources}}

    show_step(2)
    # Check ceph health
    health_info = cluster.check.get_ceph_health_status()
    assert health_info == "HEALTH_OK", (
        "Health is not OK. Will not proceed."
        "Current ceph health status: {}".format(health_info))
    mon_deploys = [
        d for d in k8sclient.deployments.list(
            namespace=rook_ns) if 'rook-ceph-mon' in d.name]
    mgr_deploys = [
        d for d in k8sclient.deployments.list(
            namespace=rook_ns) if 'rook-ceph-mgr' in d.name]
    osd_deploys = [
        d for d in k8sclient.deployments.list(
            namespace=rook_ns) if 'rook-ceph-osd' in d.name]
    all_deploys = mon_deploys + mgr_deploys + osd_deploys

    show_step(3)
    cluster.check.store_deploys_generation(deploys=all_deploys)
    cluster.patch({
        'spec': {'providerSpec': {'value': {'helmReleases': helm_r}}}})

    show_step(4)
    cluster.check.wait_deploy_generation_changed(deploys=all_deploys,
                                                 timeout=900)
    # cluster.check.check_actual_expected_pods()
    cluster.check.check_k8s_pods(target_namespaces=rook_ns)

    # Final check ceph health status
    cluster.check.wait_ceph_health_status()
    # Check that all deploys are succesfully configured with resources
    current_limits = {}
    not_configured_deploys = {}
    for deploy in all_deploys:
        for container in deploy.data['spec']['template']['spec']['containers']:
            current_limits[deploy.name] = {
                container['name']: container['resources']}

    for deploy, containers in current_limits.items():
        for container, resource in containers.items():
            if resource['limits'] and resource['requests']:
                continue
            else:
                not_configured_deploys[deploy] = {
                    'containers': {container: resource}}

    assert not not_configured_deploys, (
        "Next deploys were failed to be configured with resources limits and "
        "requests: \n{}".format(yaml.dump(not_configured_deploys)))
    LOG.info("All deploys were succesfully configured with resources "
             "requests and limits: \n{}".format(yaml.dump(current_limits)))
