import pytest
import yaml
import time

from si_tests import settings
from si_tests import logger
from si_tests.utils import waiters, templates, utils, exceptions
from si_tests.managers import openstack_manager

LOG = logger.logger

cluster_name = settings.TARGET_CLUSTER
namespace_name = settings.TARGET_NAMESPACE
region = settings.TARGET_REGION
osdpl_name = settings.OSH_DEPLOYMENT_NAME


@pytest.mark.usefixtures('log_method_time')
def test_replace_failed_os_control(kaas_manager, show_step):
    """Replace OpenStack control plane node after the outage.

    Scenario:
        1. Simulate power outage (Hard power off)
        2. Wait for NotReady node state
        3. Get info about nodeforscale node
        4. Create bmh
        5. Add labels
        6. Create BareMetal machine
        7. Recreate octavia resources and check
        8. Identity and rotate stateful applications
        9. Force delete all pods from replaced node
    """

    # Get namespace
    LOG.info(f"Namespace name: {namespace_name}")
    ns = kaas_manager.get_namespace(namespace_name)
    cluster = ns.get_cluster(cluster_name)

    # Get public key
    LOG.info(f"Public key name {cluster_name}-key")
    public_key_name = f"{cluster_name}-key"
    ns.get_publickey(public_key_name)

    child_data = yaml.safe_load(templates.render_template(
        settings.BM_CHILD_SETTINGS_YAML_PATH))[0]
    bm_hosts_data = child_data['nodes']
    show_step(3)

    # get machine data from .yaml
    replacement_node = []
    for node in bm_hosts_data:
        if 'replaceoscontrol' in node.get('si_roles', []):
            replacement_node.append(node)
            break
    assert len(replacement_node) == 1, "BMH with si_roles " \
                                       "'replaceoscontrol'" \
                                       " not found in file!"
    replacement_node = replacement_node[0]
    print(f"Node for replacement: {replacement_node['name']}")

    workers = cluster.get_machines(machine_type='worker',
                                   machine_status='Ready')
    # get cluster machine (outage candidate)
    os_control = []
    for w in workers:
        if [n for n in w.nodeLabels if 'openstack-control-plane'
                                       in n.values()]:
            os_control.append(w)

    show_step(4)
    target_node = [n for n in os_control
                   if replacement_node['name'] in n.name]
    assert len(target_node) == 1, "BM Machine not found on the cluster, " \
                                  "or Machine not in Ready status"
    target_node = target_node[0]

    # Get bmh and power off
    # Wait form k8s node status NotReady
    # Remove all labels from k8s node
    ns.set_baremetalhost_power(ns=namespace_name,
                               bmh_name=target_node.get_bmh_name(),
                               online=False)
    LOG.info("Get k8s node!!!")
    k8s_node = target_node.get_k8s_node()
    LOG.info("Delete, and start waiting k8s node status NotReady!!!")
    old_node_k8s_name = k8s_node.name
    cluster.check.wait_k8s_node_status(old_node_k8s_name,
                                       expected_status='NotReady')
    cluster.remove_k8s_node_all_labels(target_node.name)
    target_node.delete()
    cluster.wait_machine_deletion(target_node.name,
                                  retries=60, interval=30)

    # Create credentials, bmh, machine
    cred_name = replacement_node['name'] + '-cred'
    bmh_name = utils.render_bmh_name(replacement_node['name'], cluster_name,
                                     replacement_node.get('bootUEFI', True),
                                     replacement_node['bmh_labels'],
                                     si_roles=node.get('si_roles', False))
    replacement_node.update({'bmh_name': bmh_name, 'cred_name': cred_name})
    if not kaas_manager.api.secrets.present(name=cred_name,
                                            namespace=namespace_name):
        secret_data = {
            "username": replacement_node.get('ipmi').get('username'),
            "password": replacement_node.get('ipmi').get('password')
        }
        ns.create_secret(name=cred_name, data=secret_data,
                         region=region, provider="baremetal")

    osdpl_node_overrides = dict()
    node_override = replacement_node.get('osdpl_node_override', {})
    if node_override:
        osdpl_node_overrides[replacement_node['name']] = node_override

    LOG.info(f"Wait for bmh '{replacement_node['bmh_name'], bmh_name}' Ready")
    ns.wait_baremetalhosts_statuses(nodes=replacement_node['bmh_name'],
                                    wait_status='ready',
                                    retries=20,
                                    interval=60)

    labels = replacement_node.get('node_labels', {})

    show_step(6)
    l2TemplateSelector = replacement_node.get('l2TemplateSelector', dict())

    custom_bmhp = False
    if replacement_node.get('bmh_profile'):
        custom_bmhp = {
            'namespace': namespace_name,
            'name': replacement_node['bmh_profile']
        }
    replacement_node.update({'bmh_name': bmh_name+"-replaced"})
    cluster.create_baremetal_machine(
        genname=replacement_node['bmh_name'],
        node_pubkey_name=public_key_name,
        matchlabels={'kaas.mirantis.com/'
                     'baremetalhost-id':
                         replacement_node['bmh_labels']
                         ['kaas.mirantis.com/baremetalhost-id']},
        baremetalhostprofile=custom_bmhp,
        l2TemplateSelector=l2TemplateSelector,
        labels=replacement_node['machine_labels'],
        node_labels=labels
    )

    # Wait Baremetal hosts be provisioned
    ns.wait_baremetalhosts_statuses(nodes=replacement_node['bmh_name'],
                                    wait_status='provisioned',
                                    retries=30,
                                    interval=90)

    # Waiting for cluster,pods,hb are Ready after machines were added
    cluster.check.check_machines_status()
    cluster.check.check_cluster_nodes()
    cluster.check.check_k8s_nodes()
    cluster.check.check_helmbundles()

    new_k8s_node = [n for n in cluster.get_machines()
                    if replacement_node['name'] in n.name][0]
    new_node_k8s_name = new_k8s_node.get_k8s_node().name
    LOG.info(f"New node k8s name: {new_node_k8s_name}")

    show_step(7)
    LOG.info("Rerun the octavia-create-resources job")
    client_pod = cluster.k8sclient.pods.list(
        namespace="osh-system",
        name_prefix='openstack-controller')
    assert len(client_pod) > 0, ("No pods found with prefix "
                                 "openstack-controller in namespace "
                                 "osh-systme")
    client_pod = client_pod[0]
    cmd_octavia = ['/bin/sh', '-c',
                   'osctl-job-rerun octavia-create-resources openstack']
    resutl_cmd_octavia = client_pod.exec(cmd_octavia, container='osdpl')
    LOG.info(f"Octavia rerun job result: {resutl_cmd_octavia}")

    LOG.info("Wait for octavia-health-manager pod started on new node")

    def wait_octavia_pod(k8s_node_name=''):
        octavia_pods = cluster.k8sclient.pods.list_starts_with(
            "octavia-health-manager")
        octavia_pod = [op for op in octavia_pods
                       if op.data['spec']['node_name'] == k8s_node_name]
        if len(octavia_pod) > 0:
            LOG.info("octavia-health-manager was created")
            return True
        else:
            return False
    waiters.wait(
        lambda: wait_octavia_pod(k8s_node_name=new_node_k8s_name),
        timeout=600, interval=10,
        timeout_msg="Octavia pod is not created on the new node")
    oct_pods = cluster.k8sclient.pods.list_starts_with(
        "octavia-health-manager")
    octavia_hm_pod = [o for o in oct_pods
                      if o.data['spec']
                      ['node_name'] == new_node_k8s_name][0]
    LOG.info("Wait for octavia-health-manager pod in Ready state")
    try:
        octavia_hm_pod.wait_ready(timeout=150, interval=10)
    except exceptions.TimeoutError:
        LOG.error("Octavia pod is not ready. "
                  "Try to delete and wait again status (documentation step)")
        octavia_hm_pod.delete()
        time.sleep(10)
        oct_pods = cluster.k8sclient.pods.list_starts_with(
            "octavia-health-manager")
        octavia_hm_pod = [o for o in oct_pods
                          if o.data['spec']
                          ['node_name'] == new_node_k8s_name][0]
        octavia_hm_pod.wait_ready(timeout=300, interval=10)

    LOG.info("Verify that an OpenStack port for the node "
             "has been created and Active")
    client_pod = cluster.k8sclient.pods.list(
        namespace="openstack",
        name_prefix='keystone-client')
    assert len(client_pod) > 0, ("No pods found with prefix "
                                 "keystone-client in namespace "
                                 "openstack")
    client_pod = client_pod[0]
    port_cmd = ['/bin/sh', '-c', "openstack port show "
                                 "octavia-health-manager-listen-port-"
                                 f"{new_node_k8s_name} -f value -c status"]
    octavia_active_port = client_pod.exec(port_cmd).strip()
    assert octavia_active_port == "ACTIVE", \
        "Octavia listen port not found, or not not in status Active"
    LOG.info("Delete old ovtavia port")
    client_pod.exec(['/bin/sh', '-c',
                     "openstack port delete "
                     "octavia-health-manager-listen-port-"
                     f"{old_node_k8s_name}"])

    show_step(8)    # Identity and rotate stateful applications
    osdpl = cluster.k8sclient.openstackdeployment.get(osdpl_name,
                                                      namespace="openstack")
    storage_class = osdpl.data['spec']['local_volume_storage_class']
    osdpl_services_before_etcd = osdpl.data['spec']['services']
    etcd_statefulset = cluster.k8sclient.statefulsets.get(
        name="etcd-etcd",
        namespace="openstack")
    etcd_replicas = etcd_statefulset.data['spec']['replicas']
    etcd_statefulset.patch_scale(body={"spec": {"replicas": etcd_replicas-1}})
    etcd_statefulset.wait_readiness()

    LOG.info("Prepare etcd for restoring after the replacement")
    osdpl_backup_etcd = utils.merge_dicts(osdpl.data, {"spec": {"services": {
            "coordination": {"etcd": {"values": {"conf": {"etcd": {
                "ETCD_INITIAL_CLUSTER_STATE": "existing"}}}}}}}})
    cluster.k8sclient.openstackdeployment.update(osdpl_name,
                                                 namespace="openstack",
                                                 body=osdpl_backup_etcd)

    new_node_pvs = []
    all_pv = cluster.k8sclient.pvolumes.list_all()
    for pv in all_pv:
        if pv.data['spec'].get('node_affinity', {}):
            if pv.data['spec']['node_affinity']['required'][
                'node_selector_terms'][0]['match_expressions'][0][
                'values'][0] == new_node_k8s_name \
              and pv.data['spec']['storage_class_name'] == storage_class \
              and pv.data['status']['phase'] == "Available":
                LOG.info(f"Available PV found: {pv.data['metadata']['name']}")
                new_node_pvs.append(pv.data['metadata']['name'])

    old_node_pvs = [p for p in cluster.k8sclient.pvolumes.list_all()
                    if p.data['spec']['storage_class_name'] == storage_class
                    and p.data['spec']['node_affinity']['required'][
                    'node_selector_terms'][0]['match_expressions'][0][
                    'values'][0] == old_node_k8s_name]
    LOG.info(f"Available new PVs: {len(new_node_pvs)}; "
             f"Old PVCs: {len(old_node_pvs)}")
    assert len(old_node_pvs) >= 3,\
        "Should be minimum 3 stateful applications"
    assert len(old_node_pvs) <= len(new_node_pvs), \
        "No enough persistent volumes for all stateful applications"

    def pvc_create(name='', namespace='', pvname='', storclass=''):
        cluster.k8sclient.pvolumeclaims.create(
            name=name,
            namespace=namespace,
            body={
                "apiVersion": "v1",
                "kind": "PersistentVolumeClaim",
                "metadata": {
                    "name": name,
                    "namespace": namespace,
                },
                "spec": {
                    "accessModes": [
                        "ReadWriteOnce"
                    ],
                    "resources": {
                        "requests": {
                            "storage": "457Gi"
                        }
                    },
                    "storageClassName": storclass,
                    "volumeMode": "Filesystem",
                    "volumeName": pvname
                }
            })

    # Generate nev pvc with pv
    # Delete old pvc
    # Create new pvc with assigned pv from new node
    for old_pvc in old_node_pvs:
        pvc_name = old_pvc.data['spec']['claim_ref']['name']
        pvc_namespace = old_pvc.data['spec']['claim_ref']['namespace']
        pv_name = old_pvc.data['metadata']['name']
        pvc_pv = new_node_pvs[-1]
        new_node_pvs.pop()
        cluster.k8sclient.api_core.delete_namespaced_persistent_volume_claim(
            pvc_name,
            pvc_namespace)
        time.sleep(4)  # give time to deleting
        pvc_create(pvc_name, pvc_namespace, pvc_pv, storage_class)
        LOG.info(f"Pvc '{pvc_name}({pvc_namespace})' with old node "
                 f"pv({pv_name}) removed and created with the new "
                 f"pv({pvc_pv}")

    show_step(9)    # Force delete all pods from old node
    delete_pods = cluster.k8sclient.pods.list_all()
    try:
        for delete_pod in delete_pods:
            if delete_pod.data['spec']['node_name'] == old_node_k8s_name:
                LOG.info(f"Delete pod {delete_pod.data['metadata']['name']}")
                cluster.k8sclient.api_core.delete_namespaced_pod(
                    namespace=delete_pod.data['metadata']['namespace'],
                    name=delete_pod.data['metadata']['name'],
                    grace_period_seconds=0,
                    propagation_policy="Background")
    except Exception as e:
        if type(e).__name__ == 404:
            LOG.info("Skip deleting due to one of pods not found")

    # WA for ironic-conductor PRODX-14664
    iconductor_statefulset = cluster.k8sclient.statefulsets.get(
        name="ironic-conductor",
        namespace="openstack")
    iconductor_replicas = etcd_statefulset.data['spec']['replicas']
    iconductor_statefulset.patch_scale(
        body={"spec": {"replicas": 1}})
    iconductor_statefulset.wait_readiness()
    iconductor_statefulset.patch_scale(
        body={"spec": {"replicas": iconductor_replicas}})
    iconductor_statefulset.wait_readiness()

    cluster.check.check_k8s_pods()

    LOG.info("Restore etcd replicas")
    etcd_statefulset.patch_scale(body={"spec": {"replicas": etcd_replicas}})
    etcd_statefulset.wait_readiness()
    cluster.k8sclient.openstackdeployment.update(
        osdpl_name,
        namespace="openstack",
        body=osdpl_services_before_etcd)

    LOG.info("Check Openstack resources statuses")
    child_kubeconfig_name, child_kubeconfig = \
        cluster.get_kubeconfig_from_secret()
    with open('child_conf', 'w') as f:
        f.write(child_kubeconfig)
    os_manager = openstack_manager.OpenStackManager(
        kubeconfig='child_conf')
    os_manager.wait_all_osdpl_children_status()
    os_manager.wait_openstackdeployment_health_status()
    os_manager.wait_all_os_pods(timeout=900, interval=90)
