#!/usr/bin/env python3

import argparse
import os
import sys
from functools import lru_cache
from packaging import version

sys.path.append(os.getcwd())
try:
    from si_tests import settings
    from si_tests.managers.kaas_manager import Manager
except ImportError:
    print("ImportError: Run the application from the si-tests directory or "
          "set the PYTHONPATH environment variable to directory which contains"
          " ./si_tests")
    sys.exit(1)


def get_child_cluster_release_name(
        kaas_manager, release_name=settings.KAAS_CHILD_CLUSTER_RELEASE_NAME):
    """Get latest child release depending on keyword in release_name

    Recognisable keywords in release_name:

    * Empty string, 'default' or 'ucp' or 'mke' are keywords
      to get latest (by version) release with 'ucp-' or 'mke-' prefix in name
    * 'current' is a keyword to choose release from 'clusterRelease' field of
      KaasRelease
    * 'kubernetes', 'kubespray', 'k8s', 'kaas' are keywords to get latest
      (by version) release with 'kubernetes-' prefix in name
    * 'openstack' is a keyword to get latest (by version) release with
      'openstack-' prefix in name

    If none of above keywords provided release_name returned as is.
    """
    if not kaas_manager:
        kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    # Get KaasRelease
    mgm_cluster = kaas_manager.get_mgmt_cluster()
    deployed_kaas_release_name = mgm_cluster.spec[
        'providerSpec']['value']['kaas']['release']
    kaas_release = kaas_manager.get_kaasrelease(deployed_kaas_release_name)
    kaas_release = kaas_release.data
    supported_releases = kaas_release['spec']['supportedClusterReleases']
    # Try to guess correct child release name
    if not release_name or release_name in ['default', 'ucp', 'mke']:
        # Get latest ucp (mke) release,
        # which means release name has prefix 'ucp' or 'mke'
        ucp_releases = [r for r in supported_releases
                        if (r['name'].startswith('ucp-') or
                            r['name'].startswith('mke-'))]
        ucp_releases.sort(key=lambda r: version.Version(r['version']))
        release_name = ucp_releases[-1]['name']
    # Support plenty of magic words for the same thing
    elif release_name in ['kubernetes', 'kubespray', 'k8s', 'kaas']:
        # Get latest kubespray release, which means release name has prefix
        # 'kubernetes'
        kubernetes_releases = [r for r in supported_releases
                               if r['name'].startswith('kubernetes-')]
        kubernetes_releases.sort(key=lambda r: version.Version(r['version']))
        release_name = kubernetes_releases[-1]['name']
    elif release_name in ['current']:
        # Use clusterRelease from KaasRelease spec
        release_name = kaas_release['spec']['clusterRelease']
    elif release_name in ['openstack']:
        # Get latest openstack release, which means name has such prefix
        openstack_releases = [r for r in supported_releases
                              if r['name'].startswith('openstack-')]
        openstack_releases.sort(key=lambda r: version.Version(r['version']))
        release_name = openstack_releases[-1]['name']
    # Otherwise use release_name as is
    return release_name


def get_upgrade_path_combinations(**kwargs):
    """Generate upgrade path combinations from all to the latest version

    Produces 2^(n-1) lists of versions, where 'n' is amount
    of clusterreleases versions.

    :return: yields comma-separated string with release names
             where the first release is the initial version to deploy,
             and all others are the versions to upgrade in sequence.
    """
    def combinations(data):
        for i, d in enumerate(data):
            yield [d]
            for x in combinations(data[i+1:]):
                yield [d] + x

    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    cr_names = kaas_manager.get_clusterrelease_names()

    for x in combinations(cr_names[:-1]):
        path = x + [cr_names[-1]]
        yield ",".join(path)


def get_upgrade_path_single_sequence(**kwargs):
    """Generate upgrade path as a single sequence through all clusterreleases

    Produces release pairs like: "n,n+1,n+2,n+3,..."

    :return: yields comma-separated string with release name pairs
             where the first release is the initial version to deploy,
             and the second is the versions to upgrade.
    """
    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    cr_names = kaas_manager.get_clusterrelease_names()
    yield ",".join(cr_names)


def get_upgrade_path_single_sequence_pairs(**kwargs):
    """Generate upgrade path pairs

    Produces release pairs like: "n,n+1", "n+1,n+2", "n+2,n+3", ...

    :return: yields comma-separated string with release name pairs
             where the first release is the initial version to deploy,
             and the second is the versions to upgrade.
    """
    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    cr_names = kaas_manager.get_clusterrelease_names()
    for i in range(len(cr_names) - 1):
        yield f"{cr_names[i]},{cr_names[i+1]}"


def get_upgrade_path_minor_major_series(**kwargs):
    """Generate upgrade path respecting minor/major versions

    For kaas with zero-leading versions: 0.x.y
    For further kaas versions: x.y.y

    The minor versions sequence "y" is available only inside the constant
    major version "x":  0.x.y-2 > 0.x.y-1 > 0.x.y

    If the release version has the highest available minor version, it can
    be upgraded only to the highest available minor version of the next
    major release: 0.x-2.y > 0.x-1,y > 0.x.y

    For example, given that the clusterreleases are:
    - ['0.2.1', '0.2.2', '0.2.4', '0.3.1', '0.3.2', '0.3.4']

    , there will be two upgrade paths:
    - ['0.2.1', '0.2.2', '0.2.4', '0.3.4']  # latest 0.2.x -> latest 0.3.x
    - ['0.3.1', '0.3.2', '0.3.4']

    :return: yields comma-separated string with release name pairs
             where the first release is the initial version to deploy,
             and all the others are the versions to upgrade in a sequence.
    """
    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    crs = [cr.data for cr in kaas_manager.get_clusterreleases()]

    versions = {}
    for cr in crs:
        ver = cr['spec']['version'].split("+")[0]
        repr_ver = str(version.parse(ver))
        versions[repr_ver] = cr['metadata']['name']

    # sort x.y.z release versions
    pversions = sorted([version.parse(v) for v in versions.keys()])

    # generate version sequences inside each major release
    # keys: major release number,
    # values: list of release versions
    sequences = {}

    major_versions = []  # major versions in the increasing order
    for pver in pversions:
        ver = str(pver)

        if ver.startswith("0"):
            # 0.x.y : x =  major, y = minor version
            v = ver.split(".")
            major = ".".join(v[0:2])
        else:
            # x.y.y : x =  major, y.y = minor version
            v = ver.split(".")
            major = v[0]

        if major in sequences:
            sequences[major].append(ver)
        else:
            sequences[major] = [ver]
            major_versions.append(major)

    # Add the latest versions of each release sequence to the previous
    # major releases.
    # In the result, each sequences of previous major releases can
    # continue upgrade using the latest versions of the next major releases.
    major_versions.reverse()
    latest_versions = []
    for major in major_versions:
        latest_in_sequence = sequences[major][-1]
        sequences[major] += latest_versions
        latest_versions = [latest_in_sequence] + latest_versions

    for major in sequences:
        sequence = [versions[ver] for ver in sequences[major]]
        yield ",".join(sequence)


def get_upgrade_path_minor_major_pairs(**kwargs):
    """Generate upgrade path pairs respecting minor/major versions

    The same paths as get_upgrade_path_minor_major_series(), but
    in pairs of release: start child version, upgrade version

    :return: yields comma-separated string with release name pairs
             where the first release is the initial version to deploy,
             and the second is the versions to upgrade.
    """
    upgrade_paths = []
    for path in get_upgrade_path_minor_major_series():
        cr_names = path.split(",")
        for i in range(len(cr_names) - 1):
            p = f"{cr_names[i]},{cr_names[i+1]}"
            # filter out duplicate pairs of releases
            if p not in upgrade_paths:
                upgrade_paths.append(p)

    for upgrade_path in upgrade_paths:
        yield upgrade_path


def get_kaas_defined_release_upgrade_path_with_latest(**kwargs):
    for pair in get_kaas_defined_release_upgrade_path(include_latest=True,
                                                      **kwargs):
        yield pair


def get_ucp_defined_release_upgrade_path_with_latest(**kwargs):
    for pair in get_ucp_defined_release_upgrade_path(include_latest=True,
                                                     **kwargs):
        yield pair


def get_openstack_defined_release_upgrade_path_with_latest(**kwargs):
    for pair in get_openstack_defined_release_upgrade_path(
            include_latest=True, **kwargs):
        yield pair


def get_kaas_defined_release_upgrade_path(include_latest=False, **kwargs):
    def check_skipping_release(release):
        """Check if the release must be skipped"""
        # skip_list - list of release names or name prefixes
        # that should be skipped
        skip_list = [
            'mke-',
            'ucp-',
            'openstack-',
        ]
        return any((release.startswith(x) for x in skip_list))

    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    mgm_cluster = kaas_manager.get_mgmt_cluster()
    deployed_kaas_release_name = mgm_cluster.spec[
        'providerSpec']['value']['kaas']['release']
    kaas_release = kaas_manager.get_kaasrelease(deployed_kaas_release_name)
    kaas_release = kaas_release.data
    supported_releases = kaas_release['spec']['supportedClusterReleases']
    upgrade_paths = []
    for r in supported_releases:
        # var assignment for better understanding
        _current_version = r['version']
        _available_upgrade = r['availableUpgrades'][0]['version'] if \
            'availableUpgrades' in r else _current_version
        _supported_providers = r.get('providers', {}).get('supported', [])
        _needed_providers = kwargs.get('providers', {})

        # Upgrade sequences logic
        if _supported_providers:
            if not _needed_providers:
                raise ValueError(f"Release {r} has supported provider(s), "
                                 f"but pipeline ran current script without "
                                 f"--providers argument\n"
                                 f"Supported releases:\n"
                                 f"{supported_releases}")
            else:
                for provider in _needed_providers:
                    if provider in _supported_providers:
                        upgrade_paths.append((_current_version,
                                              _available_upgrade))
        else:
            upgrade_paths.append((_current_version,
                                  _available_upgrade))

    @lru_cache(maxsize=8)
    def get_releases_name(version):
        return next(r['name'] for r in supported_releases if
                    r['version'] == version)

    upgrade_paths = [(
        get_releases_name(upgrade_pair[0]),
        get_releases_name(upgrade_pair[1]))
        for upgrade_pair in upgrade_paths]

    for upgrade_pair in upgrade_paths:
        if not all(upgrade_pair):
            raise ValueError(f"Not all upgrades have a target: {upgrade_pair}")

    for upgrade_pair in upgrade_paths:
        if check_skipping_release(upgrade_pair[0]):
            print(f"Skip deploy for release: {upgrade_pair[0]}",
                  file=sys.stderr)
            continue
        elif check_skipping_release(upgrade_pair[1]):
            print(f"Skip upgrade to release: {upgrade_pair[1]}",
                  file=sys.stderr)
            yield f"{upgrade_pair[0]},{upgrade_pair[0]}"
        else:
            yield f"{upgrade_pair[0]},{upgrade_pair[1]}"


def get_ucp_defined_release_upgrade_path(include_latest=False, **kwargs):
    def check_skipping_release(release):
        """Check if the release must be skipped"""
        # skip_list - list of release names or name prefixes
        # that should be skipped
        skip_list = [
            'kubernetes-',
            'openstack-',
            'mosk-',
            'mos',
        ]
        return any((release.startswith(x) for x in skip_list))

    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    mgm_cluster = kaas_manager.get_mgmt_cluster()
    deployed_kaas_release_name = mgm_cluster.spec[
        'providerSpec']['value']['kaas']['release']
    kaas_release = kaas_manager.get_kaasrelease(deployed_kaas_release_name)
    kaas_release = kaas_release.data
    supported_releases = kaas_release['spec']['supportedClusterReleases']
    upgrade_paths = []
    for r in supported_releases:
        # var assignment for better understanding
        _current_version = r['version']
        _available_upgrade = r['availableUpgrades'][0]['version'] if \
            'availableUpgrades' in r else _current_version
        _supported_providers = r.get('providers', {}).get('supported', [])
        _needed_providers = kwargs.get('providers', {})

        # Upgrade sequences logic
        if _supported_providers:
            if not _needed_providers:
                raise ValueError(f"Release {r} has supported provider(s), "
                                 f"but pipeline ran current script without "
                                 f"--providers argument\n"
                                 f"Supported releases:\n"
                                 f"{supported_releases}")
            else:
                for provider in _needed_providers:
                    if provider in _supported_providers:
                        upgrade_paths.append((_current_version,
                                              _available_upgrade))
        else:
            upgrade_paths.append((_current_version,
                                  _available_upgrade))

    @lru_cache(maxsize=8)
    def get_releases_name(version):
        return next(r['name'] for r in supported_releases if
                    r['version'] == version)

    upgrade_paths = [(
        get_releases_name(upgrade_pair[0]),
        get_releases_name(upgrade_pair[1]))
        for upgrade_pair in upgrade_paths]

    for upgrade_pair in upgrade_paths:
        if not all(upgrade_pair):
            raise ValueError(f"Not all upgrades have a target: {upgrade_pair}")

    for upgrade_pair in upgrade_paths:
        if check_skipping_release(upgrade_pair[0]):
            print(f"Skip deploy for release: {upgrade_pair[0]}",
                  file=sys.stderr)
            continue
        elif check_skipping_release(upgrade_pair[1]):
            print(f"Skip upgrade to release: {upgrade_pair[1]}",
                  file=sys.stderr)
            yield f"{upgrade_pair[0]},{upgrade_pair[0]}"
        else:
            yield f"{upgrade_pair[0]},{upgrade_pair[1]}"


def get_openstack_defined_release_upgrade_path(include_latest=False, **kwargs):
    def check_skipping_release(release):
        """Check if the release must be skipped"""
        # skip_list - list of release names or name prefixes
        # that should be skipped
        skip_list = [
            'mke-',
            'ucp-',
            'kubernetes-',
        ]
        return any((release.startswith(x) for x in skip_list))

    kaas_manager = Manager(kubeconfig=settings.KUBECONFIG_PATH)
    mgm_cluster = kaas_manager.get_mgmt_cluster()
    deployed_kaas_release_name = mgm_cluster.spec[
        'providerSpec']['value']['kaas']['release']
    kaas_release = kaas_manager.get_kaasrelease(deployed_kaas_release_name)
    kaas_release = kaas_release.data
    supported_releases = kaas_release['spec']['supportedClusterReleases']
    upgrade_paths = []
    for r in supported_releases:
        # var assignment for better understanding
        _current_version = r['version']
        _available_upgrade = r['availableUpgrades'][0]['version'] if \
            'availableUpgrades' in r else _current_version
        _supported_providers = r.get('providers', {}).get('supported', [])
        _needed_providers = kwargs.get('providers', {})

        # Upgrade sequences logic
        if _supported_providers:
            if not _needed_providers:
                raise ValueError(f"Release {r} has supported provider(s), "
                                 f"but pipeline ran current script without "
                                 f"--providers argument\n"
                                 f"Supported releases:\n"
                                 f"{supported_releases}")
            else:
                for provider in _needed_providers:
                    if provider in _supported_providers:
                        upgrade_paths.append((_current_version,
                                              _available_upgrade))
        else:
            upgrade_paths.append((_current_version,
                                  _available_upgrade))

    @lru_cache(maxsize=8)
    def get_releases_name(version):
        return next(r['name'] for r in supported_releases if
                    r['version'] == version)

    upgrade_paths = [(
        get_releases_name(upgrade_pair[0]),
        get_releases_name(upgrade_pair[1]))
        for upgrade_pair in upgrade_paths]

    for upgrade_pair in upgrade_paths:
        if not all(upgrade_pair):
            raise ValueError(f"Not all upgrades have a target: {upgrade_pair}")

    for upgrade_pair in upgrade_paths:
        if check_skipping_release(upgrade_pair[0]):
            print(f"Skip deploy for release: {upgrade_pair[0]}",
                  file=sys.stderr)
            continue
        elif check_skipping_release(upgrade_pair[1]):
            print(f"Skip upgrade to release: {upgrade_pair[1]}",
                  file=sys.stderr)
            yield f"{upgrade_pair[0]},{upgrade_pair[0]}"
        else:
            yield f"{upgrade_pair[0]},{upgrade_pair[1]}"


def load_params():
    """
    Parse CLI arguments and environment variables

    Returns: ArgumentParser instance
    """
    parser = argparse.ArgumentParser(description=(
        'Generate upgrade paths for clusters to test cluster upgrades\n\n'
        'Returns multiple lines with comma-separated kaas clusterrelease\n'
        'versions, each line for a single '
        '"deploy->upgrade->upgrade->.." path.\n'
        'There are two upgrade strategies: inside the major release\n'
        'up to the latest minor version (X->Y->Z), and for the latest\n'
        'minor version of one major release to the latest minor version\n'
        'of the next major release:\n\n'
        '  A.X -> A.Y -> A.Z\n'
        '                 |\n'
        '                 v\n'
        '  B.X -> B.Y -> B.Z\n'
        '                 |\n'
        '                 v\n'
        '  C.X -> C.Y -> C.Z\n\n'
        'Valid --method options now are "minor_major_series" '
        'and "minor_major_pairs":\n\n'
        '"minor_major_series":\n'
        '  A-X,A-Y,A-Z,B-Z,C-Z\n'
        '  B-X,B-Y,B-Z,C-Z\n'
        '  C-X,C-Y,C-Z\n\n'
        '"minor_major_pairs":\n'
        '  A-X,A-Y\n'
        '  A-Y,A-Z\n'
        '  A-Z,B-Z\n'
        '  B-X,B-Y\n'
        '  B-Y,B-Z\n'
        '  B-Z,C-Z\n'
        '  C-X,C-Y\n'
        '  C-Y,C-Z\n\n'
        'There are several providers '
        'that can be included into upgrade procedure.\n'
        'By default upgrade path will consist of '
        'all possible upgrade sequences\n'
        'w/o releases with specified providers.\n'
        'But if you want to include some specific provider(s) \n'
        'you should use --providers argument\n'
        'and pass all needed providers separated by space.\n'
        'Valid --providers usage EXAMPLE is "--providers foo bar"\n'
    ), formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--method',
                        help=(f'Use one of the methods to generate paths, '
                              f'supported methods: {generate_methods.keys()}'),
                        default='kaas_defined_path_with_latest',
                        type=str)
    parser.add_argument('--oneline',
                        action='store_const',
                        const=";",
                        help='Show semicolon-separated release sequences '
                             'instead of newline-separated sequences',
                        default="\n")
    parser.add_argument('--providers',
                        help='Use space separated providers '
                             'from predefined list',
                        default="",
                        choices=['byo', 'baremetal', 'openstack', 'aws',
                                 'vsphere', 'equinixmetal'],
                        nargs="+",
                        type=str,
                        required=False)
    return parser


generate_methods = {
    'minor_major_series': get_upgrade_path_minor_major_series,
    'minor_major_pairs': get_upgrade_path_minor_major_pairs,
    'all_combinations': get_upgrade_path_combinations,
    'single_sequence': get_upgrade_path_single_sequence,
    'single_sequence_pairs': get_upgrade_path_single_sequence_pairs,
    'kaas_defined_path': get_kaas_defined_release_upgrade_path,
    'kaas_defined_path_with_latest':
        get_kaas_defined_release_upgrade_path_with_latest,
    'ucp_defined_path': get_ucp_defined_release_upgrade_path,
    'ucp_defined_release_upgrade_path_with_latest':
        get_ucp_defined_release_upgrade_path_with_latest,
    'openstack_defined_path': get_openstack_defined_release_upgrade_path,
    'openstack_defined_release_upgrade_path_with_latest':
        get_openstack_defined_release_upgrade_path_with_latest
}


def main():
    parser = load_params()
    opts = parser.parse_args()

    if not settings.KUBECONFIG_PATH or not os.path.isfile(
            settings.KUBECONFIG_PATH):
        print("Please set KUBECONFIG environment variable with "
              "the path to the kubeconfig file for KaaS management cluster")
        return 11

    if opts.method not in generate_methods.keys():
        print(f"Unsupported --method value, please use one of: "
              f"{generate_methods.keys()}")
        return 12

    method = generate_methods[opts.method]

    print(opts.oneline.join(method(providers=opts.providers)))


if __name__ == '__main__':
    sys.exit(main())
