#    Copyright 2019 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime, timezone
import si_tests.utils.templates as templates_utils
import yaml
from contextlib import contextmanager
from functools import wraps
import os
import shutil
from io import StringIO
import time
import string
import random
import requests
import paramiko
from retry import retry
from packaging import version
from si_tests import logger
from si_tests.utils import templates
from si_tests.settings import ARTIFACTS_DIR, ENV_NAME
from si_tests.managers import kaas_manager
from si_tests import settings
from threading import Thread, Event
import traceback

LOG = logger.logger

# kaas node-scope labeling
# Most labels hardcoded in kaas/core/frontend/src/consts.js or around
MACHINE_TYPES = {
    'cluster.sigs.k8s.io/control-plane': 'control',
    'hostlabel.bm.kaas.mirantis.com/controlplane': 'control',
    'hostlabel.bm.kaas.mirantis.com/worker': 'worker',
    'hostlabel.bm.kaas.mirantis.com/storage': 'storage',
}


class LogTime(object):
    def __init__(self, name, file_path):
        self.name = name
        self.file_path = file_path
        self.start_time = 0

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, x, y, z):
        end_time = time.time()
        total_time = end_time - self.start_time
        days = int(total_time // 86400)
        hours = int(total_time // 3600 % 24)
        minutes = int(total_time // 60 % 60)
        seconds = round(total_time % 60, 2)
        human_format = "{days}{hours}{minutes}{seconds}s".format(
            days=str(days) + "d " if days > 0 else "",
            hours=str(hours) + "h " if hours > 0 else "",
            minutes=str(minutes) + "m " if minutes > 0 else "",
            seconds=seconds)
        yaml_data = {
            "environment": ENV_NAME,
            "raw_duration": total_time,  # raw seconds
            "duration": human_format  # human readable format
        }
        with templates.YamlEditor(
                file_path=self.file_path) as editor:
            current_content = editor.content
            current_content[self.name] = yaml_data
            editor.content = current_content


class Tail(Thread):
    """
    Tail thread allows to watch remote file continuously in parallel
    with running main thread. Typical use case is watching remote log
    file (from bootstrap.sh) while main thread runs ansible script.

    Tail will read all available lines from the file during each _read()
    iteration and will try to read everything else after stop() is called
    to ensure the entire file was read.

    NOTE: It is not possible to interrupt the thread while it is in _read().
          Entire file will be read to its end if no exception occurs.
    """

    def __init__(self, filename, *args, **kwargs):
        self.filename = filename
        self.prefix = kwargs.pop('prefix', '')
        self.remote = kwargs.pop('remote', None)
        self.logger = kwargs.pop('logger', LOG)
        self.interval = float(kwargs.pop('interval', 10))
        self._stop_event = Event()
        self._position = 0
        super(Tail, self).__init__(*args, **kwargs)

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        self.join()

    def __str__(self):
        return "Tail (filename={}, interval={}, remote={})" \
            .format(self.filename, self.interval, self.remote is not None)

    def _open(self):
        if self.remote is None:
            return open(self.filename)
        else:
            return self.remote.open(self.filename)

    def _read(self):
        file_handler = None
        try:
            file_handler = self._open()
            file_handler.seek(self._position)

            for line in file_handler:
                self.logger.info(self.format(line.strip()))

            self._position = file_handler.tell()
        except OSError:
            self.logger.warning(
                self.format("Not found {}".format(self.filename)))
        except:  # noqa
            self.logger.error(self.format("Got Exception:",
                                          traceback.format_exc()))
            self.stop()
        finally:
            if file_handler:
                file_handler.close()

    def run(self):
        self.logger.info(self.format("Started {}".format(str(self))))
        try:
            while not self._stop_event.is_set():
                self._read()
                self._stop_event.wait(self.interval)
            self._read()
        finally:
            self.logger.info(self.format("Stopped {}".format(str(self))))

    def stop(self):
        self.logger.info(self.format("Stopping {} ...".format(str(self))))
        self._stop_event.set()

    def format(self, *args):
        return '{} {}'.format(self.prefix, '\n'.join(args))


@contextmanager
def pushd(path):
    current_dir = os.getcwd()
    try:
        os.chdir(os.path.expanduser(path))
        yield
    finally:
        os.chdir(current_dir)


def reduce_occurrences(items, text):
    """ Return string without items(substrings)
        Args:
            items: iterable of strings
            test: string
        Returns:
            string
        Raise:
            AssertionError if any substing not present in source text
    """
    for item in items:
        LOG.debug(
            "Verifying string {} is shown in "
            "\"\"\"\n{}\n\"\"\"".format(item, text))
        assert text.count(item) != 0
        text = text.replace(item, "", 1)
    return text


def generate_keys():
    file_obj = StringIO()
    key = paramiko.RSAKey.generate(1024)
    key.write_private_key(file_obj)
    public = key.get_base64()
    private = file_obj.getvalue()
    file_obj.close()
    return {'private': private,
            'public': public}


def load_keyfile(file_path):
    with open(file_path, 'r') as private_key_file:
        private = private_key_file.read()
    key = paramiko.RSAKey(file_obj=StringIO(private))
    public = key.get_base64()
    return {'private': private,
            'public': public}


def get_rsa_key(private_key):
    f = StringIO(private_key)
    return paramiko.rsakey.RSAKey.from_private_key(f)


def dump_keyfile(file_path, key):
    key = paramiko.RSAKey(file_obj=StringIO(key['private']))
    key.write_private_key_file(file_path)
    os.chmod(file_path, 0o644)


def clean_dir(dirpath):
    shutil.rmtree(dirpath)


def extract_name_from_mark(mark, info='name'):
    """Simple function to extract name from pytest mark

    :param mark: pytest.mark.MarkInfo
    :param info: Kwarg with information
    :rtype: string or None
    """
    if mark:
        if len(mark.args) > 0:
            return mark.args[0]
        elif info in mark.kwargs:
            return mark.kwargs[info]
    return None


def get_top_fixtures_marks(request, mark_name):
    """Order marks according to fixtures order

    When a test use fixtures that depend on each other in some order,
    that fixtures can have the same pytest mark.

    This method extracts such marks from fixtures that are used in the
    current test and return the content of the marks ordered by the
    fixture dependences.
    If the test case have the same mark, than the content of this mark
    will be the first element in the resulting list.

    :param request: pytest 'request' fixture
    :param mark_name: name of the mark to search on the fixtures and the test

    :rtype list: marks content, from last to first executed.
    """

    fixtureinfo = request.session._fixturemanager.getfixtureinfo(
        request.node, request.function, request.cls)

    top_fixtures_names = []
    for _ in enumerate(fixtureinfo.name2fixturedefs):
        parent_fixtures = set()
        child_fixtures = set()
        for name in sorted(fixtureinfo.name2fixturedefs):
            if name in top_fixtures_names:
                continue
            parent_fixtures.add(name)
            child_fixtures.update(
                fixtureinfo.name2fixturedefs[name][0].argnames)
        top_fixtures_names.extend(list(parent_fixtures - child_fixtures))

    top_fixtures_marks = []

    if mark_name in request.function.func_dict:
        # The top priority is the 'revert_snapshot' mark on the test
        top_fixtures_marks.append(
            extract_name_from_mark(
                request.function.func_dict[mark_name]))

    for top_fixtures_name in top_fixtures_names:
        fd = fixtureinfo.name2fixturedefs[top_fixtures_name][0]
        if mark_name in fd.func.func_dict:
            fixture_mark = extract_name_from_mark(
                fd.func.func_dict[mark_name])
            # Append the snapshot names in the order that fixtures are called
            # starting from the last called fixture to the first one
            top_fixtures_marks.append(fixture_mark)

    LOG.debug("Fixtures ordered from last to first called: {0}"
              .format(top_fixtures_names))
    LOG.debug("Marks ordered from most to least preffered: {0}"
              .format(top_fixtures_marks))

    return top_fixtures_marks


def gen_random_string(size):
    """Generate a random string of fixed length """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(size))


def print_pods_status(pods):
    """Print pods status in table format

    :param pods: list of K8sPod objects which produced
                 by kubectl_client.pods.list_all()
                 or kubectl_client.pods.list(namespace=...)
    """
    rows = [['NAMESPACE', 'NAME', 'READY', 'STATUS',
             'RESTARTS', 'AGE', 'IP', 'NODE', 'REASON']]
    for pod in pods:
        restarts = 0
        container_statuses = pod['status']['container_statuses'] or []
        total_containers = len(container_statuses)
        runnint_containers = sum([
            1 for x in container_statuses if x['ready'] is True
        ])
        restarts = sum([
            x['restart_count'] for x in container_statuses
        ])
        ready = "{0}/{1}".format(runnint_containers, total_containers)

        age = ''
        if pod['status']['start_time']:
            age_delta = datetime.now(timezone.utc) - \
                        pod['status']['start_time']
            m, s = divmod(age_delta.seconds, 60)
            h, m = divmod(m, 60)
            days = "{0}d ".format(age_delta.days) if age_delta.days else ''
            hours = "{0}h ".format(h) if h else ''
            minutes = "{0}m ".format(m) if (m and not days) else ''
            seconds = "{0}s".format(s) if (not days and not hours) else ''
            age = "{days}{hours}{minutes}{seconds}".format(
                days=days,
                hours=hours,
                minutes=minutes,
                seconds=seconds
            )

        rows.append([
            pod['metadata']['namespace'],
            pod['metadata']['name'],
            ready,
            pod['status']['phase'],
            str(restarts),
            str(age),
            pod['status']['pod_ip'] or '',
            pod['spec']['node_name'],
            pod['status']['reason'] or ''
        ])

    LOG.debug("Raw rows: {0}".format(rows))
    cols = zip(*rows)

    col_widths = [max(len(value or 'None') for value in col) for col in cols]

    format_str = '  '.join(['{{{0}!s:<{1}}}'.format(n, width)
                            for n, width in enumerate(col_widths)])

    LOG.debug("Format string for rows: {0}".format(format_str))
    LOG.info('\n' + '\n'.join([format_str.format(*row) for row in rows]))
    return rows


def make_export_env_strting(envs):
    envs_string = '; '.join(
        ["export {}='{}'".format(k, envs[k]) for k in envs])
    return envs_string


def merge_dicts(src, dst, path=None):
    path = path or []
    for key in dst:
        if key in src:
            if isinstance(src[key], dict) and isinstance(dst[key], dict):
                merge_dicts(src[key], dst[key], path + [str(key)])
            elif src[key] == dst[key]:
                pass  # same leaf value
            else:
                src[key] = dst[key]
        else:
            src[key] = dst[key]
    return src


def log_method_time():
    def log(func):
        @wraps(func)
        def wrapped(*args, **kwargs):
            with LogTime(func.__name__,
                         file_path=ARTIFACTS_DIR + 'time_spent.yaml'):
                result = func(*args, **kwargs)
                return result

        return wrapped

    return log


def get_expected_pods(path, target_nss=[]):
    LOG.info("Fetching lists of expected pods")
    template = templates_utils.render_template(path)
    expected_pods_file = yaml.load(template, Loader=yaml.SafeLoader)['ucp']
    expected_pods = {}
    for ns, pod_dict in expected_pods_file.items():
        if target_nss and ns not in target_nss:
            continue
        # we are in trouble if we have identical pod names
        # in different ns
        expected_pods.update(pod_dict)
    return expected_pods


def generate_list_pods(kubeconfig, target_nss=[]):
    kaas_mgr = kaas_manager.Manager(kubeconfig=kubeconfig)
    ns = kaas_mgr.get_namespace(settings.TARGET_NAMESPACE)
    cluster = ns.get_cluster(settings.TARGET_CLUSTER)
    LOG.info("Make sure we have all needed pods")
    cluster.check.check_actual_expected_pods()
    LOG.info("Generating list of pods for {0} cluster "
             "in {1} namespace".format(cluster.name, ns.name))
    kubectl_client = cluster.k8sclient
    ep = cluster.expected_pods
    expected_pods = {}
    for ns in ep:
        if not target_nss or ns in target_nss:
            pod_names = [x.split("/")[0] for x in ep[ns].keys()]
            for pod_name in sorted(pod_names, key=lambda x: len(x),
                                   reverse=True):
                pod_num = \
                    ep[ns][pod_name] if pod_name in ep[ns].keys()\
                    else ep[ns][f"{pod_name}/no_owner"]
                # order is guaranteed
                # https://stackoverflow.com/questions/39980323/are-dictionaries-ordered-in-python-3-6
                # https://docs.python.org/3.7/library/stdtypes.html#typesmapping
                expected_pods.update({pod_name: pod_num})
    actual_pods = []
    if target_nss:
        for ns in target_nss:
            actual_pods += kubectl_client.pods.list(
                namespace=ns,
                field_selector="status.phase=Running"
            )
    else:
        actual_pods = kubectl_client.pods.list_all(
            field_selector="status.phase=Running"
        )
    result = {}

    for pod_name, pod_num in expected_pods.items():
        pods = [x for x in actual_pods if x.name.startswith(pod_name)]
        # order is guaranteed
        result[pod_name] = (pod_name, pod_num, pods)
        if not target_nss:
            actual_pods = [x for x in actual_pods
                           if not x.name.startswith(pod_name)]
        else:
            actual_pods = [x for x in actual_pods if x.namespace in target_nss
                           and not x.name.startswith(pod_name)]

    LOG.debug("Generated list: {}".format(result))
    return result


@retry(requests.exceptions, delay=5, tries=5, logger=LOG)
def get_latest_k8s_image_version(url, major_version):
    LOG.info("Fetching list of available k8s conformance "
             "images from {}".format(url))
    resp = requests.get(url, verify=False)
    if not resp.ok:
        raise Exception("Request failed {}".format(resp.text))
    lst = []
    for txt in resp.text.split("\n"):
        if major_version in txt and 'att' not in txt:
            lst.append(
                txt.split('<a href="v')[1].split('/">v')[0])
    LOG.info("Available versions: {0}. "
             "Filtered by {1}".format(lst, major_version))
    return str(max([version.parse(x) for x in lst])).replace(".post", "-")


def get_labels_by_type(machine_type):
    """
    labels only in skope of MACHINE_TYPES
    terms kaas
    return: formatted dict with items
    """
    labels = dict()
    for label, m_type in MACHINE_TYPES.items():
        if m_type == machine_type:
            labels[label] = m_type
    return labels


def get_type_by_labels(labels):
    """
    labels only in scope of MACHINE_TYPES
    terms kaas.
    Types only for scope k8s node.(worker/control)
    """
    # Set default type to "worker" because worker nodes
    # don't have the related label in the openstack provider
    _machine_type = 'worker'
    # For bm case we may have multiple types for machine
    multiple_types = get_types_by_labels(labels)
    if 'worker' in multiple_types and 'storage' in multiple_types:
        _machine_type = 'worker'
    elif multiple_types:
        _machine_type = multiple_types[0]
    else:
        LOG.warning(f"Using default machine type:{_machine_type}")
    return _machine_type


def get_types_by_labels(labels):
    """
    labels and types in scope of MACHINE_TYPES
    terms kaas.
    return: list MACHINE_TYPES.values() item/items
    """
    # For bm case we may have multiple types for machine
    multiple_types = []
    for label, machine_type in MACHINE_TYPES.items():
        # label value might be in ['machinetype', 'true', str()]
        if label in labels and (any([labels[label] == machine_type,
                                     labels[label] == 'true',
                                     len(str(labels[label])) >= 1])):
            multiple_types.append(MACHINE_TYPES[label])
    if not multiple_types:
        # Workaround for missing labels for node role on worker nodes
        multiple_types = ['worker']
        LOG.warning(f"Looks like passed 'machine' dont have "
                    f"expected 'labels' for kaas. "
                    f"Using by default: {multiple_types}")
    return sorted(set(multiple_types))


def render_bmh_name(name, cname, uefi, labels, si_roles=False):
    """
    just to re-use one predictable name across all tests
    """
    if not si_roles:
        si_roles = []
    return '-'.join([name,
                     cname,
                     *get_types_by_labels(labels),
                     'efi' if uefi else 'noefi',
                     *si_roles,
                     ])


def get_available_equinix_facility(api_token, machine_type, machines_count):
    LOG.info("Fetching available Equinix Metal facility "
             "to create {} servers".format(machines_count))
    facilities = ['am6', 'sv15', 'dc13', 'da11', 'fr2',
                  'ch3', 'la4', 'ny5', 'sl1', 'sg1']
    desired_capacity = {
        'servers': [
            {
                'plan': machine_type,
                'quantity': machines_count
            }
        ]
    }

    for facility in facilities:
        try:
            if equinix_facility_capable(api_token, facility, desired_capacity):
                LOG.info("{} facility suitable for {} servers"
                         .format(facility, machines_count))
                return facility
        except Exception as err:
            LOG.warning("{} facility unavailable: {}, skipping"
                        .format(facility, err))

    raise Exception("Failed to get available Equinix facility for {}"
                    " instances".format(machines_count))


@retry(Exception, delay=3, tries=3, logger=LOG)
def equinix_facility_capable(api_token, facility_name, servers_meta):
    LOG.info("Checking hardware availability in facility {}"
             .format(facility_name))
    servers_meta['servers'][0]['facility'] = facility_name
    headers = {'Accept': 'application/json', 'X-Auth-Token': api_token,
               'Content-Type': 'application/json'}

    resp = requests.post('https://api.equinix.com/metal/v1/capacity',
                         headers=headers, json=servers_meta)
    if not resp.ok:
        raise Exception("Request failed: {}".format(resp.text))

    result = resp.json()['servers'][0]
    return result['available']
